/************************************************************************
 *   IRC - Internet Relay Chat, include/standards.h
 *   Copyright (C) 1990 Jarkko Oikarinen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef	__config_include__
#define	__config_include__

/**** MAKE SURE THIS IS CORRECT ****/
/*
 * ONLY EDIT "HARD_FDLIMIT_" and "INIT_MAXCLIENTS" 
 */
#define HARD_FDLIMIT_   2048
#define INIT_MAXCLIENTS 2000

/*
 * This is how many 'buffer connections' we allow... Remember,
 * MAX_BUFFER + MAX_CLIENTS can't exceed HARD_FDLIMIT :)
 */
#define MAX_BUFFER	48

/* Also define this for SunOS 4.1.4_DBE */
#undef SUNDBE
#ifdef SUNDBE
#define IRC_FD_SETSIZE HARD_FDLIMIT_
#define NOFILE HARD_FDLIMIT_
#endif /* SUNDBE */

/*
 * DPATH SPATH CPATH MPATH KPATH - directoy and files locations Full
 * pathnames and defaults of irc system's support files. Please note
 * that these are only the recommended names and paths. Change as
 * needed. You must define these to something, even if you don't really
 * want them.
 * 
 * DPATH = directory, SPATH = server executable, CPATH = conf file, MPATH
 * = MOTD KPATH = kline conf file
 * 
 * For /restart to work, SPATH needs to be a full pathname (unless "." is
 * in your exec path). -Rodder Leave KPATH undefined if you want klines
 * in main conf file. HPATH is the opers help file, seen by opers on
 * /quote help.
 * 
 * -Dianora
 */
#define DPATH   "/usr/home/ircd/ircd/"
#define SPATH   "/usr/home/ircd/ircd/ircd"

/*
 * SHORT_MOTD There are client ignoring the FORCE_MOTD MOTD numeric,
 * there is no point forcing MOTD on connecting clients IMO. Give them
 * a short NOTICE telling them they should read the motd, and leave it
 * at that.
 */
#undef SHORT_MOTD

/*
 * HUB - enable server-server routing If your server is running as a a
 * HUB Server then define this. A HUB Server has many servers connect
 * to it at the same as opposed to a leaf which just has 1 server
 * (typically the uplink). Define this correctly for performance
 * reasons.
 */
#undef HUB

/* The following two are defaults only and may be overridden
 * with T: lines in the configuration file
 */

/* Define this if you want a notice sent to connecting users
 * about wingate scanning bots.  A URL is included in the 
 * notice
 */
#define WINGATE_NOTICE
/* For this to compile, you must define the hostname of your 
 * wingate bot - which is included in the notice.
 */
#ifdef WINGATE_NOTICE
#define MONITOR_HOST "bots.someisp.com"
#endif

/*
 * USE_SYSLOG - log errors and such to syslog() If you wish to have the
 * server send 'vital' messages about server through syslog, define
 * USE_SYSLOG. Only system errors and events critical to the server are
 * logged although if this is defined with FNAME_USERLOG, syslog() is
 * used instead of the above file. It is not recommended that this
 * option is used unless you tell the system administrator beforehand
 * and obtain their permission to send messages to the system log
 * files.
 * 
 * IT IS STRONGLY RECOMMENDED THAT YOU *DO* USE SYSLOG.  Many fatal ircd
 * errors are only logged to syslog.
 */
#define	USE_SYSLOG
#ifdef	USE_SYSLOG
/*
 * SYSLOG_KILL SYSLOG_SQUIT SYSLOG_CONNECT SYSLOG_USERS SYSLOG_OPER If
 * you use syslog above, you may want to turn some (none) of the
 * spurious log messages for KILL,SQUIT,etc off.
 */
#undef	SYSLOG_KILL		/* log all operator kills */
#undef	SYSLOG_SQUIT		/* log all remote squits */
#undef	SYSLOG_CONNECT		/* log remote connect messages */
#undef	SYSLOG_USERS		/* send userlog stuff to syslog */
#undef	SYSLOG_OPER		/* log all users who successfully oper */
#undef   SYSLOG_BLOCK_ALLOCATOR	/* debug block allocator */

/*
 * LOG_FACILITY - facility to use for syslog() Define the facility you
 * want to use for syslog().  Ask your sysadmin which one you should
 * use.
 */
#define LOG_FACILITY LOG_LOCAL4
#endif /* USE_SYSLOG */

/*
 * MAXSENDQLENGTH - Max amount of internal send buffering Max amount of
 * internal send buffering when socket is stuck (bytes)
 */
#define MAXSENDQLENGTH 5050000	/* Recommended value: 5050000 for efnet 
				 */

/*
 * BUFFERPOOL - the maximum size of the total of all sendq's.
 * Recommended value is four times MAXSENDQLENGTH.
 */
#define	BUFFERPOOL     (4 * MAXSENDQLENGTH)

/* 
 * HIDEULINEDSERVS 
 * Define this if you want to hide the location of U:lined servers (and 
 * then clients on them) from nonopers. With this defined, no non-oper 
 * should be able to find out which server the U:lined server is connected
 * to. If you are connected to the main DALnet network, you MUST have this
 * enabled.
 */
#define HIDEULINEDSERVS 1

/*
 * comstud and I have both noted that solaris 2.5 at least, takes a
 * hissy fit if you don't read a fd that becomes ready right away.
 * Unfortunately the dog3 priority code relies upon not having to read
 * a ready fd right away. If you have HTM mode set low as it is
 * normally, the server will eventually grind to a halt. Personally, I
 * think the server is faster without some of the CPU expensive
 * manipulation some of the priority code does. Your choice. but it has
 * to be defined for SOLARIS Try it and see. compare. Don't complain if
 * Solaris lags if you don't define this. I warned you.
 * 
 * -Dianora
 */
#ifdef OS_SOLARIS
#define NO_PRIORITY
#else
#undef NO_PRIORITY
#endif

/* Services Name */
#define SERVICES_NAME "services.irctr.gen.tr"	/* DALnet services */

/* 
    * CRYPT_OPER_PASSWORD - use crypted oper passwords in the ircd.conf 
    * define this if you want to use crypted passwords for operators in 
    * your ircd.conf file. 
    */

#undef  CRYPT_OPER_PASSWORD

#define MAXCHANNELSPERUSER  10	/* Recommended value: 10 */

#define USE_FAKEHOST

/* these should be set to your local and global kline addresses */

#ifndef SERVER_KLINE_ADDRESS
#define SERVER_KLINE_ADDRESS "bad_admin@poorly.configured.server.com"
#endif

#ifndef NETWORK_KLINE_ADDRESS
#define NETWORK_KLINE_ADDRESS "bad_admin@poorly.configured.network.com"
#endif

