/************************************************************************
 *   IRC - Internet Relay Chat, include/struct.h
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

/* $Id: struct.h,v 1.18 2001/11/24 23:01:18 tr-ircd Exp $ */

#ifndef	__struct_include__
#define __struct_include__

#include "config.h"
#if !defined(CONFIG_H_LEVEL_12)
#error Incorrect config.h for this revision of ircd.
#endif

#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#ifdef OSF
#include "/usr/include/netdb.h"
#else
#include <netdb.h>
#endif
#if defined( HAVE_STDDEF_H )
#include <stddef.h>
#endif
#ifdef ORATIMING
#include <sys/time.h>
#endif

#ifdef USE_SYSLOG
#include <syslog.h>
#if defined( HAVE_SYS_SYSLOG_H )
#include <sys/syslog.h>
#endif
#endif

#define REPORT_DO_DNS_		":%s NOTICE AUTH :*** Looking up your hostname...\n"
#define REPORT_FIN_DNS_		":%s NOTICE AUTH :*** Found your hostname\n"
#define REPORT_FIN_DNSC_	":%s NOTICE AUTH :*** Found your hostname, cached\n"
#define REPORT_FAIL_DNS_	":%s NOTICE AUTH :*** Couldn't look up your hostname\n"
#define REPORT_DO_ID_		":%s NOTICE AUTH :*** Checking Ident\n"
#define REPORT_FIN_ID_		":%s NOTICE AUTH :*** Got Ident response\n"
#define REPORT_FAIL_ID_		":%s NOTICE AUTH :*** No Ident response\n"

extern char REPORT_DO_DNS[256], REPORT_FIN_DNS[256],
   REPORT_FIN_DNSC[256], REPORT_FAIL_DNS[256], REPORT_DO_ID[256],
   REPORT_FIN_ID[256], REPORT_FAIL_ID[256];

#include "hash.h"

typedef struct ConfItem aConfItem;
typedef struct Client aClient;
typedef struct Channel aChannel;
typedef struct User anUser;
typedef struct Server aServer;
typedef struct Service aService;
typedef struct SLink Link;
typedef struct ChanLink chanMember;
typedef struct SMode Mode;
typedef struct Watch aWatch;
typedef struct Ban aBan;
typedef struct Invite anInvite;
typedef struct BanEx aBanEx;
typedef struct ListOptions LOpts;
typedef long ts_val;

typedef struct MotdItem aMotd;

#include "class.h"
#include "dbuf.h"		/* THIS REALLY SHOULDN'T BE HERE!!!
				 * --msa */

#define	HOSTLEN		63	/* Length of hostname.  Updated to */

/* comply with RFC1123 */

#define HOSTIPLEN	15	/* Length of dotted quad form of IP */

/* - Dianora  */

#define	NICKLEN		30

/* Necessary to put 9 here instead of 10  if  
 * s_msg.c/m_nick has been corrected.  This 
 * preserves compatibility with old * servers --msa 
 */

#define MAX_DATE_STRING 32	/* maximum string length for a date
				 * string */

#define	USERLEN		    10
#define	REALLEN	 	    50
#define	TOPICLEN	    307
#define KILLLEN		    400
#define	CHANNELLEN	    32
#define	PASSWDLEN 	    63
#define	KEYLEN		    23
#define	BUFSIZE		    512	/* WARNING: *DONT* CHANGE THIS!!!! */
#define	MAXRECIPIENTS       20
#define	MAXBANS	 	    100
#define MAXINVITES          100
#define MAXEXCEPTIONS       100

#define MOTDLINELEN	    90

#define MAXSILES            10
#define MAXSILELENGTH       128

#define MAXDCCALLOW         10

#define DCC_LINK_ME	0x01	/* This is my dcc allow */
#define DCC_LINK_REMOTE 0x02	/* I need to remove these dcc allows
				 * from these clients when I die */

#define	USERHOST_REPLYLEN	(NICKLEN+HOSTLEN+USERLEN+5)

/*
 * 'offsetof' is defined in ANSI-C. The following definition * is not
 * absolutely portable (I have been told), but so far * it has worked
 * on all machines I have needed it. The type * should be size_t but...
 * --msa
 */

#ifndef offsetof
#define	offsetof(t,m) (int)((&((t *)0L)->m))
#endif

#define	elementsof(x) (sizeof(x)/sizeof(x[0]))

/* flags for bootup options (command line flags) */

#define	BOOT_CONSOLE 	 1
#define	BOOT_QUICK	 2
#define	BOOT_DEBUG	 4
#define	BOOT_INETD	 8
#define	BOOT_TTY	 16
#define	BOOT_OPER	 32
#define	BOOT_AUTODIE 	 64
#define BOOT_STDERR	 128
#define BOOT_HIDELINKS	 256	/* uses the new /links */

#define	STAT_LOG	-6	/* logfile for -x */
#define	STAT_MASTER	-5	/* Local ircd master before
				 * identification */
#define	STAT_CONNECTING	-4
#define	STAT_HANDSHAKE	-3
#define	STAT_ME		-2
#define	STAT_UNKNOWN	-1
/* the line of truth lies here (truth == registeredness) */
#define	STAT_SERVER	0
#define	STAT_CLIENT	1
#define STAT_SERVICE    2

/* status macros. */

#define	IsRegisteredUser(x)	((x)->status == STAT_CLIENT)
#define	IsRegistered(x)		((x)->status >= STAT_SERVER)
#define	IsConnecting(x)		((x)->status == STAT_CONNECTING)
#define	IsHandshake(x)		((x)->status == STAT_HANDSHAKE)
#define	IsMe(x)			((x)->status == STAT_ME)
#define	IsUnknown(x)		((x)->status == STAT_UNKNOWN || (x)->status == STAT_MASTER)
#define	IsServer(x)		((x)->status == STAT_SERVER)
#define	IsClient(x)		((x)->status == STAT_CLIENT)
#define	IsLog(x)		((x)->status == STAT_LOG)
#define IsService(x)		((x)->status == STAT_SERVICE)

#define	SetMaster(x)		((x)->status = STAT_MASTER)
#define	SetConnecting(x)	((x)->status = STAT_CONNECTING)
#define	SetHandshake(x)		((x)->status = STAT_HANDSHAKE)
#define	SetMe(x)		((x)->status = STAT_ME)
#define	SetUnknown(x)		((x)->status = STAT_UNKNOWN)
#define	SetServer(x)		((x)->status = STAT_SERVER)
#define	SetClient(x)		((x)->status = STAT_CLIENT)
#define	SetLog(x)		((x)->status = STAT_LOG)
#define SetService(x)           ((x)->status = STAT_SERVICE)

#define	FLAGS_PINGSENT     0x000001	/* Unreplied ping sent */
#define	FLAGS_DEADSOCKET   0x000002	/* Local socket is
					 * dead--Exiting soon */
#define	FLAGS_KILLED       0x000004	/* Prevents "QUIT" from being
					 * sent for this */
#define	FLAGS_BLOCKED      0x000008	/* socket is in a blocked
					 * condition */
#define FLAGS_REJECT_HOLD  0x000010	/* client has been klined */
#define	FLAGS_CLOSING      0x000020	/* set when closing to suppress 
					 * errors */
#define	FLAGS_LISTEN       0x000040	/* used to mark clients which
					 * we listen() on */
#define	FLAGS_CHKACCESS    0x000080	/* ok to check clients access
					 * if set */
#define	FLAGS_DOINGDNS	   0x000100	/* client is waiting for a DNS
					 * response */
#define	FLAGS_AUTH	   0x000200	/* client is waiting on rfc931
					 * response */
#define	FLAGS_WRAUTH	   0x000400	/* set if we havent writen to
					 * ident server */
#define	FLAGS_LOCAL	   0x000800	/* set for local clients */
#define	FLAGS_GOTID	   0x001000	/* successful ident lookup
					 * achieved */
#define	FLAGS_DOID	   0x002000	/* I-lines say must use ident
					 * return */
#define	FLAGS_NONL	   0x004000	/* No \n in buffer */
#define FLAGS_NORMALEX     0x008000	/* Client exited normally */
#define FLAGS_SENDQEX      0x010000	/* Sendq exceeded */
#define FLAGS_IPHASH       0x020000	/* iphashed this client */
#define FLAGS_ULINE 	   0x040000	/* client is U-lined */
#define FLAGS_USERBURST	   0x080000	/* server in nick/channel
					 * netburst */
#define FLAGS_TOPICBURST   0x100000	/* server in topic netburst */
#define FLAGS_BURST	(FLAGS_USERBURST | FLAGS_TOPICBURST)
#define FLAGS_SOBSENT      0x200000	/* we've sent an SOB, just have 
					 * to send an EOB */
#define FLAGS_EOBRECV      0x400000	/* we're waiting on an EOB */
#define FLAGS_BAD_DNS	   0x800000	/* spoofer-guy */
#define FLAGS_SERV_NEGO	   0x1000000	/* This is a server that has
					 * passed connection tests, but 
					 * is a stat < 0 for handshake
					 * purposes */
#define FLAGS_RC4IN        0x2000000	/* This link is rc4 encrypted. */
#define FLAGS_RC4OUT       0x4000000	/* This link is rc4 encrypted. */
#define FLAGS_ZIPPED_IN	   0x8000000	/* This link is gzipped. */
#define FLAGS_ZIPPED_OUT   0x10000000	/* This link is gzipped. */


/* Capabilities of the ircd or clients */

#define CAPAB_TS3     0x0000001	/* Supports the TS3 Protocal */
#define CAPAB_NOQUIT  0x0000002	/* Supports NOQUIT */
#define CAPAB_NSJOIN  0x0000004	/* server supports new smart sjoin */
#define CAPAB_BURST   0x0000008	/* server supports BURST command */
#define CAPAB_UNCONN  0x0000010	/* server supports UNCONNECT */
#define CAPAB_DKEY    0x0000020	/* server supports dh-key exchange
				 * using "DKEY" */
#define CAPAB_ZIP     0x0000040	/* server supports gz'd links */
#define CAPAB_DOZIP   0x0000080	/* output to this link shall be gzipped 
				 */
#define CAPAB_DODKEY  0x0000100	/* do I do dkey with this link? */
#define CAPAB_NICKIP  0x0000200	/* IP in the NICK line? */

#define CAPAB_DTONE     0x0001000	/* DarkstarTimeMr14C Protocol
					 * Change */
#define CAPAB_SERVICES  0x0002000	/* Server is capable of SERVICE 
					 */
#define CAPAB_EXCLUDE   0x0004000	/* Capable of autokill
					 * exclusions */
#define CAPAB_CANHIDE	0x0008000	/* Capable of server name 
					 * stealth */
#define CAPAB_HIDENAME	0x0010000	/* The name of this server
					 * may be hidden */

#define IsTS3       	((x)->capabilities & CAPAB_TS3)
#define IsNoQuit(x) 	((x)->capabilities & CAPAB_NOQUIT)
#define IsSSJoin(x)	((x)->capabilities & CAPAB_NSJOIN)
#define IsBurst(x)	((x)->capabilities & CAPAB_BURST)
#define IsUnconnect(x)	((x)->capabilities & CAPAB_UNCONN)
#define CanDoDKEY(x)    ((x)->capabilities & CAPAB_DKEY)
#define WantDKEY(x)	((x)->capabilities & CAPAB_DODKEY)	
			/* N:Line flag E */
#define IsZipCapable(x)	((x)->capabilities & CAPAB_ZIP)
#define DoZipThis(x) 	((x)->capabilities & CAPAB_DOZIP)	
			/* N:Line flag Z */
#define IsNICKIP(x)     ((x)->capabilities & CAPAB_NICKIP)
#define IsDTone(x)      ((x)->capabilities & CAPAB_DTONE)
#define HasServices(x)  ((x)->capabilities & CAPAB_SERVICES)
#define CanExclude(x)   ((x)->capabilities & CAPAB_EXCLUDE)
#define CanHide(x)	((x)->capabilities & CAPAB_CANHIDE)
#define WillHide(x)	((x)->capabilities & CAPAB_HIDENAME)
#define SetHidename(x)	((x)->capabilities |= CAPAB_HIDENAME)

#define IsULine(x) ((x)->flags & FLAGS_ULINE)

/* defines for check_ctcp results */
#define CTCP_NONE       0
#define CTCP_YES        1
#define CTCP_DCC        2
#define CTCP_DCCSEND    3

/* User Modes */
#define UMODE_o     0x00001	/* umode +o - Oper */
#define UMODE_O     0x00002	/* umode +O - Local Oper */
#define UMODE_i     0x00004	/* umode +i - Invisible */
#define UMODE_w     0x00008	/* umode +w - Get wallops */
#define UMODE_r     0x00010	/* umode +r - registered nick */
#define UMODE_g     0x00020	/* umode +g - Globops */
#define UMODE_a     0x00040	/* umode +a - Services Admin */
#define UMODE_A     0x00080	/* umode +A - Server Admin */
#define UMODE_n     0x00100	/* umode +n - Routing Notices */
#define UMODE_h     0x00200	/* umode +h - Helper */
#define UMODE_R     0x00400	/* umode +R - Routing staff -REMOVED
				 * -READDED: listen to registered */
#define UMODE_D     0x00800	/* umode +D - pseudo/hidden, has seen
				 * dcc warning message */
#define UMODE_x     0x01000	/* umode +x - hidden hostnames */
#define UMODE_q     0x02000	/* umode +q - anonymous nicknames */
#define UMODE_z     0x04000	/* Umode +z - zombie clients */

/* for sendto_ops_lev */

#define CCONN_LEV	1
#define REJ_LEV		2
#define SKILL_LEV	3
#define SPY_LEV		4
#define DEBUG_LEV	5
#define FLOOD_LEV 	6
#define SPAM_LEV 	7
#define DCCSEND_LEV	8
#define SERVICE_LEV	9	/* TRIRCD */
#define ULINE_LEV       10	/* TRIRCD */

#define T_CCONN_LEV     "SERVER Channel: Clients connecting and exiting"
#define T_REJ_LEV       "SERVER Channel: Rejected connections"
#define T_SKILL_LEV     "SERVER Channel: Server Kills"
#define T_SPY_LEV       "SERVER Channel: Issued stats/info/admin/trace/links commands"
#define T_DCCSEND_LEV   "SERVER Channel: Blocked dcc sends"
#define T_FLOOD_LEV     "SERVER Channel: Flood notices"
#define T_SPAM_LEV      "SERVER Channel: Possible Spambots"
#define T_DEBUG_LEV     "SERVER Channel: Debug information notices"
#define T_NOTICE_LEV    "SERVER Channel: Server notices"
#define T_SERVICE_LEV   "SERVER Channel: Notices about connecting services"

#define N_NOTICE        "&NOTICE"
#define N_CONNECT       "&CONNECTS"
#define N_REJECT        "&REJECTED"
#define N_KILL          "&KILLS"
#define N_STAT          "&STATS"
#define N_DCCSEND       "&DCCSEND"
#define N_FLOOD         "&FLOOD"
#define N_SPAMBOT       "&SPAMBOTS"
#define N_DEBUG         "&DEBUG"
#define N_ULINE         "#services"
#define N_SERVICE       "&SERVICES"

#define	SEND_UMODES (UMODE_i|UMODE_o|UMODE_w|UMODE_r|UMODE_a|UMODE_A|UMODE_h|UMODE_x|UMODE_R)
#define ALL_UMODES (SEND_UMODES|UMODE_g|UMODE_n|UMODE_O)

#ifdef DEFAULT_HELP_MODE
#define OPER_UMODES (UMODE_o|UMODE_w|UMODE_g|UMODE_n|UMODE_h)
#else
#define OPER_UMODES (UMODE_o|UMODE_w|UMODE_g|UMODE_n)
#endif

#define LOCOP_UMODES (UMODE_O|UMODE_w|UMODE_g|UMODE_n|UMODE_h)

#ifdef USE_FAKEHOST
#define       UFLAGS  (UMODE_i|UMODE_w|UMODE_x)
#else
#define       UFLAGS  (UMODE_i|UMODE_w)
#endif

#define	FLAGS_ID (FLAGS_DOID|FLAGS_GOTID)

#define	IsOper(x)		((x)->umode & UMODE_o)
#define	IsLocOp(x)		((x)->umode & UMODE_O)
#define	IsInvisible(x)		((x)->umode & UMODE_i)
#define	IsAnOper(x)		((x)->umode & (UMODE_o|UMODE_O))
#define IsRegNick(x)    	((x)->umode & UMODE_r)
#define IsSAdmin(x)   		((x)->umode & UMODE_a)
#define IsAdmin(x)    		((x)->umode & UMODE_A)
#define IsUmoden(x)   		((x)->umode & UMODE_n)
#define IsUmodeh(x)   		((x)->umode & UMODE_h)
#define IsNoNonReg(x)           ((x)->umode & UMODE_R)
#define SeenDCCNotice(x)     	((x)->umode & UMODE_D)
#define	SendWallops(x)		((x)->umode & UMODE_w)
#define SendGlobops(x)	 	((x)->umode & UMODE_g)
#define SendRnotice(x) 		((x)->umode & UMODE_n)
#define IsZombie(x)     	((x)->umode & UMODE_z)
#define IsFake(x)       	((x)->umode & UMODE_x)
#define IsAnon(x)       	((x)->umode & UMODE_q)

#define	SetOper(x)		((x)->umode |= UMODE_o)
#define SetRegNick(x)   	((x)->umode |= UMODE_r)
#define SetSAdmin(x)  		((x)->umode |= UMODE_a)
#define	SetLocOp(x)    		((x)->umode |= UMODE_O)
#define	SetInvisible(x)		((x)->umode |= UMODE_i)
#define	SetWallops(x)  		((x)->umode |= UMODE_w)
#define SetDCCNotice(x)         ((x)->umode |= UMODE_D)
#define SetAnon(x)      	((x)->umode |= UMODE_q)
#define SetZombie(x)    	((x)->umode |= UMODE_z)

#define ClearSAdmin(x)          ((x)->umode &= ~UMODE_a)
#define ClearAdmin(x)           ((x)->umode &= ~UMODE_A)
#define ClearUmoden(x)          ((x)->umode &= ~UMODE_n)
#define ClearUmodeh(x)          ((x)->umode &= ~UMODE_h)
#define ClearRouting(x)         ((x)->umode &= ~UMODE_R)
#define ClearOper(x)            ((x)->umode &= ~UMODE_o)
#define ClearLocOp(x)           ((x)->umode &= ~UMODE_O)
#define ClearInvisible(x)       ((x)->umode &= ~UMODE_i)
#define ClearWallops(x)         ((x)->umode &= ~UMODE_w)
#define ClearFake(x)    	((x)->umode &= ~UMODE_x)
#define ClearZombie(x)  	((x)->umode &= ~UMODE_z)
#define ClearAnon(x)    	((x)->umode &= ~UMODE_q)

#define IsPerson(x)             ((x)->user && IsClient(x))
#define IsPrivileged(x)         (IsAnOper(x) || IsServer(x) || IsService(x) || IsULine(x))

/* #define SetUnixSock(x)  	((x)->flags |= FLAGS_UNIX) */
#define	SetDNS(x)		((x)->flags |= FLAGS_DOINGDNS)
#define	SetAccess(x)		((x)->flags |= FLAGS_CHKACCESS)
#define SetNegoServer(x)        ((x)->flags |= FLAGS_SERV_NEGO)
#define SetRC4OUT(x)            ((x)->flags |= FLAGS_RC4OUT)
#define SetRC4IN(x)             ((x)->flags |= FLAGS_RC4IN)
#define SetZipIn(x)             ((x)->flags |= FLAGS_ZIPPED_IN)
#define SetZipOut(x)            ((x)->flags |= FLAGS_ZIPPED_OUT)
#define SetRestricted(x)        ((x)->flags |= FLAGS_RESTRICTED)

#define IsRestricted(x)         ((x)->flags & FLAGS_RESTRICTED)
#define IsListening(x)          ((x)->flags & FLAGS_LISTEN)
#define DoAccess(x)             ((x)->flags & FLAGS_CHKACCESS)
#define IsLocal(x)              ((x)->flags & FLAGS_LOCAL)
#define IsDead(x)               ((x)->flags & FLAGS_DEADSOCKET)
#define	DoingAuth(x)		((x)->flags & FLAGS_AUTH)
#define	NoNewLine(x)		((x)->flags & FLAGS_NONL)
#define DoingDNS(x)             ((x)->flags & FLAGS_DOINGDNS)
#define IsNegoServer(x)		((x)->flags & FLAGS_SERV_NEGO)
#define IsRC4OUT(x)             ((x)->flags & FLAGS_RC4OUT)
#define IsRC4IN(x)              ((x)->flags & FLAGS_RC4IN)
#define ZipIn(x)                ((x)->flags & FLAGS_ZIPPED_IN)
#define ZipOut(x)               ((x)->flags & FLAGS_ZIPPED_OUT)

#define ClearDNS(x)             ((x)->flags &= ~FLAGS_DOINGDNS)
#define ClearAuth(x)            ((x)->flags &= ~FLAGS_AUTH)
#define ClearAccess(x)          ((x)->flags &= ~FLAGS_CHKACCESS)
#define ClearNegoServer(x)      ((x)->flags &= ~FLAGS_SERV_NEGO)

#define RC4EncLink(x)		(((x)->flags & (FLAGS_RC4IN|FLAGS_RC4OUT)) == (FLAGS_RC4IN|FLAGS_RC4OUT))

/* Oper flags -TimeMr14C: Some are STILL left, in order to see what had existed */

#define OFLAG_REHASH	0x00000001	/* Oper can /rehash server */
#define OFLAG_DIE	0x00000002	/* Oper can /die the server */
#define OFLAG_RESTART	0x00000004	/* Oper can /restart the server 
					 */
#define OFLAG_HELPOP	0x00000010	/* Oper can send /HelpOps */
#define OFLAG_GLOBOP	0x00000020	/* Oper can send /GlobOps */
#define OFLAG_WALLOP	0x00000040	/* Oper can send /WallOps */
#define OFLAG_LOCOP	0x00000080	/* Oper can send /LocOps */
#define OFLAG_LROUTE	0x00000100	/* Oper can do local routing */
#define OFLAG_GROUTE	0x00000200	/* Oper can do global routing */
#define OFLAG_LKILL	0x00000400	/* Oper can do local kills */
#define OFLAG_GKILL	0x00000800	/* Oper can do global kills */
#define OFLAG_KLINE	0x00001000	/* Oper can /kline users */
#define OFLAG_UNKLINE	0x00002000	/* Oper can /unkline users */
#define OFLAG_LNOTICE	0x00004000	/* Oper can send local serv
					 * notices */
#define OFLAG_GNOTICE	0x00008000	/* Oper can send global notices 
					 */
#define OFLAG_ADMIN	0x00010000	/* Admin */
#define OFLAG_UMODEC	0x00020000	/* Oper can set umode +c */
#define OFLAG_UMODEF	0x00040000	/* Oper can set umode +f */
#define OFLAG_SADMIN    0x00080000	/* Oper can be a services admin 
					 */
#define OFLAG_ZLINE	0x00100000	/* Oper can use /zline and
					 * /unzline */
#define OFLAG_UMODEY    0x00200000	/* Oper can set umode +y */
#define OFLAG_UMODED    0x00400000	/* Oper can set umode +d */
#define OFLAG_UMODEB    0x00800000	/* Oper can set umode +b */
#define OFLAG_RSTAFF    0x01000000	/* Oper can view server IPs
					 * (routing staff) */
#define OFLAG_LOCAL	(OFLAG_REHASH|OFLAG_HELPOP|OFLAG_GLOBOP|OFLAG_WALLOP|OFLAG_LOCOP|OFLAG_LROUTE|OFLAG_LKILL|OFLAG_KLINE|OFLAG_UNKLINE|OFLAG_LNOTICE|OFLAG_UMODEC|OFLAG_UMODEF)
#define OFLAG_GLOBAL	(OFLAG_LOCAL|OFLAG_GROUTE|OFLAG_GKILL|OFLAG_GNOTICE)
#define OFLAG_ISGLOBAL	(OFLAG_GROUTE|OFLAG_GKILL|OFLAG_GNOTICE)

#define OPCanZline(x)		((x)->oflag & OFLAG_SADMIN)
#define OPCanRehash(x)		((x)->oflag & OFLAG_REHASH)
#define OPCanDie(x)		((x)->oflag & OFLAG_DIE)
#define OPCanRestart(x)		((x)->oflag & OFLAG_RESTART)
#define OPCanHelpOp(x)		((x)->oflag & OFLAG_HELPOP)
#define OPCanGlobOps(x)		((x)->oflag & OFLAG_GLOBOP)
#define OPCanWallOps(x)		((x)->oflag & OFLAG_WALLOP)
#define OPCanLocOps(x)		((x)->oflag & OFLAG_LOCOP)
#define OPCanLRoute(x)		((x)->oflag & OFLAG_LROUTE)
#define OPCanGRoute(x)		((x)->oflag & OFLAG_GROUTE)
#define OPCanLKill(x)		((x)->oflag & OFLAG_LKILL)
#define OPCanGKill(x)		((x)->oflag & OFLAG_GKILL)
#define OPCanKline(x)		((x)->oflag & OFLAG_KLINE)
#define OPCanUnKline(x)		((x)->oflag & OFLAG_UNKLINE)
#define OPCanLNotice(x)		((x)->oflag & OFLAG_LNOTICE)
#define OPCanGNotice(x)		((x)->oflag & OFLAG_GNOTICE)
#define OPIsAdmin(x)		((x)->oflag & OFLAG_ADMIN)
#define OPIsSAdmin(x)		((x)->oflag & OFLAG_SADMIN)
#define OPIsRStaff(x)   	((x)->oflag & OFLAG_RSTAFF)
#define OPCanUModeC(x)		((x)->oflag & OFLAG_UMODEC)
#define OPCanUModeF(x)		((x)->oflag & OFLAG_UMODEF)
#define OPCanUModeY(x)  	((x)->oflag & OFLAG_UMODEY)
#define OPCanUModeD(x)  	((x)->oflag & OFLAG_UMODED)
#define OPCanUModeB(x)  	((x)->oflag & OFLAG_UMODEB)

#define OPSetRehash(x)		((x)->oflag |= OFLAG_REHASH)
#define OPSetDie(x)		((x)->oflag |= OFLAG_DIE)
#define OPSetRestart(x)		((x)->oflag |= OFLAG_RESTART)
#define OPSetHelpOp(x)		((x)->oflag |= OFLAG_HELPOP)
#define OPSetGlobOps(x)		((x)->oflag |= OFLAG_GLOBOP)
#define OPSetWallOps(x)		((x)->oflag |= OFLAG_WALLOP)
#define OPSetLocOps(x)		((x)->oflag |= OFLAG_LOCOP)
#define OPSetLRoute(x)		((x)->oflag |= OFLAG_LROUTE)
#define OPSetGRoute(x)		((x)->oflag |= OFLAG_GROUTE)
#define OPSetLKill(x)		((x)->oflag |= OFLAG_LKILL)
#define OPSetGKill(x)		((x)->oflag |= OFLAG_GKILL)
#define OPSetKline(x)		((x)->oflag |= OFLAG_KLINE)
#define OPSetUnKline(x)		((x)->oflag |= OFLAG_UNKLINE)
#define OPSetLNotice(x)		((x)->oflag |= OFLAG_LNOTICE)
#define OPSetGNotice(x)		((x)->oflag |= OFLAG_GNOTICE)
#define OPSSetAdmin(x)		((x)->oflag |= OFLAG_ADMIN)
#define OPSSetSAdmin(x) 	((x)->oflag |= OFLAG_SADMIN)
#define OPSetRStaff(x)  	((x)->oflag |= OFLAG_RSTAFF)
#define OPSetUModeC(x)		((x)->oflag |= OFLAG_UMODEC)
#define OPSetUModeF(x)		((x)->oflag |= OFLAG_UMODEF)
#define OPSetUModeY(x)  	((x)->oflag |= OFLAG_UMODEY)
#define OPSetUModeD(x)  	((x)->oflag |= OFLAG_UMODED)
#define OPSetUModeB(x)  	((x)->oflag |= OFLAG_UMODEB)
#define OPSetZLine(x)		((x)->oflag |= OFLAG_ZLINE)

#define OPClearRehash(x)	((x)->oflag &= ~OFLAG_REHASH)
#define OPClearDie(x)		((x)->oflag &= ~OFLAG_DIE)
#define OPClearRestart(x)	((x)->oflag &= ~OFLAG_RESTART)
#define OPClearHelpOp(x)	((x)->oflag &= ~OFLAG_HELPOP)
#define OPClearGlobOps(x)	((x)->oflag &= ~OFLAG_GLOBOP)
#define OPClearWallOps(x)	((x)->oflag &= ~OFLAG_WALLOP)
#define OPClearLocOps(x)	((x)->oflag &= ~OFLAG_LOCOP)
#define OPClearLRoute(x)	((x)->oflag &= ~OFLAG_LROUTE)
#define OPClearGRoute(x)	((x)->oflag &= ~OFLAG_GROUTE)
#define OPClearLKill(x)		((x)->oflag &= ~OFLAG_LKILL)
#define OPClearGKill(x)		((x)->oflag &= ~OFLAG_GKILL)
#define OPClearKline(x)		((x)->oflag &= ~OFLAG_KLINE)
#define OPClearUnKline(x)	((x)->oflag &= ~OFLAG_UNKLINE)
#define OPClearLNotice(x)	((x)->oflag &= ~OFLAG_LNOTICE)
#define OPClearGNotice(x)	((x)->oflag &= ~OFLAG_GNOTICE)
#define OPClearAdmin(x)		((x)->oflag &= ~OFLAG_ADMIN)
#define OPClearSAdmin(x)	((x)->oflag &= ~OFLAG_SADMIN)
#define OPClearRStaff(x) 	((x)->oflag &= ~OFLAG_RSTAFF)
#define OPClearUModeC(x)	((x)->oflag &= ~OFLAG_UMODEC)
#define OPClearUModeF(x)	((x)->oflag &= ~OFLAG_UMODEF)
#define OPClearUModeY(x)        ((x)->oflag &= ~OFLAG_UMODEY)
#define OPClearUModeD(x)        ((x)->oflag &= ~OFLAG_UMODED)
#define OPClearUModeB(x)        ((x)->oflag &= ~OFLAG_UMODEB)
#define OPClearZLine(x)		((x)->oflag &= ~OFLAG_ZLINE)

/* defined debugging levels */
#define	DEBUG_FATAL  0
#define	DEBUG_ERROR  1		/* report_error() and other errors that 
				 * are found */
#define	DEBUG_NOTICE 3
#define	DEBUG_DNS    4		/* used by all DNS related routines - a 
				 * *lot* */
#define	DEBUG_INFO   5		/* general usful info */
#define	DEBUG_NUM    6		/* numerics */
#define	DEBUG_SEND   7		/* everything that is sent out */
#define	DEBUG_DEBUG  8		/* anything to do with debugging, ie
				 * unimportant :) */
#define	DEBUG_MALLOC 9		/* malloc/free calls */
#define	DEBUG_LIST  10		/* debug list use */

struct Counter {
   int server;			/* servers */
   int service;			/* services */
   int myserver;		/* my servers */
   int myulined;		/* my ulined servers */
   int myservice;		/* services connected to me */
   int oper;			/* Opers */
   int chan;			/* Channels */
   int local;			/* Local Clients */
   int total;			/* total clients */
   int invisi;			/* invisible clients */
   int unknown;			/* unknown connections */
   int max_loc;			/* MAX local clients */
   int max_tot;			/* MAX global clients */
   ts_val start;		/* when we started collecting info */
   u_long today;		/* Client Connections today */
   ts_val day;			/* when today started */
   u_long weekly;		/* connections this week */
   ts_val week;			/* when this week started */
   u_long monthly;		/* connections this month */
   ts_val month;		/* when this month started */
   u_long yearly;		/* this is gonna be big */
   ts_val year;			/* when this year started (HEH!) */
};

struct MotdItem {
   char line[MOTDLINELEN];
   struct MotdItem *next;
};

/*
 * lets speed this up... also removed away information. *tough* Dianora
 */
typedef struct Whowas {
   int hashv;
   char name[NICKLEN + 1];
   char username[USERLEN + 1];
   char hostname[HOSTLEN + 1];	/* we only show fakehost here */
   char morehost[HOSTLEN + 1];	/* shown to ircops, the realhost */
   char *servername;
   char realname[REALLEN + 1];
   time_t logoff;
   struct Client *online;	/* Pointer to new nickname for chasing
				 * or NULL */

   struct Whowas *next;		/* for hash table... */

   struct Whowas *prev;		/* for hash table... */
   struct Whowas *cnext;	/* for client struct linked list */
   struct Whowas *cprev;	/* for client struct linked list */
} aWhowas;

struct ConfItem {
   unsigned int status;		/* If CONF_ILLEGAL, delete when no
				 * clients */
   unsigned int flags;		/* i-lines and akills use this */
   int clients;			/* Number of *LOCAL* clients using this 
				 */
   struct in_addr ipnum;	/* ip number of host field */
   char *host;
   char *localhost;
   char *passwd;
   char *name;
   int port;
   time_t hold;			/* Hold action until this time
				 * (calendar time) */
   aClass *class;		/* Class of connection */
   struct ConfItem *next;
};

#define	CONF_CLIENT		0x0002
#define	CONF_CONNECT_SERVER	0x0004
#define	CONF_NOCONNECT_SERVER	0x0008
#define	CONF_LOCOP		0x0010
#define	CONF_OPERATOR		0x0020
#define	CONF_ME			0x0040
#define	CONF_KILL		0x0080
#define	CONF_ADMIN		0x0100
#define	CONF_CLASS		0x0200
#define	CONF_SERVICE		0x0400
/* #define	CONF_LEAF		0x0800 */
#define	CONF_LISTEN_PORT	0x1000
#define	CONF_HUB		0x2000
#define CONF_ELINE		0x4000
#define CONF_FLINE		0x8000
#define	CONF_ZLINE		0x20000
#define CONF_QUARANTINED_NICK 	0x40000
#define CONF_ULINE 		0x80000
#define CONF_DRPASS		0x100000	/* die/restart pass,
						 * from df465 */
#define CONF_AKILL		0x200000
#define CONF_SQLINE     	0x400000
#define CONF_MONINFO		0x800000	/* proxy monitor info */
#define CONF_GCOS               0x1000000
#define CONF_SGLINE             0x2000000
#define CONF_SZLINE             0x4000000	/* Services placed
						 * zlines */
#define CONF_QUARANTINED_CHAN   0x8000000
#define CONF_MATCH              0x40000000
#define CONF_ILLEGAL            0x80000000

#define CONF_JUPECHAN           0x0001	/* Wasn't used :P -TimeMr14C */
#define CONF_SJUPECHAN          0x10000	/* Wasn't used :P -TimeMr14C */
#define CONF_EXCLUDE            0x20000000	/* Wasn't used :P
						 * -TimeMr14C */

#define CONF_QUARANTINE         (CONF_QUARANTINED_NICK | CONF_QUARANTINED_CHAN)
#define	CONF_OPS		(CONF_OPERATOR | CONF_LOCOP)
#define	CONF_SERVER_MASK	(CONF_CONNECT_SERVER | CONF_NOCONNECT_SERVER)
#define	CONF_CLIENT_MASK	(CONF_CLIENT | CONF_SERVICE | CONF_OPS | CONF_SERVER_MASK)

#define	IsIllegal(x)	((x)->status & CONF_ILLEGAL)

#define CONF_FLAGS_I_OPERPORT      0x0002	/* did the password
						 * field specify OPER? */
#define CONF_FLAGS_I_NAME_HAS_AT   0x0004	/* does NAME in
						 * I:HOST::NAME have an 
						 * @? */
#define CONF_FLAGS_I_HOST_HAS_AT   0x0008	/* does HOST in
						 * I:HOST::NAME have an 
						 * @? */

/* Client structures */

struct User {
   Link *channel;		/* chain of channel pointer blocks */
   Link *invited;		/* chain of invite pointer blocks */
   char *away;			/* pointer to away message */
   time_t last;
   int joined;			/* number of channels joined */
   char username[USERLEN + 1];
   char host[HOSTLEN + 1];
#ifdef USE_FAKEHOST
   char fakehost[HOSTLEN + 1];
#endif
   char *server;		/* pointer to scached server name */
#ifdef OSF
   uint_t servicestamp;		/* solaris is gay -epi */
#else
   u_int32_t servicestamp;	/* Services id - Raistlin */
#endif
   /*
    * In a perfect world the 'server' name should not be needed, a
    * pointer to the client describing the server is enough. 
    * Unfortunately, in reality, server may not yet be in links while
    * USER is introduced... --msa
    */
   Link *silence;		/* chain of silenced users */
   LOpts *lopt;			/* Saved /list options */
   Link *dccallow;		/* chain of dcc send allowed users */
   char *real_oper_host;
   char *real_oper_username;
   char *real_oper_ip;
};

struct Server {
   char *up;			/* Pointer to scache name */
   char bynick[NICKLEN + 1];
   char byuser[USERLEN + 1];
   char byhost[HOSTLEN + 1];
   aConfItem *nline;		/* N-line pointer for this server */
   int dkey_flags;		/* dkey flags */
#ifdef HAVE_ENCRYPTION_ON
   void *sessioninfo_in;	/* pointer to opaque sessioninfo
				 * structure */
   void *sessioninfo_out;	/* pointer to opaque sessioninfo
				 * structure */
   void *rc4_in;		/* etc */
   void *rc4_out;		/* etc */
#endif
   void *zip_out;
   void *zip_in;
};

struct Service {
   int sflags;			/* Flags of this service */
   int enable;			/* Services is filled with information
				 * of ... */
   char *server;		/* uplink server */
   Link *channel;		/* channels monitored by this service */
   aClient *scptr;		/* Client Structure the service belongs 
				 * to */
   struct Service *next, *prev;
};

#define SMODE_U         0x0001
#define SMODE_H         0x0002
#define SMODE_A         0x0004

#define IsServiceModeH(x)       (((x)->status & STAT_SERVICE) && ((x)->service->sflags & SMODE_H))
#define IsServiceModeA(x)       (((x)->status & STAT_SERVICE) && ((x)->service->sflags & SMODE_A))

#define SERVICE_MODE_NONE       0x000000
#define SERVICE_SEE_PREFIX      0x000001
#define SERVICE_SEE_OPERS       0x000002
#define SERVICE_SEE_NICKS       0x000004
#define SERVICE_SEE_QUITS       0x000008
#define SERVICE_SEE_KILLS       0x000010
#define SERVICE_SEE_SERVERS     0x000020
#define SERVICE_SEE_SERVICES    0x000040
#define SERVICE_SEE_SQUITS      0x000080
#define SERVICE_SEE_RNICKS      0x000100
#define SERVICE_SEE_UMODES      0x000200
#define SERVICE_SEE_JOINS       0x000400
#define SERVICE_SEE_KICKS       0x000800
#define SERVICE_SEE_PARTS       0x001000
#define SERVICE_SEE_MODES       0x002000
#define SERVICE_SEE_TOPIC       0x004000

struct Client {
   struct Client *next, *prev, *hnext;
   anUser *user;		/* ...defined, if this is a User */
   aServer *serv;		/* ...defined, if this is a server */
   aService *service;		/* Guess what this is -TimeMr14C */
   aWhowas *whowas;		/* Pointers to whowas structs */
   time_t lasttime;		/* ...should be only LOCAL clients?
				 * --msa */
   time_t firsttime;		/* time client was created */
   time_t since;		/* last time we parsed something */
   ts_val tsinfo;		/* TS on the nick, SVINFO on servers */
   long flags;			/* client flags */
   long umode;			/* We can illeviate overflow this way */
   aClient *from;		/* == self, if Local Client, *NEVER*
				 * NULL! */
   aClient *uplink;		/* this client's uplink to the network */
   int fd;			/* >= 0, for local clients */
   int hopcount;		/* number of servers to this 0 = local */
   short status;		/* Client type */
   char nicksent;
   char name[HOSTLEN + 1];	/* Unique name of the client, nick or
				 * host */
   char info[REALLEN + 1];	/* Free form additional client
				 * information */
   char username[USERLEN + 1];  /* username here now for auth stuff */
   struct in_addr ip;           /* keep real ip# too */
   char hostip[HOSTIPLEN + 1];  /* Keep real ip as string too - Dianora
                                 */
   char sockhost[HOSTLEN + 1];

#ifdef FLUD
   Link *fludees;
#endif


/***********************************************************************
 ** FIELDS BELOW ARE LOCAL ONLY ********* FIELDS BELOW ARE LOCAL ONLY **
 ***********************************************************************/

   /*
    * The following fields are allocated only for local clients 
    * (directly connected to this server with a socket.  The first
    * of them MUST be the "count"--it is the field to which the
    * allocation is tied to! Never refer to  these fields, if (from != self).
    */
   int count;			/* Amount of data in buffer */
#ifdef FLUD
   time_t fludblock;
   struct fludbot *fluders;
#endif
#ifdef ANTI_SPAMBOT
   time_t last_join_time;	/* when this client last joined a
				 * channel */
   time_t last_leave_time;	/* when this client last left a channel 
				 */
   int join_leave_count;	/* count of JOIN/LEAVE in less than
				 * MIN_JOIN_LEAVE_TIME seconds */
   int oper_warn_count_down;
#endif
   char buffer[BUFSIZE];	/* Incoming message buffer */
   short lastsq;		/* # of 2k blocks when sendqueued
				 * called last */
   struct DBuf sendQ;		/* Outgoing message queue--if socket
				 * full */
   struct DBuf recvQ;		/* Hold for data incoming yet to be
				 * parsed */
   long sendM;			/* Statistics: protocol messages send */
   long sendK;			/* Statistics: total k-bytes send */
   long receiveM;		/* Statistics: protocol messages
				 * received */
   long receiveK;		/* Statistics: total k-bytes received */
   u_short sendB;		/* counters to count upto 1-k lots of
				 * bytes */
   u_short receiveB;		/* sent and received. */
   long lastrecvM;		/* to check for activity --Mika */
   int priority;
   aClient *acpt;		/* listening client which we accepted
				 * from */
   Link *confs;			/* Configuration record associated */
   int authfd;			/* fd for rfc931 authentication */
   unsigned short port;		/* and the remote port# too :-) */
   struct hostent *hostp;
#ifdef ANTI_NICK_FLOOD
   time_t last_nick_change;
   int number_of_nick_changes;
#endif
#ifdef NO_AWAY_FLUD
   time_t alas;			/* last time of away set */
   int acount;			/* count of away settings */
#endif

   char passwd[PASSWDLEN + 1];
   int oflag;			/* Operator Flags */
   Link *watch;			/* user's watch list */
   int sockerr;			/* what was the last error returned for 
				 * this socket? */
   int watches;			/* how many watches this user has set */
   int capabilities;		/* what this server/client supports */
   int pingval;			/* cache client class ping value here */
   int sendqlen;		/* cache client max sendq here */
};

#define	CLIENT_LOCAL_SIZE sizeof(aClient)
#define	CLIENT_REMOTE_SIZE offsetof(aClient,count)

/* statistics structures */
struct stats {
   unsigned int is_cl;		/* number of client connections */
   unsigned int is_sv;		/* number of server connections */
   unsigned int is_ni;		/* connection but no idea who it was */
   unsigned short is_cbs;	/* bytes sent to clients */
   unsigned short is_cbr;	/* bytes received to clients */
   unsigned short is_sbs;	/* bytes sent to servers */
   unsigned short is_sbr;	/* bytes received to servers */
   unsigned long is_cks;	/* k-bytes sent to clients */
   unsigned long is_ckr;	/* k-bytes received to clients */
   unsigned long is_sks;	/* k-bytes sent to servers */
   unsigned long is_skr;	/* k-bytes received to servers */
   time_t is_cti;		/* time spent connected by clients */
   time_t is_sti;		/* time spent connected by servers */
   unsigned int is_ac;		/* connections accepted */
   unsigned int is_ref;		/* accepts refused */
   unsigned int is_unco;	/* unknown commands */
   unsigned int is_wrdi;	/* command going in wrong direction */
   unsigned int is_unpf;	/* unknown prefix */
   unsigned int is_empt;	/* empty message */
   unsigned int is_num;		/* numeric message */
   unsigned int is_kill;	/* number of kills generated on
				 * collisions */
   unsigned int is_fake;	/* MODE 'fakes' */
   unsigned int is_asuc;	/* successful auth requests */
   unsigned int is_abad;	/* bad auth requests */
   unsigned int is_udp;		/* packets recv'd on udp port */
   unsigned int is_loc;		/* local connections made */
#ifdef FLUD
   unsigned int is_flud;	/* users/channels flood protected */
#endif
};

/* Message table structure */

struct Message {
   char *cmd;
   int (*func) ();
   unsigned int count;		/* number of times command used */
   int parameters;
   char flags;

   /*
    * bit 0 set means that this command is allowed to be used only on
    * the average of once per 2 seconds -SRB I could have defined
    * other bit maps to above instead of the next two flags that I
    * added. so sue me. -Dianora 
    */

   char allow_unregistered_use;	/* flag if this command can be used if
				 * unregistered */
   char reset_idle;		/* flag if this command causes idle
				 * time to be reset */
   char allow_call_for_service;	/* flag if this command from a service
				 * can be called */
   unsigned long bytes;
};

typedef struct msg_tree {
   char *final;
   struct Message *msg;
   struct msg_tree *pointers[26];
} MESSAGE_TREE;

/*
 * Move BAN_INFO information out of the SLink struct its _only_ used
 * for bans, no use wasting the memory for it in any other type of
 * link. Keep in mind, doing this that it makes it slower as more
 * Malloc's/Free's have to be done, on the plus side bans are a smaller
 * percentage of SLink usage. Over all, the th+hybrid coding team came
 * to the conclusion it was worth the effort.
 * 
 * - Dianora
 */

struct Ban {
   char *banstr;
   char *who;
   time_t when;
   u_char type;
   aBan *next;
};

struct Invite {
   char *invitestr;
   char *who;
   time_t when;
   u_char type;
   anInvite *next;
};

struct BanEx {
   char *banexstr;
   char *who;
   time_t when;
   u_char type;
   aBanEx *next;
};

/* channel member link structure, used for chanmember chains */
struct ChanLink {
   struct ChanLink *next;
   aClient *cptr;
   int flags;
   int bans;			/* for bquiet: number of bans against
				 * this user */
   int invites;
   int banexceptions;
};

/* general link structure used for chains */

struct SLink {
   struct SLink *next;
   union {
      aClient *cptr;
      aChannel *chptr;
      aConfItem *aconf;
      aWatch *wptr;
      char *cp;
   } value;
   int flags;
};

/* mode structure for channels */

struct SMode {
   unsigned int mode;
   int limit;
   char key[KEYLEN + 1];
};

/* channel structure */

struct Channel {
   struct Channel *nextch, *prevch, *hnextch;
   int hashv;			/* raw hash value */
   Mode mode;
   char topic[TOPICLEN + 1];
   char topic_nick[NICKLEN + 1];
   time_t topic_time;
   int users;
   chanMember *members;
   Link *invites;
   aBan *banlist;
   anInvite *invitelist;
   aBanEx *banexlist;
   ts_val channelts;
#ifdef FLUD
   time_t fludblock;
   struct fludbot *fluders;
#endif
   char chname[CHANNELLEN + 1];
};

#define	TS_CURRENT	3	/* current TS protocol version */
#define	TS_MIN		1	/* minimum supported TS protocol
				 * version */
#define	TS_DOESTS	0x20000000	/* What the hell is that ?
					 * -TimeMr14C */
#define	DoesTS(x)	((x)->tsinfo == TS_DOESTS)

/* Channel related flags */

#define	CHFL_CHANOP     0x0001	/* Channel operator */
#define	CHFL_VOICE      0x0002	/* the power to speak */
#define	CHFL_DEOPPED 	0x0004	/* deopped by us, modes need to be
				 * bounced */
#define	CHFL_BAN	0x0008	/* ban channel flag */

#define CHFL_INVITE     0x10000	/* invites */
#define CHFL_EXCEPTION  0x20000	/* ban exceptions */
#define CHFL_PROTECT	0x40000	/* kick protected users */
#define CHFL_OWNER	0x80000	/* channel owner (kick + deop
				 * protection) */

/* mask types */

#define MTYP_FULL      0x01	/* mask is nick!user@host */
#define MTYP_USERHOST  0x02	/* mask is user@host */
#define MTYP_HOST      0x04	/* mask is host only */

/* Channel Visibility macros */

#define	MODE_CHANOP	CHFL_CHANOP
#define	MODE_VOICE	CHFL_VOICE
#define	MODE_DEOPPED  	CHFL_DEOPPED
#define	MODE_PRIVATE  	0x0008
#define	MODE_SECRET   	0x0010
#define	MODE_MODERATED  0x0020
#define	MODE_TOPICLIMIT 0x0040
#define	MODE_INVITEONLY 0x0080
#define	MODE_NOPRIVMSGS 0x0100
#define	MODE_KEY	0x0200
#define	MODE_BAN	0x0400
#define	MODE_LIMIT	0x0800
#define MODE_REGISTERED	0x1000
#define MODE_REGONLY	0x2000
#define MODE_NOCOLOR	0x4000
#define MODE_OPERONLY   0x8000

#define MODE_INVITE     CHFL_INVITE
#define MODE_EXCEPTION  CHFL_EXCEPTION
#define MODE_PROTECT	CHFL_PROTECT
#define MODE_OWNER	CHFL_OWNER
#define MODE_ANONYMOUS  0x1000000

/* mode flags which take another parameter (With PARAmeterS) */

#define MODE_SPEAKABLE  (MODE_CHANOP|MODE_VOICE|MODE_OWNER)
#define	MODE_WPARAS	(MODE_SPEAKABLE|MODE_BAN|MODE_KEY|MODE_LIMIT|MODE_EXCEPTION|MODE_INVITES)
#define CHFL_SPEAKABLE  (CHFL_CHANOP|CHFL_VOICE|CHFL_OWNER)

#define	HoldChannel(x)		(!(x))	/* name invisible */
#define	SecretChannel(x)	((x) && ((x)->mode.mode & MODE_SECRET))
#define	HiddenChannel(x)	((x) && ((x)->mode.mode & MODE_PRIVATE))
#define	ShowChannel(v,c)	(PubChannel(c) || IsMember((v),(c)))
#define	PubChannel(x)		((!x) || ((x)->mode.mode & (MODE_PRIVATE | MODE_SECRET)) == 0)

#define IsChanModerated(x)      ((x)->mode.mode & MODE_MODERATED)
#define IsChanOpTopic(x)        ((x)->mode.mode & MODE_TOPICLIMIT)
#define IsChanInviteonly(x)     ((x)->mode.mode & MODE_INVITEONLY)
#define IsChanNoPriv(x)         ((x)->mode.mode & MODE_NOPRIVMSGS)
#define IsChanKey(x)            ((x)->mode.mode & MODE_KEY)
#define IsChanLimited(x)        ((x)->mode.mode & MODE_LIMIT)
#define IsChanRegistered(x)     ((x)->mode.mode & MODE_REGISTERED)
#define IsChanRegOnly(x)        ((x)->mode.mode & MODE_REGONLY)
#define IsChanNoColor(x)        ((x)->mode.mode & MODE_NOCOLOR)
#define IsChanOperOnly(x)       ((x)->mode.mode & MODE_OPERONLY)
#define IsChanAnon(x)           ((x)->mode.mode & MODE_ANONYMOUS)

#define IsMember(b,c) 		((b && b->user && find_channel_link((b->user)->channel, c)) ? 1 : 0)

#define	IsChannelName(name) 	((name) && (*(name) == '#' || *(name) == '&'))
#define IsLocalChan(x)          ((x) && (*(x) == '&'))
#define IsGlobalChan(x)         ((x) && (*(x) == '#'))

/* Misc macros */

#define	BadPtr(x) (!(x) || (*(x) == '\0'))

#define	isvalid(c) (((c) >= 'A' && (c) < '~') || isdigit(c) || (c) == '-')

#define	MyConnect(x)		((x)->fd >= 0)
#define	MyClient(x)		(MyConnect(x) && IsClient(x))
#define	MyOper(x)		(MyConnect(x) && IsOper(x))

/* String manipulation macros */

/* strncopynt --> strncpyzt to avoid confusion, sematics changed N must
 * be now the number of bytes in the array --msa */

#define	strncpyzt(x, y, N) do{(void)strncpy(x,y,N);x[N-1]='\0';}while(0)
#define	StrEq(x,y)	(!strcmp((x),(y)))

/* used in SetMode() in channel.c and m_umode() in s_msg.c */

#define	MODE_NULL      0
#define	MODE_ADD       0x40000000
#define	MODE_DEL       0x20000000

/* return values for hunt_server() */

#define	HUNTED_NOSUCH	(-1)	/* if the hunted server is not found */
#define	HUNTED_ISME	0	/* if this server should execute the
				 * command */
#define	HUNTED_PASS	1	/* if message passed onwards
				 * successfully */

/* used when sending to #mask or $mask */

#define	MATCH_SERVER  1
#define	MATCH_HOST    2

/* used for async dns values */

#define	ASYNC_NONE	(-1)
#define	ASYNC_CLIENT	0
#define	ASYNC_CONNECT	1
#define	ASYNC_CONF	2
#define	ASYNC_SERVER	3

/* misc variable externs */

extern char version[128], *infotext[];
extern char *generation, *creation;

/* misc defines */

#define ZIP_NEXT_BUFFER -4
#define RC4_NEXT_BUFFER -3
#define	FLUSH_BUFFER	-2
#define	UTMP		"/etc/utmp"
#define	COMMA		","

#ifdef ORATIMING

/*
 * Timing stuff (for performance measurements): compile with
 * -DORATIMING and put a TMRESET where you want the counter of time
 * spent set to 0, a TMPRINT where you want the accumulated results,
 * and TMYES/TMNO pairs around the parts you want timed -orabidoo
 */

extern struct timeval tsdnow, tsdthen;
extern unsigned long tsdms;

#define TMRESET tsdms=0;
#define TMYES gettimeofday(&tsdthen, NULL);
#define TMNO gettimeofday(&tsdnow, NULL); if (tsdnow.tv_sec!=tsdthen.tv_sec) tsdms+=1000000*(tsdnow.tv_sec-tsdthen.tv_sec); tsdms+=tsdnow.tv_usec; tsdms-=tsdthen.tv_usec;
#define TMPRINT sendto_ops("Time spent: %ld ms", tsdms);
#else
#define TMRESET
#define TMYES
#define TMNO
#define TMPRINT
#endif

/* allow 5 minutes after server rejoins the network before allowing
 * chanops new channel */

#ifdef NO_CHANOPS_WHEN_SPLIT
#define MAX_SERVER_SPLIT_RECOVERY_TIME 5
#endif

#ifdef FLUD
struct fludbot {
   struct Client *fluder;
   int count;
   time_t first_msg, last_msg;
   struct fludbot *next;
};
#endif

struct Watch {
   aWatch *hnext;
   time_t lasttime;
   Link *watch;
   char nick[1];
};

struct ListOptions {
   LOpts *next;
   Link *yeslist, *nolist;
   int starthash;
   short int showall;
   unsigned short usermin;
   int usermax;
   time_t currenttime;
   time_t chantimemin;
   time_t chantimemax;
   time_t topictimemin;
   time_t topictimemax;
};

typedef struct SearchOptions {
   int umodes;
   char *nick;
   char *user;
   char *host;
   char *gcos;
   char *ip;
   char *fakehost;
   aChannel *channel;
   aClient *server;
   char umode_plus:1;
   char nick_plus:1;
   char user_plus:1;
   char host_plus:1;
   char gcos_plus:1;
   char ip_plus:1;
   char fhost_plus:1;
   char chan_plus:1;
   char serv_plus:1;
   char away_plus:1;
   char check_away:1;
   char check_umode:1;
   char show_chan:1;
   char search_chan:1;
   char spare:4;		/* spare space for more stuff(?) */
} SOpts;

#define IsSendable(x)      	(DBufLength(&x->sendQ) < 16384)
#define DoList(x)    		(((x)->user) && ((x)->user->lopt))

/* internal defines for cptr->sockerr */
#define IRCERR_BUFALLOC         -11
#define IRCERR_ZIP              -12

#endif /* __struct_include__ */
