/*
 *   IRC - Internet Relay Chat, src/cmd_oper.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Co Center
 *   
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *   
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: cmd_oper.c,v 1.13 2001/12/05 18:31:50 tr-ircd Exp $ 
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "nameser.h"
#include "resolv.h"
#include "dh.h"
#include "zlink.h"

#if defined(AIX) || defined(DYNIXPTX) || defined(SVR3)
# include <time.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "h.h"
#if defined( HAVE_STRING_H )
# include <string.h>
#else
  /*
   * older unices don't have strchr/strrchr .. help them out 
   */
# include <strings.h>
# undef strchr
# define strchr index
#endif
#include "dich_conf.h"
#include "fdlist.h"
#include "throttle.h"

extern fdlist serv_fdlist;
extern int lifesux;
static char buf[BUFSIZE], buf2[BUFSIZE];
static char modebuf[REALMODEBUFLEN], parabuf[REALMODEBUFLEN];
extern int rehashed;
extern int zline_in_progress;

#ifdef HTM_LOCK_ON_NETBURST
extern int HTMLOCK;
#endif

#ifdef HIGHEST_CONNECTION
int max_connection_count = 1, max_client_count = 1;
#endif

/*
 * external variables 
 */

/*
 * external functions 
 */

#ifdef MAXBUFFERS
extern void reset_sock_opts();
#endif

extern char *smalldate(time_t);	/* defined in s_misc.c */
extern void outofmemory(void);	/* defined in list.c */
extern void s_die(void);	/* defined in ircd.c as VOIDSIG */
extern int match(char *, char *);	/* defined in match.c */
extern char *find_restartpass();
extern char *find_diepass();

/*
 * Local function prototypes 
 */

static int isnumber(char *);	/* return 0 if not, else return number */
static char *cluster(char *);

int send_motd(aClient *, aClient *, int, char **);
void read_motd(char *);

#ifdef SHORT_MOTD
void read_shortmotd(char *);
#endif

char motd_last_changed_date[MAX_DATE_STRING];	/* enough room for date */

static int flush_write(aClient *, char *, int, char *, int, char *);

#ifdef LOCKFILE
/*
 * Shadowfax's lockfile code 
 */
void do_pending_klines(void);

struct pkl {
    char *comment;		/* Kline Comment */
    char *kline;		/* Actual Kline */
    struct pkl *next;		/* Next Pending Kline */
} *pending_klines = NULL;

time_t pending_kline_time = 0;

#endif /* LOCKFILE */

/*
 * m_samode - Just bout the same as df
 * *  - Raistlin
 * * parv[0] = sender
 * * parv[1] = channel
 * * parv[2] = modes 
 */
int m_samode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int sendts;
    aChannel *chptr;

    if (check_registered(cptr))
	return 0;
    if (!IsPrivileged(cptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (!IsSAdmin(cptr) || parc < 2)
	return 0;
    chptr = find_channel(parv[1], NullChn);
    if (chptr == NullChn)
	return 0;
    if (IsChanAnon(chptr)) {
	sendto_one(sptr,
		   ":%s NOTICE %s :SAMODE Cannot be used on anonymous channels.",
		   me.name, parv[0]);
	sendto_serv_butone(NULL,
			   ":%s GLOBOPS :%s tried to use SAMODE on anonymous channel %s",
			   me.name, sptr->name, chptr->chname);
	return 0;
    }

    if (!check_channelname(sptr, (unsigned char *) parv[1]))
	return 0;
    sendts =
	set_mode(cptr, sptr, chptr, 2, parc - 2, parv + 2, modebuf, parabuf);

    if (strlen(modebuf) > (size_t) 1) {
	sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s",
			       parv[0], chptr->chname, modebuf, parabuf);
	sendto_match_servs(chptr, cptr, ":%s MODE %s %s %s",
			   parv[0], chptr->chname, modebuf, parabuf);
	if (MyClient(sptr)) {
	    sendto_serv_butone(NULL,
			       ":%s GLOBOPS :%s used SAMODE (%s %s%s%s)",
			       me.name, sptr->name, chptr->chname, modebuf,
			       (*parabuf != 0 ? " " : ""), parabuf);
	    send_globops("from %s: %s used SAMODE (%s %s%s%s)", me.name,
			 sptr->name, chptr->chname, modebuf,
			 (*parabuf != 0 ? " " : ""), parabuf);
	}
    }
    return 0;
}

/*
 * * m_wallops (write to *all* opers currently online)
 *      parv[0] = sender prefix
 *      parv[1] = message text
 */
int m_wallops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (BadPtr(message)) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "WALLOPS");
	return 0;
    }

    if (!IsServer(sptr) && MyConnect(sptr) && !OPCanWallOps(sptr)
	&& !IsServiceModeH(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return (0);
    }

    sendto_wallops_butone(IsServer(cptr) ? cptr : NULL, sptr, ":%s WALLOPS :%s",
			  parv[0], message);
    return 0;
}

/*
 * * m_locops (write to *all* local opers currently online)
 *      parv[0] = sender prefix
 *      parv[1] = message text 
 */
int m_locops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (BadPtr(message)) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "LOCOPS");
	return 0;
    }

    if (!IsServer(sptr) && MyConnect(sptr) && !OPCanLocOps(sptr)
	&& !IsServiceModeH(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return (0);
    }
    sendto_locops("from %s: %s", parv[0], message);
    return (0);
}

/*
 * m_goper  (Russell) sort of like wallop, but only to ALL +o clients
 * on every server.
 *      parv[0] = sender prefix
 *      parv[1] = message text 
 * Taken from df465, ported to hybrid. -mjs
 */
int m_goper(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;

    if (BadPtr(message)) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "GOPER");
	return 0;
    }
    if ((!IsServer(sptr) && !IsServiceModeH(sptr)) || !IsULine(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    sendto_serv_butone(IsServer(cptr) ? cptr : NULL, ":%s GOPER :%s", parv[0],
		       message);
    sendto_ops("from %s: %s", parv[0], message);
    return 0;
}

/*
 * m_gnotice  (Russell) sort of like wallop, but only to +g clients on *
 * this server.
 *      parv[0] = sender prefix
 *      parv[1] = message text 
 * ported from df465 to hybrid -mjs
 * 
 * This function itself doesnt need any changes for the move to +n routing
 * notices, to sendto takes care of it.  Now only sends to +n clients -epi
 */
int m_gnotice(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;

    if (BadPtr(message)) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "GNOTICE");
	return 0;
    }
    if (!IsServer(sptr) && MyConnect(sptr) && !IsServiceModeH(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    sendto_serv_butone(IsServer(cptr) ? cptr : NULL, ":%s GNOTICE :%s", parv[0],
		       message);
    sendto_gnotice("from %s: %s", parv[0], message);
    return 0;
}

int m_globops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    /*
     * a few changes, servers weren't able to globop -mjs 
     */

    if (check_registered(sptr))
	return 0;

    if (BadPtr(message)) {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		       "GLOBOPS");
	return 0;
    }

    /*
     * must be a client, must be an oper or a Ulined server -mjs 
     */

    if (MyClient(sptr) && !OPCanGlobOps(sptr) && !IsULine(sptr)
	&& !IsServiceModeH(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (strlen(message) > TOPICLEN)
	message[TOPICLEN] = '\0';
    sendto_serv_butone(IsServer(cptr) ? cptr : NULL, ":%s GLOBOPS :%s", parv[0],
		       message);
    send_globops("from %s: %s", parv[0], message);
    return 0;
}

/*
 * This function is not necessary -TimeMr14C
 * int
 * m_chatops(aClient *cptr, aClient *sptr, int parc, char *parv[])
 * {
 * char       *message = parc > 1 ? parv[1] : NULL;
 * 
 * if (check_registered(sptr))
 * return 0;
 * if (BadPtr(message))
 * {
 * if (MyClient(sptr))
 * sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS),
 * me.name, parv[0], "CHATOPS");
 * return 0;
 * }
 * 
 * if (MyClient(sptr) && (!IsAnOper(sptr) || !SendChatops(sptr)) && !IsULine(sptr))
 * {
 * sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
 * return 0;
 * }
 * 
 * if (strlen(message) > TOPICLEN)
 * message[TOPICLEN] = '\0';   
 * sendto_serv_butone(IsServer(cptr) ? cptr : NULL, ":%s CHATOPS :%s",
 * parv[0], message);
 * send_chatops("from %s: %s", parv[0], message);
 * return 0;
 * }
 */
/*
 * m_htm - high traffic mode info
 */
int m_htm(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
#define LOADCFREQ 5
    char *command;

    extern int lifesux, LRV, LCF, noisy_htm;	/* in ircd.c */
    extern int currlife;

    if (!IsAnOper(sptr) && !IsServiceModeH(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    sendto_one(sptr,
	       ":%s NOTICE %s :HTM is %s(%d), %s. Max rate = %dk/s. Current = %dk/s",
	       me.name, parv[0], lifesux ? "ON" : "OFF", lifesux,
	       noisy_htm ? "NOISY" : "QUIET", LRV, currlife);
    if (parc > 1) {
	command = parv[1];
	if (!strcasecmp(command, "TO")) {
	    if (parc > 2) {
    int new_value = atoi(parv[2]);

		if (new_value < 0)
		    sendto_one(sptr,
			       ":%s NOTICE %s :\002Cannot set LRV < 0!\002",
			       me.name, parv[0]);
		else
		    LRV = new_value;

		sendto_one(sptr,
			   ":%s NOTICE %s :NEW Max rate = %dk/s. Current = %dk/s",
			   me.name, parv[0], LRV, currlife);
		sendto_realops
		    ("%s!%s@%s set new HTM rate to %dk/s (%dk/s current)",
		     parv[0],
		     (sptr->user ? sptr->user->username : sptr->username),
		     sptr->sockhost, LRV, currlife);
	    } else
		sendto_one(sptr,
			   ":%s NOTICE %s :LRV command needs an integer parameter",
			   me.name, parv[0]);
	} else {
	    if (!strcasecmp(command, "ON")) {
		lifesux = 1;
		sendto_one(sptr, ":%s NOTICE %s :HTM is now ON.", me.name,
			   parv[0]);
		sendto_ops("Entering high-traffic mode: Forced by %s!%s@%s",
			   parv[0],
			   (sptr->user ? sptr->user->username : sptr->username),
			   sptr->sockhost);
		LCF = 30;	/* 30s */
	    } else if (!strcasecmp(command, "OFF")) {
#ifdef HTM_LOCK_ON_NETBURST
		if (HTMLOCK == YES) {
		    if (!IsAdmin(sptr)) {
			sendto_one(sptr,
				   ":%s NOTICE %s :Cannot change HTM - Currently LOCKED");
			return 0;
		    } else {
			sendto_one(sptr, ":%s NOTICE %s :Removing HTM LOCK");
			sendto_ops
			    ("Resuming standard operation through HTM LOCK: Forced by %s!%s@%s",
			     parv[0],
			     (sptr->user ? sptr->user->username : sptr->
			      username), sptr->sockhost);
			lifesux = 0;
			LCF = LOADCFREQ;
			return 0;
		    }
		}
#endif

		lifesux = 0;
		LCF = LOADCFREQ;
		sendto_one(sptr, ":%s NOTICE %s :HTM is now OFF.", me.name,
			   parv[0]);
		sendto_ops("Resuming standard operation: Forced by %s!%s@%s",
			   parv[0],
			   (sptr->user ? sptr->user->username : sptr->username),
			   sptr->sockhost);
	    } else if (!strcasecmp(command, "QUIET")) {
		sendto_ops("HTM is now QUIET");
		noisy_htm = NO;
	    } else if (!strcasecmp(command, "NOISY")) {
		sendto_ops("HTM is now NOISY");
		noisy_htm = YES;
	    } else
		sendto_one(sptr,
			   ":%s NOTICE %s :Commands are:HTM [ON] [OFF] [TO int] [QUIET] [NOISY]",
			   me.name, parv[0]);
	}
    }
    return 0;
}

/*
 * cluster() input
 * - pointer to a hostname output
 * pointer to a static of the hostname masked for use in a kline. side
 * effects - NONE
 *
 * reworked a tad -Dianora
 */

static char *cluster(char *hostname)
{
    static char result[HOSTLEN + 1];	/* result to return */
    char temphost[HOSTLEN + 1];	/* workplace */
    char *ipp;			/* used to find if host is ip # only */
    char *host_mask;		/* used to find host mask portion to '*' */
    char *zap_point = (char *) NULL;	/* used to zap last nnn portion of an
					 * ip # */
    char *tld;			/* Top Level Domain */
    int is_ip_number;		/* flag if its an IP # */
    int number_of_dots;		/* count # of dots for ip# and domain klines */

    if (!hostname)
	return (char *) NULL;	/* EEK! */

    /*
     * If a '@' is found in the hostname, this is bogus and must have
     * been introduced by server that doesn't check for bogus domains
     * (dns spoof) very well. *sigh* just return it... I could also
     * legitimately return (char *)NULL as above.
     *
     * -Dianora
     */

    if (strchr(hostname, '@')) {
	strncpyzt(result, hostname, HOSTLEN);
	return (result);
    }

    strncpyzt(temphost, hostname, HOSTLEN);

    is_ip_number = YES;		/* assume its an IP# */
    ipp = temphost;
    number_of_dots = 0;

    while (*ipp) {
	if (*ipp == '.') {
	    number_of_dots++;
	    if (number_of_dots == 3)
		zap_point = ipp;
	    ipp++;
	} else if (!isdigit(*ipp)) {
	    is_ip_number = NO;
	    break;
	}
	ipp++;
    }

    if (is_ip_number && (number_of_dots == 3)) {
	zap_point++;
	*zap_point++ = '*';	/* turn 111.222.333.444 into ... */
	*zap_point = '\0';	/* 111.222.333.* */
	strncpy(result, temphost, HOSTLEN);
	return (result);
    } else {
	tld = strrchr(temphost, '.');
	if (tld) {
	    number_of_dots = 2;
	    if (tld[3])		/* its at least a 3 letter tld */
		number_of_dots = 1;
	    if (tld != temphost)	/* in these days of dns spoofers ... */
		host_mask = tld - 1;	/* Look for host portion to '*' */
	    else
		host_mask = tld;	/* degenerate case hostname is '.com'
					 * ect. */

	    while (host_mask != temphost) {
		if (*host_mask == '.')
		    number_of_dots--;
		if (number_of_dots == 0) {
		    result[0] = '*';
		    strncpy(result + 1, host_mask, HOSTLEN - 1);
		    return (result);
		}
		host_mask--;
	    }
	    result[0] = '*';	/* foo.com => *foo.com */
	    strncpy(result + 1, temphost, HOSTLEN);
	} else {		/* no tld found oops. just return it as is */
	    strncpy(result, temphost, HOSTLEN);
	    return (result);
	}
    }

    return (result);
}

/*
 * isnumber()
 *
 * inputs
 * - pointer to ascii string in output
 * - 0 if not an integer number, else the number side effects
 * - none return -1 if not an integer.
 * (if someone types in maxint, oh well..) - lucas
 */

static int isnumber(char *p)
{
    int result = 0;

    while (*p) {
	if (isdigit(*p)) {
	    result *= 10;
	    result += ((*p) & 0xF);
	    p++;
	} else
	    return (-1);
    }
    /*
     * in the degenerate case where oper does a /quote kline 0 user@host
     * :reason i.e. they specifically use 0, I am going to return 1
     * instead as a return value of non-zero is used to flag it as a
     * temporary kline
     */

    /*
     * er, no. we return 0 because 0 means that it's a permanent kline. -lucas
     * oh, and we only do this if DEFAULT_KLINE_TIME is specified.
     */

#ifndef DEFAULT_KLINE_TIME
    if (result == 0)
	result = 1;
#endif

    return (result);
}

/*
 * flush_write()
 *
 * inputs
 * - pointer to client structure of oper requesting unkline
 * - out is the file descriptor
 * - buf is the buffer to write
 * - ntowrite is the expected number of character to be written
 * - temppath is the temporary file name to be written output
 *   YES for error on write
 *   NO for success side effects
 * - if successful, the buf is written to output file if a write failure
 *   happesn, and the file pointed to by temppath, if its non NULL, is
 *   removed.
 * 
 * The idea here is, to be as robust as possible when writing to the kline
 * file.
 * 
 * Yes, I could have dug the opernick out of sptr. I didn't feel like it.
 * so sue me.
 *
 * -Dianora
 */

static int
flush_write(aClient *sptr, char *opernick, int out, char *buf, int ntowrite,
	    char *temppath)
{
    int nwritten;
    int error_on_write = NO;

    nwritten = write(out, buf, ntowrite);
    if (nwritten != ntowrite) {
	sendto_one(sptr, ":%s NOTICE %s :Unable to write to %s", me.name,
		   opernick, temppath);
	error_on_write = YES;
	(void) close(out);
	if (temppath != (char *) NULL)
	    (void) unlink(temppath);
    }
    return (error_on_write);
}

/*
 * m_kline()
 *
 * re-worked a tad ... -Dianora
 *
 * Added comstuds SEPARATE_QUOTE_KLINES_BY_DATE code -Dianora
 */

int m_kline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

#if defined (LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
    struct pkl *k;
#else
    int out;
#endif

    char buffer[1024];

#ifdef SEPARATE_QUOTE_KLINES_BY_DATE
    char timebuffer[MAX_DATE_STRING];
    char filenamebuf[1024];
    struct tm *tmptr;
#endif
    char *filename;		/* filename to use for kline */
    char *user, *host;
    char *reason;
    char *current_date;
    aClient *acptr;
    char tempuser[USERLEN + 2];
    char temphost[HOSTLEN + 1];
    aConfItem *aconf;
    int temporary_kline_time = 0;	/* -Dianora */
    time_t temporary_kline_time_seconds = 0;
    int time_specified = 0;
    char *argv;

    if (!IsAnOper(sptr) && !IsServiceModeA(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "KLINE");
	return 0;
    }

    argv = parv[1];

    if ((temporary_kline_time = isnumber(argv)) >= 0) {
	if (parc < 3) {
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		       "KLINE");
	    return 0;
	}
	if (temporary_kline_time > (24 * 60))
	    temporary_kline_time = (24 * 60);	/* Max it at 24 hours */

    temporary_kline_time_seconds = (time_t) temporary_kline_time *(time_t) 60;

	/*
	 * turn it into minutes 
	 */
	argv = parv[2];
	parc--;
	time_specified = 1;
    } else {
	temporary_kline_time = 0;	/* -1 minute klines are bad... :) -
					 * lucas */
    }

    if (strchr(argv, ' ')) {
	sendto_one(sptr,
		   ":%s NOTICE %s :Poorly formatted hostname (contains spaces). Be sure you are using the form: "
		   "/quote KLINE [time] <user@host/nick> :<reason>", me.name,
		   parv[0]);
	return 0;
    }

    if ((host = strchr(argv, '@')) || *argv == '*') {
	/*
	 * Explicit user@host mask given 
	 */

	if (host) {		/* Found user@host */
	    user = argv;	/* here is user part */
	    *(host++) = '\0';	/* and now here is host */
	} else {
	    user = "*";		/* no @ found, assume its *@somehost */
	    host = argv;
	}

	if (!*host)		/* duh. no host found, assume its '*' * * * * * 
				 * * * * * * * * * * host */
	    host = "*";
	strncpyzt(tempuser, user, USERLEN + 2);	/* allow for '*' in front */
	strncpyzt(temphost, host, HOSTLEN);
	user = tempuser;
	host = temphost;
    } else {
	/*
	 * Try to find user@host mask from nick 
	 */

	if (!(acptr = find_chasing(sptr, argv, NULL)))
	    return 0;

	if (!acptr->user)
	    return 0;

	if (IsServer(acptr)) {
	    sendto_one(sptr,
		       ":%s NOTICE %s :Can't KLINE a server, use @'s where appropriate",
		       me.name, parv[0]);
	    return 0;
	}
	/*
	 * turn the "user" bit into "*user", blow away '~' if found in
	 * original user name (non-idented)
	 */

	tempuser[0] = '*';
	if (*acptr->user->username == '~')
	    strcpy(tempuser + 1, (char *) acptr->user->username + 1);
	else
	    strcpy(tempuser + 1, acptr->user->username);
	user = tempuser;
	host = cluster(acptr->user->host);
    }

    if (time_specified)
	argv = parv[3];
    else
	argv = parv[2];

#ifdef DEFAULT_KLINE_TIME
    if (time_specified == 0) {
	temporary_kline_time = DEFAULT_KLINE_TIME;
    temporary_kline_time_seconds = (time_t) temporary_kline_time *(time_t) 60;
    }
#endif

    if (parc > 2) {
	if (strchr(argv, ':')) {
	    sendto_one(sptr, ":%s NOTICE %s :Invalid character ':' in comment",
		       me.name, parv[0]);
	    return 0;
	}

	if (strchr(argv, '#')) {
	    sendto_one(sptr, ":%s NOTICE %s :Invalid character '#' in comment",
		       me.name, parv[0]);
	    return 0;
	}
	if (*argv)
	    reason = argv;
	else
	    reason = "No reason";
    } else
	reason = "No reason";

    if (!match(user, "akjhfkahfasfjd") &&
	!match(host, "ldksjfl.kss...kdjfd.jfklsjf")) {
	sendto_one(sptr, ":%s NOTICE %s :Can't K-Line *@*", me.name, parv[0]);
	return 0;
    }
#ifdef NON_REDUNDANT_KLINES
    if ((aconf = find_is_klined(host, user, (aClient *) NULL))) {
    char *reason;

	reason = aconf->passwd ? aconf->passwd : "<No Reason>";

	sendto_one(sptr,
		   ":%s NOTICE %s :[%s@%s] already K-lined by [%s@%s] - %s",
		   me.name, parv[0], user, host, aconf->name, aconf->host,
		   reason);
	return 0;

    }
#endif

    current_date = smalldate((time_t) 0);

    aconf = make_conf();
    aconf->status = CONF_KILL;
    DupString(aconf->host, host);

    (void) ircsprintf(buffer, "%s (%s)", reason, current_date);

    DupString(aconf->passwd, buffer);
    DupString(aconf->name, user);
    aconf->port = 0;

    if (temporary_kline_time) {
	aconf->hold = timeofday + temporary_kline_time_seconds;
	add_temp_kline(aconf);
	rehashed = YES;
	zline_in_progress = NO;
	sendto_realops
	    ("%s added temporary %d min. K-Line for [%s@%s] [%s]", parv[0],
	     temporary_kline_time, user, host, reason);
	return 0;
    }

    Class (aconf) = find_class(0);

    /*
     * when can aconf->host ever be a NULL pointer though? or for that
     * matter, when is aconf->status ever not going to be CONF_KILL at
     * this point?  -Dianora
     */

    switch (sortable(host)) {
	case 0:
	    l_addto_conf_list(&KList3, aconf, host_field);
	    break;
	case 1:
	    addto_conf_list(&KList1, aconf, host_field);
	    break;
	case -1:
	    addto_conf_list(&KList2, aconf, rev_host_field);
	    break;
    }
    /*
     * comstud's SEPARATE_QUOTE_KLINES_BY_DATE code 
     */
    /*
     * Note, that if SEPARATE_QUOTE_KLINES_BY_DATE is defined, it
     * doesn't make sense to have LOCKFILE on the kline file
     */

#ifdef SEPARATE_QUOTE_KLINES_BY_DATE
    tmptr = localtime(&NOW);
    strftime(timebuffer, MAX_DATE_STRING, "%y%m%d", tmptr);
    (void) sprintf(filenamebuf, "%s.%s", klinefile, timebuffer);
    filename = filenamebuf;
    sendto_one(sptr, ":%s NOTICE %s :Added K-Line [%s@%s] to %s",
	       me.name, parv[0], user, host, filenamebuf);
#else
    filename = klinefile;

# ifdef KPATH
    sendto_one(sptr,
	       ":%s NOTICE %s :Added K-Line [%s@%s] to server klinefile",
	       me.name, parv[0], user, host);
# else
    sendto_one(sptr,
	       ":%s NOTICE %s :Added K-Line [%s@%s] to server configfile",
	       me.name, parv[0], user, host);
# endif	/* KPATH */
#endif /* SEPARATE_QUOTE_KLINES_BY_DATE */

    rehashed = YES;
    zline_in_progress = NO;
    sendto_realops("%s added K-Line for [%s@%s] [%s]", parv[0], user, host,
		   reason);
#if defined(LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
    /*
     * MDP - Careful, don't cut & paste that ^O 
     */

    if ((k = (struct pkl *) malloc(sizeof(struct pkl))) == NULL) {
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name,
		   parv[0]);
	return (0);
    }

    (void) ircsprintf(buffer, "#%s!%s@%s K'd: %s@%s:%s\n",
		      sptr->name, sptr->user->username, sptr->user->host, user,
		      host, reason);

    if ((k->comment = strdup(buffer)) == NULL) {
	free(k);
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name,
		   parv[0]);
	return (0);
    }

    (void) ircsprintf(buffer, "K:%s:%s (%s):%s\n", host, reason, current_date,
		      user);

    if ((k->kline = strdup(buffer)) == NULL) {
	free(k->comment);
	free(k);
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name,
		   parv[0]);
	return (0);
    }
    k->next = pending_klines;
    pending_klines = k;

    do_pending_klines();
    return (0);

#else /* LOCKFILE - MDP and not SEPARATE_KLINES_BY_DATE */

    if ((out = open(filename, O_RDWR | O_APPEND | O_CREAT)) == -1) {
	sendto_one(sptr, ":%s NOTICE %s :Problem opening %s ", me.name, parv[0],
		   filename);
	return 0;
    }
# ifdef SEPARATE_QUOTE_KLINES_BY_DATE
    fchmod(out, 0660);
# endif
    (void) ircsprintf(buffer, "#%s!%s@%s K'd: %s@%s:%s\n",
		      sptr->name, sptr->user->username, sptr->user->host, user,
		      host, reason);

    if (write(out, buffer, strlen(buffer)) <= 0) {
	sendto_one(sptr, ":%s NOTICE %s :Problem writing to %s", me.name,
		   parv[0], filename);
	(void) close(out);
	return 0;
    }

    (void) ircsprintf(buffer, "K:%s:%s (%s):%s\n", host, reason, current_date,
		      user);

    if (write(out, buffer, strlen(buffer)) <= 0) {
	sendto_one(sptr, ":%s NOTICE %s :Problem writing to %s", me.name,
		   parv[0], filename);
	(void) close(out);
	return 0;
    }

    (void) close(out);

# ifdef USE_SYSLOG
    syslog(LOG_NOTICE, "%s added K-Line for [%s@%s] [%s]", parv[0], user, host,
	   reason);
# endif

    return 0;
#endif /* LOCKFILE */
}

/*
 * * m_unkline
 * Added Aug 31, 1997 
 * common (Keith Fralick) fralick@gate.net
 *
 *      parv[0] = sender
 *      parv[1] = address to remove
 *
 * re-worked and cleanedup for use in hybrid-5 -Dianora
 *   
 * Added comstuds SEPARATE_QUOTE_KLINES_BY_DATE
 */
int m_unkline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int in, out;
    int pairme = NO;
    char buf[256], buff[256];
    char temppath[256];

#ifdef SEPARATE_QUOTE_KLINES_BY_DATE
    char timebuffer[MAX_DATE_STRING];
    char filenamebuf[1024];
    struct tm *tmptr;
#endif

    char *filename;		/* filename to use for unkline */
    char *user, *host;
    char *p;
    int nread;
    int error_on_write = NO;
    aConfItem *aconf;
    struct stat oldfilestat;
    mode_t oldumask;

    ircsprintf(temppath, "%s.tmp", klinefile);

    if (check_registered(sptr))
	return -1;

    if (!IsAnOper(sptr) && !IsServiceModeA(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "UNKLINE");
	return 0;
    }

    if ((host = strchr(parv[1], '@')) || *parv[1] == '*') {
	/*
	 * Explicit user@host mask given 
	 */

	if (host) {		/* Found user@host */
	    user = parv[1];	/* here is user part */
	    *(host++) = '\0';	/* and now here is host */
	} else {
	    user = "*";		/* no @ found, assume its *@somehost */
	    host = parv[1];
	}
    } else {
	sendto_one(sptr, ":%s NOTICE %s :Invalid parameters", me.name, parv[0]);
	return 0;
    }

    if ((user[0] == '*') && (user[1] == '\0') && (host[0] == '*')
	&& (host[1] == '\0')) {
	sendto_one(sptr, ":%s NOTICE %s :Cannot UNK-Line everyone", me.name,
		   parv[0]);
	return 0;
    }
# if defined(LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
    if (lock_kline_file() < 0) {
	sendto_one(sptr,
		   ":%s NOTICE %s :%s is locked try again in a few minutes",
		   me.name, parv[0], klinefile);
	return -1;
    }
# endif

# ifdef SEPARATE_QUOTE_KLINES_BY_DATE
    tmptr = localtime(&NOW);
    strftime(timebuffer, MAX_DATE_STRING, "%y%m%d", tmptr);
    (void) sprintf(filenamebuf, "%s.%s", klinefile, timebuffer);
    filename = filenamebuf;
# else
    filename = klinefile;
# endif

    if ((in = open(filename, O_RDONLY)) == -1) {
	sendto_one(sptr, ":%s NOTICE %s :Cannot open %s", me.name, parv[0],
		   filename);
# if defined(LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
	(void) unlink(LOCKFILE);
# endif
	return 0;
    }
    if (fstat(in, &oldfilestat) < 0)	/* Save the old file mode */
	oldfilestat.st_mode = 0644;

    oldumask = umask(0);	/* ircd is normally too paranoid */

    if ((out = open(temppath, O_WRONLY | O_CREAT, oldfilestat.st_mode)) == -1) {
	sendto_one(sptr, ":%s NOTICE %s :Cannot open %s", me.name, parv[0],
		   temppath);
	(void) close(in);
# if defined (LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
	(void) unlink(LOCKFILE);
# endif
	umask(oldumask);	/* Restore the old umask */
	return 0;
    }
    umask(oldumask);		/* Restore the old umask */

    /*
     * #Dianora!db@ts2-11.ottawa.net K'd: foo@bar:No reason K:bar:No
     * reason (1997/08/30 14.56):foo
     */

    clear_conf_list(&KList1);
    clear_conf_list(&KList2);
    clear_conf_list(&KList3);

    while ((nread = dgets(in, buf, sizeof(buf))) > 0) {
	buf[nread] = '\0';

	if ((buf[1] == ':') && ((buf[0] == 'k') || (buf[0] == 'K'))) {
	    /*
	     * its a K: line 
	     */

    char *found_host;
    char *found_user;
    char *found_comment;

	    strcpy(buff, buf);

	    p = strchr(buff, '\n');
	    if (p)
		*p = '\0';
	    p = strchr(buff, '\r');
	    if (p)
		*p = '\0';

	    found_host = buff + 2;	/* point past the K: */

	    p = strchr(found_host, ':');
	    if (p == (char *) NULL) {
		sendto_one(sptr, ":%s NOTICE %s :K-Line file corrupted",
			   me.name, parv[0]);
		sendto_one(sptr, ":%s NOTICE %s :Couldn't find host", me.name,
			   parv[0]);
		if (!error_on_write)
		    error_on_write =
			flush_write(sptr, parv[0], out, buf, strlen(buf),
				    temppath);
		continue;	/* This K line is corrupted ignore */
	    }
	    *p = '\0';
	    p++;

	    found_comment = p;
	    p = strchr(found_comment, ':');
	    if (p == (char *) NULL) {
		sendto_one(sptr, ":%s NOTICE %s :K-Line file corrupted",
			   me.name, parv[0]);
		sendto_one(sptr, ":%s NOTICE %s :Couldn't find comment",
			   me.name, parv[0]);
		if (!error_on_write)
		    error_on_write =
			flush_write(sptr, parv[0], out, buf, strlen(buf),
				    temppath);
		continue;	/* This K line is corrupted ignore */
	    }
	    *p = '\0';
	    p++;

	    found_user = p;
	    /*
	     * Ok, if its not an exact match on either the user or the
	     * host then, write the K: line out, and I add it back to the
	     * K line tree
	     */
	    if (strcasecmp(host, found_host)
		|| strcasecmp(user, found_user)) {
		if (!error_on_write)
		    error_on_write =
			flush_write(sptr, parv[0], out, buf, strlen(buf),
				    temppath);
		aconf = make_conf();
		aconf->status = CONF_KILL;
		DupString(aconf->host, found_host);
		DupString(aconf->name, found_user);
		DupString(aconf->passwd, found_comment);
		aconf->port = 0;
    Class (aconf) = find_class(0);

		switch (sortable(found_host)) {
		    case 0:
			l_addto_conf_list(&KList3, aconf, host_field);
			break;
		    case 1:
			addto_conf_list(&KList1, aconf, host_field);
			break;
		    case -1:
			addto_conf_list(&KList2, aconf, rev_host_field);
			break;
		}
	    } else
		pairme++;
	} else if (buf[0] == '#') {
    char *userathost;
    char *found_user;
    char *found_host;

	    strcpy(buff, buf);
	    /*
	     * #Dianora!db@ts2-11.ottawa.net K'd: foo@bar:No reason
	     * K:bar:No reason (1997/08/30 14.56):foo
	     *
	     * If its a comment coment line, i.e. #ignore this line Then just
	     * ignore the line
	     */
	    p = strchr(buff, ':');
	    if (p == (char *) NULL) {
		if (!error_on_write)
		    error_on_write =
			flush_write(sptr, parv[0], out, buf, strlen(buf),
				    temppath);
		continue;
	    }
	    *p = '\0';
	    p++;

	    userathost = p;
	    p = strchr(userathost, ':');

	    if (p == (char *) NULL) {
		if (!error_on_write)
		    error_on_write =
			flush_write(sptr, parv[0], out, buf, strlen(buf),
				    temppath);
		continue;
	    }
	    *p = '\0';

	    while (*userathost == ' ')

		userathost++;

	    found_user = userathost;
	    p = strchr(found_user, '@');
	    if (p == (char *) NULL) {
		if (!error_on_write)
		    error_on_write =
			flush_write(sptr, parv[0], out, buf, strlen(buf),
				    temppath);
		continue;
	    }
	    *p = '\0';
	    found_host = p;
	    found_host++;

	    if ((strcasecmp(found_host, host))
		|| (strcasecmp(found_user, user))) {
		if (!error_on_write)
		    error_on_write =
			flush_write(sptr, parv[0], out, buf, strlen(buf),
				    temppath);
	    }
	} else {		/* its the ircd.conf file, and not a Kline or
				 * comment */
	    if (!error_on_write)
		error_on_write =
		    flush_write(sptr, parv[0], out, buf, strlen(buf), temppath);
	}
    }

    (void) close(in);

    /*
     * The result of the rename should be checked too... oh well
     */
    /*
     * If there was an error on a write above, then its been reported
     * and I am not going to trash the original kline /conf file
     * -Dianora
     */
    if ((!error_on_write) && (close(out) >= 0))
	(void) rename(temppath, filename);
    else {
	sendto_one(sptr,
		   ":%s NOTICE %s :Couldn't write temp kline file, aborted",
		   me.name, parv[0]);
# if defined (LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
	(void) unlink(LOCKFILE);
# endif
	return -1;
    }

# if defined (LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
    (void) unlink(LOCKFILE);
# endif

    if (pairme == NO) {
	if (!(pairme = remove_temp_kline(host, user, CONF_KILL))) {
	    sendto_one(sptr, ":%s NOTICE %s :No K-Line for %s@%s", me.name,
		       parv[0], user, host);
	    return 0;
	}
    }
    sendto_one(sptr, ":%s NOTICE %s :K-Line for [%s@%s] is removed", me.name,
	       parv[0], user, host);
    sendto_ops("%s has removed the K-Line for: [%s@%s] (%d matches)", parv[0],
	       user, host, pairme);
    return 0;
}

/*
 * re-worked a tad added Rodders dated KLINE code -Dianora
 *
 * BUGS:        There is a concern here with LOCKFILE handling the LOCKFILE
 * code only knows how to talk to the kline file. Having an extra
 * define allowing D lines to go into ircd.conf or the kline file
 * complicates life. LOCKFILE code should still respect any lock
 * placed.. The fix I believe is that we are going to have to also pass
 * along in the struct pkl struct, which file the entry is to go
 * into... or.. just remove the ZLINES_IN_KPATH option. -Dianora
 */

int m_zline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
#if defined(LOCKFILE) && defined(ZLINES_IN_KPATH)
    struct pkl *k;
#else
    int out;
#endif
    char buffer[1024];
    char *host, *reason, *p;
    char *current_date;
    int number_of_dots = 0;
    aClient *acptr;
    aConfItem *aconf;

    if (!MyClient(sptr) || !IsOper(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    if (parc < 2) {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		       "ZLINE");
	return 0;
    }

    host = parv[1];

    p = host;
    while (*p) {
	if (*p == '.') {
	    number_of_dots++;
	    p++;
	} else if (*p == '*') {
	    if (number_of_dots != 3) {
		sendto_one(sptr,
			   ":%s NOTICE %s :Z-line format error, use xxx.xxx.xxx.xxx or xxx.xxx.xxx.*",
			   me.name, parv[0]);
		return 0;
	    } else {
		*(p + 1) = '\0';
		break;
	    }
	} else {
	    if (!isdigit(*p)) {
		if (!(acptr = find_chasing(sptr, parv[1], NULL)))
		    return 0;

		if (!acptr->user)
		    return 0;

		if (IsServer(acptr)) {
		    sendto_one(sptr,
			       ":%s NOTICE %s :Can't ZLINE a server silly",
			       me.name, parv[0]);
		    return 0;
		}

		if (!MyConnect(acptr)) {
		    sendto_one(sptr,
			       ":%s NOTICE :%s :Can't ZLINE nick on another server",
			       me.name, parv[0]);
		    return 0;
		}

		host = cluster(acptr->hostip);
		break;
	    }
	    p++;
	}
    }

    if (parc > 2) {		/* host :reason */
	if (strchr(parv[2], ':')) {
	    sendto_one(sptr, ":%s NOTICE %s :Invalid character ':' in comment",
		       me.name, parv[0]);
	    return 0;
	}

	if (strchr(parv[2], '#')) {
	    sendto_one(sptr, ":%s NOTICE %s :Invalid character '#' in comment",
		       me.name, parv[0]);
	    return 0;
	}

	if (*parv[2])
	    reason = parv[2];
	else
	    reason = "No reason";
    } else
	reason = "No reason";

#ifdef NON_REDUNDANT_KLINES
    if ((aconf = find_is_zlined(host))) {
    char *reason = aconf->passwd ? aconf->passwd : "<No Reason>";

	sendto_one(sptr,
		   ":%s NOTICE %s :[%s] already Z-lined by [%s] - %s",
		   me.name, parv[0], host, aconf->host, reason);
	return 0;

    }
#endif

    current_date = smalldate((time_t) 0);

    aconf = make_conf();
    aconf->status = CONF_ZLINE;
    DupString(aconf->host, host);

    (void) ircsprintf(buffer, "%s (%s)", reason, current_date);

    DupString(aconf->passwd, buffer);
    aconf->name = (char *) NULL;
    aconf->port = 0;
    Class (aconf) = find_class(0);

    /*
     * since a quote zline always ensures a sortable pattern there is no
     * need to check sortable() - Dianora
     */

    addto_conf_list(&ZList1, aconf, host_field);

    sendto_realops("%s added Z-Line for [%s] [%s]", parv[0], host, reason);

#ifdef ZLINES_IN_KPATH
    sendto_one(sptr,
	       ":%s NOTICE %s :Added Z-Line [%s] to server klinefile", me.name,
	       parv[0], host);
#else
    sendto_one(sptr,
	       ":%s NOTICE %s :Added Z-Line [%s] to server configfile", me.name,
	       parv[0], host);
#endif
    /*
     * * I moved the following 2 lines up here because we still want
     * the server to hunt for 'targetted' clients even if there are 
     * problems adding the Z-line the the appropriate file. -ThemBones
     */
    rehashed = YES;
    zline_in_progress = YES;

#if defined(LOCKFILE) && defined(ZLINES_IN_KPATH)

    if ((k = (struct pkl *) MyMalloc(sizeof(struct pkl))) == NULL) {
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name,
		   parv[0]);
	return (0);
    }

    (void) ircsprintf(buffer, "#%s!%s@%s D'd: %s:%s (%s)\n",
		      sptr->name, sptr->user->username,
		      sptr->user->host, host, reason, current_date);

    if ((k->comment = strdup(buffer)) == NULL) {
	free(k);
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name,
		   parv[0]);
	return (0);
    }

    (void) ircsprintf(buffer, "D:%s:%s (%s):\n", host, reason, current_date);

    if ((k->kline = strdup(buffer)) == NULL) {
	free(k->comment);
	free(k);
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name,
		   parv[0]);
	return (0);
    }
    k->next = pending_klines;
    pending_klines = k;

    do_pending_klines();
    return (0);
#else /* LOCKFILE - MDP */

    if ((out = open(zlinefile, O_RDWR | O_APPEND | O_CREAT)) == -1) {
	sendto_one(sptr, ":%s NOTICE %s :Problem opening %s ", me.name, parv[0],
		   zlinefile);
	return 0;
    }

    (void) ircsprintf(buffer, "#%s!%s@%s Z'd: %s:%s (%s)\n",
		      sptr->name, sptr->user->username,
		      sptr->user->host, host, reason, current_date);

    if (write(out, buffer, strlen(buffer)) <= 0) {
	sendto_one(sptr, ":%s NOTICE %s :Problem writing to %s", me.name,
		   parv[0], zlinefile);
	close(out);
	return 0;
    }

    (void) ircsprintf(buffer, "Z:%s:%s (%s):\n", host, reason, current_date);

    if (write(out, buffer, strlen(buffer)) <= 0) {
	sendto_one(sptr, ":%s NOTICE %s :Problem writing to %s", me.name,
		   parv[0], zlinefile);
	close(out);
	return 0;
    }

    (void) close(out);
    return 0;
#endif /* LOCKFILE */
}

/*
 * * m_rehash *
 * 
 */
int m_rehash(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (!OPCanRehash(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    if (parc > 1) {
	if (mycmp(parv[1], "DNS") == 0) {
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "DNS");
	    flush_cache();	/* flush the dns cache */
	    res_init();		/* re-read /etc/resolv.conf */
	    sendto_ops("%s is rehashing DNS while whistling innocently",
		       parv[0]);
	    return 0;
	} else if (mycmp(parv[1], "SZLINES") == 0) {
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0],
		       "SLINES");
	    remove_szline("*", 1);	/* flush the dns cache */
	    sendto_ops("%s is removing SZLINES while whistling innocently",
		       parv[0]);
	    return 0;
	} else if (mycmp(parv[1], "TKLINES") == 0) {
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0],
		       "temp klines");
	    flush_temp_klines();
	    sendto_ops("%s is clearing temp klines while whistling innocently",
		       parv[0]);
	    return 0;
	} else if (mycmp(parv[1], "GC") == 0) {
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0],
		       "garbage collecting");
	    block_garbage_collect();
	    sendto_ops("%s is garbage collecting while whistling innocently",
		       parv[0]);
	    return 0;
	} else if (mycmp(parv[1], "MOTD") == 0) {
	    sendto_ops("%s is forcing re-reading of MOTD file", parv[0]);
	    read_motd(MOTD);
#ifdef SHORT_MOTD
	    read_shortmotd(SHORTMOTD);
#endif
	    return (0);
	} else if (mycmp(parv[1], "IP") == 0) {
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0],
		       "ip hash");
	    rehash_ip_hash();
	    sendto_ops("%s is rehashing iphash while whistling innocently",
		       parv[0]);
	} else if (mycmp(parv[1], "AKILLS") == 0) {
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0],
		       "akills");
	    do_rehash_akills();
	    sendto_ops("%s is rehashing akills", parv[0]);
	    return 0;
	} else if (mycmp(parv[1], "EXCLUDES") == 0) {
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0],
		       "excludes");
	    do_rehash_excludes();
	    sendto_ops("%s is rehashing excludes", parv[0]);
	    return 0;
	} else if (mycmp(parv[1], "THROTTLES") == 0) {
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0],
		       "throttles");
	    throttle_rehash();
	    sendto_ops("%s is rehashing throttles", parv[0]);
	    return 0;
	}
    } else {
	sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], configfile);
	sendto_ops
	    ("%s is rehashing Server config file while whistling innocently",
	     parv[0]);
# ifdef USE_SYSLOG
	syslog(LOG_INFO, "REHASH From %s\n", get_client_name(sptr, FALSE));
# endif
	return rehash(cptr, sptr, (parc > 1) ? ((*parv[1] == 'q') ? 2 : 0) : 0);
    }
    return 0;			/* shouldn't ever get here */
}

/*
 * * m_restart *
 * 
 */
int m_restart(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *pass = NULL;

    if (!OPCanRestart(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    /*
     * m_restart is now password protected as in df465 only change --
     * this one doesn't allow a reason to be specified. future changes:
     * crypt()ing of password, reason to be re-added -mjs
     */
    if ((pass = (char *) find_restartpass())) {
	if (parc < 2) {
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		       "RESTART");
	    return 0;
	}
	if (strcmp(pass, parv[1])) {
	    sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
	    return 0;
	}
    }
#ifdef USE_SYSLOG
    syslog(LOG_WARNING, "Server RESTART by %s\n", get_client_name(sptr, FALSE));
#endif
    sprintf(buf, "Server RESTART by %s", get_client_name(sptr, TRUE));
    restart(buf);
    return 0;			/* NOT REACHED */
}

/*
 * * m_close - added by Darren Reed Jul 13 1992.
 */
int m_close(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    int i;
    int closed = 0;

    if (!MyOper(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    for (i = highest_fd; i; i--) {
	if (!(acptr = local[i]))
	    continue;
	if (!IsUnknown(acptr) && !IsConnecting(acptr) && !IsHandshake(acptr))
	    continue;
	sendto_one(sptr, rpl_str(RPL_CLOSING), me.name, parv[0],
		   get_client_name(acptr, TRUE), acptr->status);
	(void) exit_client(acptr, acptr, acptr, "Oper Closing");
	closed++;
    }
    sendto_one(sptr, rpl_str(RPL_CLOSEEND), me.name, parv[0], closed);
    return 0;
}

int m_die(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    int i;
    char *pass = NULL;

    if (!OPCanDie(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    /*
     * X line -mjs 
     */

    if ((pass = (char *) find_diepass())) {
	if (parc < 2) {
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		       "DIE");
	    return 0;
	}
	if (strcmp(pass, parv[1])) {
	    sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
	    return 0;
	}
    }

    for (i = 0; i <= highest_fd; i++) {
	if (!(acptr = local[i]))
	    continue;
	if (IsClient(acptr))
	    sendto_one(acptr,
		       ":%s NOTICE %s :Server Terminating. %s", me.name,
		       acptr->name, sptr->name);
	else if (IsServer(acptr))
	    sendto_one(acptr, ":%s ERROR :Terminated by %s", me.name,
		       get_client_name(sptr, TRUE));
    }
    (void) s_die();
    return 0;
}

/*
 * Shadowfax's LOCKFILE code
 */
#ifdef LOCKFILE

int lock_kline_file()
{
int fd;

    /*
     * Create Lockfile 
     */

    if ((fd = open(LOCKFILE, O_WRONLY | O_CREAT | O_EXCL, 0666)) < 0) {
	sendto_realops("%s is locked, klines pending", klinefile);
	pending_kline_time = time(NULL);
	return (-1);
    }
    (void) close(fd);
    return (1);
}

void do_pending_klines()
{
int fd;
char s[20];
struct pkl *k, *ok;

    if (!pending_klines)
	return;

    /*
     * Create Lockfile 
     */
    if ((fd = open(LOCKFILE, O_WRONLY | O_CREAT | O_EXCL, 0666)) < 0) {
	sendto_realops("%s is locked, klines pending", klinefile);
	pending_kline_time = time(NULL);
	return;
    }
    (void) ircsprintf(s, "%d\n", getpid());
    (void) write(fd, s, strlen(s));
    close(fd);

    /*
     * Open klinefile 
     */
    if ((fd = open(klinefile, O_WRONLY | O_APPEND)) == -1) {
	sendto_realops("Pending klines cannot be written, cannot open %s",
		       klinefile);
	unlink(LOCKFILE);
	return;
    }
    /*
     * Add the Pending Klines 
     */

    k = pending_klines;
    while (k) {
	write(fd, k->comment, strlen(k->comment));
	write(fd, k->kline, strlen(k->kline));
	free(k->comment);
	free(k->kline);
	ok = k;
	k = k->next;
	free(ok);
    }
    pending_klines = NULL;
    pending_kline_time = 0;

    close(fd);

    /*
     * Delete the Lockfile 
     */
    unlink(LOCKFILE);
}
#endif

/*
 * * m_kill * parv[0] = sender prefix *       parv[1] = kill victim *
 * parv[2] = kill path
 */
int m_kill(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    char *user, *path, *killer, *p, *nick;
    char mypath[KILLLEN + 1];
    char *unknownfmt = "<Unknown>";	/* 
					 * AFAIK this shouldnt happen
					 * * but -Raist
					 */
    int chasing = 0, kcount = 0;

    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "KILL");
	return 0;
    }

    user = parv[1];
    path = parv[2];		/* 
				 * Either defined or NULL (parc >= 2!!)
				 */
    if (path == NULL)
	path = ")";

    if (!IsPrivileged(cptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (IsAnOper(cptr)) {
	if (!BadPtr(path))
	    if (strlen(path) > (size_t) KILLLEN)
		path[KILLLEN] = '\0';
    }
    if (MyClient(sptr))
	user = canonize(user);
    for (p = NULL, nick = strtoken(&p, user, ","); nick;
	 nick = strtoken(&p, NULL, ",")) {
	chasing = 0;
	if (!(acptr = find_client(nick, NULL))) {
	    /*
	     * * If the user has recently changed nick, we automaticly *
	     * rewrite the KILL for this new nickname--this keeps *
	     * servers in synch when nick change and kill collide
	     */
	    if (!(acptr = get_history(nick, (long) KILLCHASETIMELIMIT))) {
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0],
			   nick);
		return 0;
	    }
	    sendto_one(sptr, ":%s NOTICE %s :KILL changed from %s to %s",
		       me.name, parv[0], nick, acptr->name);
	    chasing = 1;
	}
	if ((!MyConnect(acptr) && MyClient(cptr) && !OPCanGKill(cptr)) ||
	    (MyConnect(acptr) && MyClient(cptr) && !OPCanLKill(cptr))) {
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    continue;
	}
	if (IsServer(acptr) || IsMe(acptr)) {
	    sendto_one(sptr, err_str(ERR_CANTKILLSERVER), me.name, parv[0]);
	    continue;
	}
	kcount++;
	if (!IsServer(sptr) && (kcount > MAXKILLS)) {
	    sendto_one(sptr,
		       ":%s NOTICE %s :Too many targets, kill list was truncated. Maximum is %d.",
		       me.name, sptr->name, MAXKILLS);
	    break;
	}

	if (MyClient(sptr)) {
    char myname[HOSTLEN + 1], *s;
    int slen;

	    strncpy(myname, me.name, 80);
	    if ((s = index(myname, '.')))
		*s = 0;

	    /*
	     * "<myname>!<sptr->user->host>!<sptr->name> (path)" 
	     */
	    slen =
		KILLLEN - (strlen(sptr->name) + strlen(sptr->user->host) +
			   strlen(myname) + 8);
	    if (slen < 0)
		slen = 0;

	    if (strlen(path) > slen)
		path[slen] = '\0';

	    ircsprintf(mypath, "%s!%s!%s (%s)", myname,
#ifdef USE_FAKEHOST
		       IsFake(sptr) ? sptr->user->fakehost : sptr->user->host,
#else
		       sptr->user->host,
#endif
		       sptr->name, path);
	    mypath[KILLLEN] = '\0';
	} else
	    strncpy(mypath, path, KILLLEN + 1);
	/*
	 * * Notify all *local* opers about the KILL, this includes the
	 * one * originating the kill, if from this server--the special
	 * numeric * reply message is not generated anymore. *
	 * 
	 * Note: "acptr->name" is used instead of "user" because we may *
	 * ave changed the target because of the nickname change.
	 */
	if (IsLocOp(sptr) && !MyConnect(acptr)) {
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
	if (IsAnOper(sptr))
	    sendto_ops_lev(0,
			   "Received KILL message for %s!%s@%s. From %s Path: %s",
			   acptr->name,
			   acptr->user ? acptr->user->username : unknownfmt,
#ifdef USE_FAKEHOST
			   acptr->user ? (IsFake(acptr) ? acptr->user->
					  fakehost : acptr->user->
					  host) : unknownfmt,
#else
			   acptr->user ? acptr->user->host : unknownfmt,
#endif
			   parv[0], mypath);
	else
	    sendto_ops_lev(SKILL_LEV,
			   "Received KILL message for %s!%s@%s. From %s Path: %s",
			   acptr->name,
			   acptr->user ? acptr->user->username : unknownfmt,
#ifdef USE_FAKEHOST
			   acptr->user ? (IsFake(acptr) ? acptr->user->
					  fakehost : acptr->user->
					  host) : unknownfmt,
#else
			   acptr->user ? acptr->user->host : unknownfmt,
#endif
			   parv[0], mypath);

#if defined(USE_SYSLOG) && defined(SYSLOG_KILL)
	if (IsOper(sptr))
	    syslog(LOG_INFO, "KILL From %s!%s@%s For %s Path %s",
		   parv[0], acptr->name,
		   acptr->user ? acptr->user->username : unknownfmt,
		   acptr->user ? acptr->user->host : unknownfmt, mypath);
#endif
	/*
	 * * And pass on the message to other servers. Note, that if KILL *
	 * was changed, the message has to be sent to all links, also *
	 * back. * Suicide kills are NOT passed on --SRB
	 */
	if (!MyConnect(acptr) || !MyConnect(sptr) || !IsAnOper(sptr)) {
	    sendto_serv_butone(cptr, ":%s KILL %s :%s", parv[0], acptr->name,
			       mypath);
	    if (chasing && IsServer(cptr))
		sendto_one(cptr, ":%s KILL %s :%s", me.name, acptr->name,
			   mypath);
	    acptr->flags |= FLAGS_KILLED;

	    sendto_service(SERVICE_SEE_KILLS, SERVICE_SEE_PREFIX,
			   ":%s KILL %s :%s", parv[0], acptr->name, mypath);
	    if (sptr->user)
		sendto_service(SERVICE_SEE_KILLS | SERVICE_SEE_PREFIX, 0,
			       ":%s!%s@%s KILL %s :%s", parv[0],
			       sptr->user->username, sptr->user->host,
			       acptr->name, mypath);
	}
	/*
	 * * Tell the victim she/he has been zapped, but *only* if * the
	 * victim is on current server--no sense in sending the *
	 * notification chasing the above kill, it won't get far * anyway
	 * (as this user don't exist there any more either)
	 */
	if (MyConnect(acptr)) {
	    sendto_prefix_one(acptr, sptr, ":%s KILL %s :%s", parv[0],
			      acptr->name, mypath);
	    sendto_service(SERVICE_SEE_KILLS, SERVICE_SEE_PREFIX,
			   ":%s KILL %s :%s", parv[0], acptr->name, mypath);
	    if (sptr->user)
		sendto_service(SERVICE_SEE_KILLS | SERVICE_SEE_PREFIX, 0,
			       ":%s!%s@%s KILL %s :%s",
			       parv[0], sptr->user->username, sptr->user->host,
			       acptr->name, mypath);
	}

	/*
	 * * Set FLAGS_KILLED. This prevents exit_one_client from sending *
	 * the unnecessary QUIT for this. ,This flag should never be *
	 * set in any other place...
	 */
	if (MyConnect(acptr) && MyConnect(sptr) && IsAnOper(sptr))
	    (void) ircsprintf(buf2, "Local kill by %s (%s)", sptr->name,
			      BadPtr(parv[2]) ? sptr->name : parv[2]);
	else {
	    killer = strchr(mypath, '(');
	    if (killer == NULL)
		killer = "()";
	    (void) ircsprintf(buf2, "Killed (%s %s)", sptr->name, killer);
	}
	if (exit_client(cptr, acptr, sptr, buf2) == FLUSH_BUFFER)
	    return FLUSH_BUFFER;
    }
    return 0;
}

/***********************************************************************
 * m_connect() - Added by Jto 11 Feb 1989
 ***********************************************************************/
/*
 * * m_connect
 *      parv[0] = sender prefix
 *      parv[1] = servername
 *      parv[2] = port number
 *      parv[3] = remote server
 */
int m_connect(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int port, tmpport, retval;
    aConfItem *aconf;
    aClient *acptr;

    if (!IsPrivileged(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return -1;
    }

    if ((MyClient(sptr) && !OPCanGRoute(sptr) && parc > 3) ||
	(MyClient(sptr) && !OPCanLRoute(sptr) && parc <= 3)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    if (hunt_server(cptr, sptr, ":%s CONNECT %s %s :%s", 3, parc, parv) !=
	HUNTED_ISME)
	return 0;

    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "CONNECT");
	return -1;
    }

    if ((acptr = find_server(parv[1], NULL))) {
	sendto_one(sptr, ":%s NOTICE %s :Connect: Server %s %s %s.",
		   me.name, parv[0], parv[1], "already exists from",
		   acptr->from->name);
	return 0;
    }

    for (aconf = conf; aconf; aconf = aconf->next)
	if (aconf->status == CONF_CONNECT_SERVER &&
	    match(parv[1], aconf->name) == 0)
	    break;

    /*
     * Checked first servernames, then try hostnames. 
     */
    if (!aconf)
	for (aconf = conf; aconf; aconf = aconf->next)
	    if (aconf->status == CONF_CONNECT_SERVER &&
		(match(parv[1], aconf->host) == 0 ||
		 match(parv[1], strchr(aconf->host, '@') + 1) == 0))
		break;

    if (!aconf) {
	sendto_one(sptr, "NOTICE %s :Connect: No C line found for %s.", parv[0],
		   parv[1]);
	return 0;
    }
    /*
     * * Get port number from user, if given. If not specified, use
     * the default form configuration structure. If missing from   
     * there, then use the precompiled default.
     */
    tmpport = port = aconf->port;
    if (parc > 2 && !BadPtr(parv[2])) {
	if ((port = atoi(parv[2])) <= 0) {
	    sendto_one(sptr, "NOTICE %s :Connect: Illegal port number",
		       parv[0]);
	    return 0;
	}
    }
    /*
     * * Notify all operators about remote connect requests
     * Let's notify about local connects, too. - lucas
     * sendto_ops_butone -> sendto_serv_butone(), like in df. -mjs
     */
    sendto_gnotice("from %s: %s CONNECT %s %s from %s",
		   me.name, IsAnOper(cptr) ? "Local" : "Remote",
		   parv[1], parv[2] ? parv[2] : "", get_client_name(sptr,
								    HIDEME));
    sendto_serv_butone(NULL, ":%s GNOTICE :%s CONNECT %s %s from %s", me.name,
		       IsAnOper(cptr) ? "Local" : "Remote", parv[1],
		       parv[2] ? parv[2] : "", get_client_name(sptr, HIDEME));

#if defined(USE_SYSLOG) && defined(SYSLOG_CONNECT)
    syslog(LOG_DEBUG, "CONNECT From %s : %s %s", parv[0], parv[1],
	   parv[2] ? parv[2] : "");
#endif

    aconf->port = port;
    switch (retval = connect_server(aconf, sptr, NULL)) {
	case 0:
	    sendto_one(sptr, ":%s NOTICE %s :*** Connecting to %s.", me.name,
		       parv[0], aconf->name);
	    break;
	case -1:
	    sendto_one(sptr, ":%s NOTICE %s :*** Couldn't connect to %s.",
		       me.name, parv[0], aconf->name);
	    break;
	case -2:
	    sendto_one(sptr, ":%s NOTICE %s :*** Host %s is unknown.",
		       me.name, parv[0], aconf->name);
	    break;
	default:
	    sendto_one(sptr,
		       ":%s NOTICE %s :*** Connection to %s failed: %s",
		       me.name, parv[0], aconf->name, strerror(retval));
    }
    aconf->port = tmpport;
    return 0;
}
