/************************************************************************
 *   IRC - Internet Relay Chat, src/cmd_serv.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: cmd_serv.c,v 1.10 2001/11/24 23:01:19 tr-ircd Exp $ 
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "nameser.h"
#include "resolv.h"
#include "dh.h"
#include "zlink.h"

#if defined(AIX) || defined(DYNIXPTX) || defined(SVR3)
# include <time.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "h.h"
#if defined( HAVE_STRING_H )
# include <string.h>
#else
  /*
   * older unices don't have strchr/strrchr .. help them out 
   */
# include <strings.h>
# undef strchr
# define strchr index
#endif
#include "dich_conf.h"
#include "fdlist.h"

extern fdlist serv_fdlist;
extern int lifesux;
static char buf[BUFSIZE];
extern int rehashed;
extern int zline_in_progress;

#ifdef HIGHEST_CONNECTION
int max_connection_count = 1, max_client_count = 1;
#endif

/*
 * external variables 
 */

/*
 * external functions 
 */

#ifdef MAXBUFFERS
extern void reset_sock_opts();
#endif

extern char *smalldate(time_t);	/* defined in s_misc.c */
extern void outofmemory(void);	/* defined in list.c */
extern void s_die(void);	/* defined in ircd.c as VOIDSIG */
extern int match(char *, char *);	/* defined in match.c */
extern void report_conf_links(aClient *, aConfList *, int, char);
extern void show_opers(aClient *, char *);
extern void show_servers(aClient *, char *);
extern void count_memory(aClient *, char *);
extern void rehash_ip_hash();	/* defined in s_conf.c */
extern char *find_restartpass();
extern char *find_diepass();

/*
 * Local function prototypes 
 */

int send_motd(aClient *, aClient *, int, char **);
void read_motd(char *);

#ifdef SHORT_MOTD
void read_shortmotd(char *);
#endif

char motd_last_changed_date[MAX_DATE_STRING];	/* enough room for date */

/*
 * m_svskill - Just about the same as outta df
 * *  - Raistlin
 * * parv[0] = servername
 * * parv[1] = client
 * * parv[2] = nick stamp
 * * parv[3] = kill message
 */

int m_svskill(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    char *comment;
    char reason[TOPICLEN + 1];
    ts_val ts = 0;

    if (parc > 3) {
	comment = parv[3] ? parv[3] : parv[0];
	ts = atol(parv[2]);
    } else
	comment = (parc > 2 && parv[2]) ? parv[2] : parv[0];

    if (!IsULine(sptr))
	return -1;
    if ((acptr = find_client(parv[1], NULL))
	&& (!ts || ts == acptr->tsinfo)) {
	if (MyClient(acptr)) {
	    strcpy(reason, "Quit: ");
	    strncpy(reason + 6, comment, TOPICLEN - 11);
	    reason[TOPICLEN] = '\0';
	    return exit_client(acptr, acptr, sptr, reason);
	}
	if (acptr->from == cptr) {
	    sendto_ops_lev(DEBUG_LEV,
			   "Received wrong-direction SVSKILL for %s (behind %s) from %s",
			   acptr->name, cptr->name, get_client_name(sptr,
								    HIDEME));
	    return 0;
	} else
	    sendto_serv_butone(cptr, ":%s SVSKILL %s :%s", parv[0], parv[1],
			       comment);
    }
    return 0;
}

/*
 * m_akill -
 * * Parse AKILL command
 * * parv[1]=host, parv[2]=user, parv[3]=length, parv[4]=akiller, parv[5]=time set, parv[6]=reason
 */
int m_akill(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aConfItem *aconf, *ac2;
    char *user, *host, *reason, *akiller, buffer[1024], *current_date;
    time_t length = 0, timeset = 0;

    if ((!IsServer(sptr) && !IsServiceModeA(sptr)) || (parc < 6))
	return 0;

    if (!IsULine(sptr)) {
	sendto_serv_butone(&me,
			   ":%s GLOBOPS :Non-ULined server %s trying to AKILL!",
			   me.name, sptr->name);
	send_globops("From %s: Non-ULined server %s trying to AKILL!", me.name,
		     sptr->name);
	return 0;
    }

    host = parv[1];
    user = parv[2];
    akiller = parv[4];
    length = atoi(parv[3]);
    timeset = atoi(parv[5]);
    reason = (parv[6] ? parv[6] : "<no reason>");

    current_date = smalldate((time_t) timeset);
    /*
     * cut reason down a little, eh? 
     */
    /*
     * 250 chars max 
     */
    if (strlen(reason) > 250)
	reason[251] = 0;

    /*
     * if this is already klined, don't akill it now 
     */
    ac2 = find_is_klined(host, user, (aClient *) NULL);
    if (ac2 != NULL) {
	/*
	 * pass along the akill anyways 
	 */
	sendto_serv_butone(cptr, ":%s AKILL %s %s %d %s %d :%s",
			   sptr->name, host, user, length, akiller, timeset,
			   reason);
	return 0;
    }
    /*
     * fill out aconf 
     */
    aconf = make_conf();
    aconf->status = CONF_AKILL;
    DupString(aconf->host, host);
    DupString(aconf->name, user);

    ircsprintf(buffer, "%s (%s)", reason, current_date);

    DupString(aconf->passwd, buffer);

    /*
     * whether or not we have a length, this is still a temporary akill 
     */
    /*
     * if the length is over a day, or nonexistant, we call this a 'forever'
     * * akill and set ->hold to 0xFFFFFFFF to indicate such, this is a hack to prevent
     * * forever akills from being removed unless by an explicit /rehash 
     */
    if (length > 86400 || !length)
	aconf->hold = 0xFFFFFFFF;
    else
	aconf->hold = timeset + length;

    add_temp_kline(aconf);
    rehashed = YES;
    zline_in_progress = NO;

    /*
     * now finally send it off to any other servers! 
     */
    sendto_serv_butone(cptr, ":%s AKILL %s %s %d %s %d :%s",
		       sptr->name, host, user, length, akiller, timeset,
		       reason);
    return 0;
}

int m_rakill(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (!IsServer(sptr) && !IsServiceModeA(sptr))
	return 0;
    /*
     * just quickly find the akill and be rid of it! 
     */
    if (parc < 3) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "RAKILL");
	return 0;
    }
    if (!IsULine(sptr)) {
	sendto_serv_butone(&me,
			   ":%s GLOBOPS :Non-ULined server %s trying to RAKILL!",
			   me.name, sptr->name);
	send_globops("From %s: Non-ULined server %s trying to RAKILL!", me.name,
		     sptr->name);
	return 0;
    }
    remove_temp_kline(parv[1], parv[2], CONF_AKILL);
    sendto_serv_butone(cptr, ":%s RAKILL %s %s", sptr->name, parv[1], parv[2]);
    return 0;
}

int m_sqline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aConfItem *aconf;

    if (!(IsServer(sptr) || IsULine(sptr)))
	return 0;

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "SQLINE");
	return 0;
    }

    /*
     * get rid of redundancies 
     */
    parv[1] = collapse(parv[1]);

    /*
     * if we have any Q:lines (SQ or Q) that match
     * * this Q:line, just return (no need to waste cpu 
     */

    if (!
	(aconf =
	 find_conf_name(parv[1],
			*parv[1] ==
			'#' ? CONF_QUARANTINED_CHAN : CONF_QUARANTINED_NICK))) {
	/*
	 * okay, it doesn't suck, build a new conf for it 
	 */
	aconf = make_conf();

	/*
	 * Q:line and SQline, woo 
	 */
	if (*parv[1] == '#')
	    aconf->status = (CONF_QUARANTINED_CHAN | CONF_SQLINE);
	else
	    aconf->status = (CONF_QUARANTINED_NICK | CONF_SQLINE);
	DupString(aconf->name, parv[1]);
	DupString(aconf->passwd, (parv[2] != NULL ? parv[2] : "Reserved"));
	aconf->next = conf;
	conf = aconf;
    }
    sendto_serv_butone(cptr, ":%s SQLINE %s :%s", sptr->name, parv[1],
		       aconf->passwd);
    return 0;
}

int m_unsqline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int matchit = 0;
    char *mask;
    aConfItem *aconf, *ac2, *ac3;

    if (!(IsServer(sptr) || IsULine(sptr)))
	return 0;

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "UNSQLINE");
	return 0;
    }

    if (parc == 3) {
	matchit = atoi(parv[1]);
	mask = parv[2];
    } else
	mask = parv[1];

    /*
     * find the sqline and remove it 
     */
    /*
     * this was way too complicated and ugly. Fixed. -lucas 
     */
    /*
     * Changed this to use match.  Seems to make more sense? 
     */
    /*
     * Therefore unsqline * will clear the sqline list. 
     */

    ac2 = ac3 = aconf = conf;
    while (aconf) {
	if ((aconf->status & (CONF_QUARANTINED_NICK))
	    && (aconf->status & (CONF_SQLINE)) &&
	    aconf->name && (matchit ? !match(mask, aconf->name)
			    : !mycmp(mask, aconf->name))) {
	    if (conf == aconf)
		ac2 = ac3 = conf = aconf->next;
	    else
		ac2 = ac3->next = aconf->next;
	    MyFree(aconf->passwd);
	    MyFree(aconf->name);
	    MyFree(aconf);
	    aconf = ac2;
	} else {
	    ac3 = aconf;
	    aconf = aconf->next;
	}
    }
    if (parc == 3)
	sendto_serv_butone(cptr, ":%s UNSQLINE %d :%s", sptr->name, matchit,
			   mask);
    else
	sendto_serv_butone(cptr, ":%s UNSQLINE :%s", sptr->name, mask);
    return 0;
}

int m_sgline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aConfItem *aconf;
    int len;
    char *mask, *reason;

    if (!(IsServer(sptr) || IsULine(sptr)))
	return 0;

    if (parc < 3) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "SGLINE");
	return 0;
    }

    len = atoi(parv[1]);
    mask = parv[2];
    if ((strlen(mask) > len) && (mask[len]) == ':') {
	mask[len] = '\0';
	reason = mask + len + 1;
    } else {			/* Bogus */
	return 0;
    }

    /*
     * if we have any G:lines (SG or G) that match
     * * this G:line, just return (no need to waste cpu 
     */

    if (!(aconf = find_conf_name(mask, CONF_GCOS))) {
	/*
	 * okay, it doesn't suck, build a new conf for it 
	 */
	aconf = make_conf();

	aconf->status = (CONF_GCOS | CONF_SGLINE);	/* G:line and SGline,
							 * woo */
	DupString(aconf->name, mask);
	DupString(aconf->passwd, reason ? reason : "Reason Unspecified");
	aconf->next = conf;
	conf = aconf;
    }
    sendto_serv_butone(cptr, ":%s SGLINE %d :%s:%s", sptr->name, len,
		       aconf->name, aconf->passwd);
    return 0;
}

int m_unsgline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int matchit = 0;
    char *mask;
    aConfItem *aconf, *ac2, *ac3;

    if (!(IsServer(sptr) || IsULine(sptr)))
	return 0;

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "UNSGLINE");
	return 0;
    }

    if (parc == 3) {
	matchit = atoi(parv[1]);
	mask = parv[2];
    } else
	mask = parv[1];

    /*
     * find the sgline and remove it 
     */
    /*
     * this was way too complicated and ugly. Fixed. -lucas 
     */
    /*
     * Changed this to use match.  Seems to make more sense? 
     */
    /*
     * Therefore unsgline * will clear the sgline list. 
     */
    ac2 = ac3 = aconf = conf;
    while (aconf) {
	if ((aconf->status & (CONF_GCOS))
	    && (aconf->status & (CONF_SGLINE)) &&
	    aconf->name && (matchit ? !match(mask, aconf->name)
			    : !mycmp(mask, aconf->name))) {
	    if (conf == aconf)
		ac2 = ac3 = conf = aconf->next;
	    else
		ac2 = ac3->next = aconf->next;
	    MyFree(aconf->passwd);
	    MyFree(aconf->name);
	    MyFree(aconf);
	    aconf = ac2;
	} else {
	    ac3 = aconf;
	    aconf = aconf->next;
	}
    }

    if (parc == 3)
	sendto_serv_butone(cptr, ":%s UNSGLINE %d :%s", sptr->name, matchit,
			   mask);
    else
	sendto_serv_butone(cptr, ":%s UNSGLINE :%s", sptr->name, mask);
    return 0;
}

int m_szline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aConfItem *aconf;

    if (!(IsServer(sptr) || IsULine(sptr)))
	return 0;

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "SZLINE");
	return 0;
    }

    if (!(aconf = find_is_zlined(parv[1]))) {
	/*
	 * okay, it doesn't suck, build a new conf for it 
	 */
	aconf = make_conf();
	aconf->status = (CONF_ZLINE | CONF_SZLINE);	/* Z:line and SZline,
							 * woo */
	DupString(aconf->host, parv[1]);
	DupString(aconf->passwd, (parv[2] != NULL ? parv[2] : "No Reason"));
	aconf->name = (char *) NULL;
	add_szline(aconf);
    }
    sendto_serv_butone(cptr, ":%s SZLINE %s :%s", sptr->name, parv[1],
		       aconf->passwd);
    return 0;
}

int m_unszline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int matchit = 0;
    char *mask;

    if (!IsServer(sptr) || !IsULine(sptr))
	return 0;
    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "UNSZLINE");
	return 0;
    }
    if (parc == 3) {
	matchit = atoi(parv[1]);
	mask = parv[2];
    } else
	mask = parv[1];

    remove_szline(mask, matchit ? 1 : 0);
    if (parc == 3)
	sendto_serv_butone(cptr, ":%s UNSZLINE %d %s", sptr->name, matchit,
			   mask);
    else
	sendto_serv_butone(cptr, ":%s UNSZLINE %s", sptr->name, mask);
    return 0;
}

/*
 * the services aliases. * 
 *
 * NICKSERV     - /nickserv * CHANSERV  - /chanserv * OPERSERV  -
 * /operserv * MEMOSERV         - /memoserv * SERVICES  - /services *
 * IDENTIFY     - /identify * taz's code -mjs
 */

int m_chanserv(cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
    if ((acptr = find_person(ChanServ, NULL)))
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], ChanServ,
		   SERVICES_NAME, parv[1]);
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], ChanServ);
    return 0;
}

/*
 * m_nickserv
 */
int m_nickserv(cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
    if ((acptr = find_person(NickServ, NULL)))
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], NickServ,
		   SERVICES_NAME, parv[1]);
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], NickServ);
    return 0;
}

/*
 * m_memoserv
 */
int m_memoserv(cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
    if ((acptr = find_person(MemoServ, NULL)))
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], MemoServ,
		   SERVICES_NAME, parv[1]);
    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], MemoServ);
    return 0;
}

/*
 * m_operserv
 */
int m_operserv(cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
    if ((acptr = find_person(OperServ, NULL)))
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], OperServ,
#ifdef OPERSERV_OTHER_HOST
		   OPERSERV_OTHER_HOST,
#else
		   SERVICES_NAME,
#endif
		   parv[1]);

    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], OperServ);
    return 0;
}

/*
 * m_statserv
 */
int m_statserv(cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
aClient *acptr;

    if (check_registered_user(sptr))
	return 0;
    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
    if ((acptr = find_person(StatServ, NULL)))
	sendto_one(acptr, ":%s PRIVMSG %s@%s :%s", parv[0], StatServ,
#ifdef OPERSERV_OTHER_HOST
		   OPERSERV_OTHER_HOST,
#else
		   SERVICES_NAME,
#endif
		   parv[1]);

    else
	sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0], OperServ);
    return 0;
}

/*
 * m_services -- see df465+taz
 */
int m_services(cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
char *tmps;

    if (check_registered_user(sptr))
	return 0;

    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
    if ((strlen(parv[1]) >= 4) && (!myncmp(parv[1], "help", 4))) {
	sendto_one(sptr, ":services!service@%s NOTICE %s :For ChanServ "
		   "help use: /chanserv help", SERVICES_NAME, sptr->name);
	sendto_one(sptr, ":services!service@%s NOTICE %s :For NickServ "
		   "help use: /nickserv help", SERVICES_NAME, sptr->name);
	sendto_one(sptr, ":services!service@%s NOTICE %s :For MemoServ "
		   "help use: /memoserv help", SERVICES_NAME, sptr->name);
	return 0;
    }
    if ((tmps = (char *) strchr(parv[1], ' '))) {
	for (; *tmps == ' '; tmps++);	/* er.. before this for loop, the next
					 * * comparison would always compare
					 * '#' with ' '.. oops. - lucas */
	if (*tmps == '#')
	    return m_chanserv(cptr, sptr, parc, parv);
	else
	    return m_nickserv(cptr, sptr, parc, parv);
    }
    return m_nickserv(cptr, sptr, parc, parv);
}

/*
 * m_identify  df465+taz
 */
int m_identify(cptr, sptr, parc, parv)
     aClient *cptr, *sptr;
     int parc;
     char *parv[];
{
aClient *acptr;

    if (check_registered_user(sptr))
	return 0;

    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }
    if (*parv[1]) {
	if ((*parv[1] == '#') && ((char *) strchr(parv[1], ' '))) {
	    if ((acptr = find_person(ChanServ, NULL)))
		sendto_one(acptr, ":%s PRIVMSG %s@%s :IDENTIFY %s ",
			   parv[0], ChanServ, SERVICES_NAME, parv[1]);
	    else
		sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0],
			   ChanServ);
	} else {
	    if ((acptr = find_person(NickServ, NULL)))
		sendto_one(acptr, ":%s PRIVMSG %s@%s :IDENTIFY %s", parv[0],
			   NickServ, SERVICES_NAME, parv[1]);
	    else
		sendto_one(sptr, err_str(ERR_SERVICESDOWN), me.name, parv[0],
			   NickServ);
	}
    }
    return 0;
}

/*
 * s_svsnick - Pretty straight forward.  Mostly straight outta df
 * *  - Raistlin
 * * parv[0] = sender
 * * parv[1] = old nickname
 * * parv[2] = new nickname
 * * parv[3] = timestamp
 */
int m_svsnick(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (!IsULine(sptr) || parc < 4 || (strlen(parv[2]) > NICKLEN))
	return 0;
    /*
     * if we can't SVSNICK them to something because the nick is in use, KILL them 
     */
    acptr = find_person(parv[2], NULL);
    if (acptr != NULL) {
	/*
	 * send a kill out for the person instead --wd 
	 */
	sendto_serv_butone(cptr, ":%s KILL %s :%s (SVSNICK Collide)",
			   sptr->name, parv[1], sptr->name);
	/*
	 * now send back a kill for the nick they were presumably changed to 
	 */
	sendto_one(cptr, ":%s KILL %s :%s (SVSNICK Collide)", sptr->name,
		   parv[2], sptr->name);
	return 0;
    }
    if (!hunt_server(cptr, sptr, ":%s SVSNICK %s %s :%s", 1, parc, parv)
	!= HUNTED_ISME) {
	if ((acptr = find_person(parv[1], NULL)) != NULL) {
	    acptr->umode &= ~UMODE_r;
#ifdef ANTI_NICK_FLOOD
	    acptr->last_nick_change = atoi(parv[3]);
#endif
	    sendto_common_channels(acptr, ":%s NICK :%s", parv[1], parv[2]);
	    if (IsPerson(acptr))
		add_history(acptr, 1);
	    sendto_serv_butone(NULL, ":%s NICK %s :%i", parv[1], parv[2],
			       atoi(parv[3]));
	    if (acptr->name[0])
		del_from_client_hash_table(acptr->name, acptr);
	    strcpy(acptr->name, parv[2]);
	    add_to_client_hash_table(parv[2], acptr);
	}
    }
    return 0;
}

/*
 * send_svsmode_out - Once again, a df function that fits in pretty nicely
 * *  - Raistlin
 */
void send_svsmode_out(aClient *cptr, aClient *sptr, aClient *bsptr, int old)
{
    send_umode(NULL, sptr, old, SEND_UMODES, buf);
    sendto_serv_butone(cptr, ":%s SVSMODE %s :%s", bsptr->name, sptr->name,
		       buf);
}

/*
 * channel_svsmode:
 * * parv[0] sender  
 * * parv[1] channel 
 * * parv[2] modes   
 * * parv[3] nick    
 * * parv[4] nickts  
 * * currently, only a mode of -b is supported.
 * * services should use MODE for regular channel modes.
 * * 2/5/00 lucas
 * * preconditions: parc >= 3, sptr is ulined
 */
int channel_svsmode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aChannel *chptr;
    aClient *acptr = NULL;
    char *m, *nick = NULL;
    char change = '+';
    ts_val nickts = 0;
    int sendmsg = 1;

    if (!(chptr = find_channel(parv[1], NULL)))
	return 0;

    if (parc >= 4) {
	nick = parv[3];
	if (parc > 4)
	    nickts = atol(parv[4]);
    }

    if (nick) {
	acptr = find_person(nick, NULL);
	if (!acptr || (nickts && acptr->tsinfo != nickts))
	    return 0;
    }

    for (m = parv[2]; *m; m++)
	switch (*m) {
	    case '+':
	    case '-':
		change = *m;
		break;

	    case 'b':
		if (nick && MyClient(acptr) && change == '-') {
		    remove_matching_bans(chptr, acptr, &me);
		    sendmsg--;
		}
		break;

	    case 'e':
		if (nick && MyClient(acptr) && change == '-') {
		    remove_matching_banex(chptr, acptr, &me);
		    sendmsg--;
		}
		break;

	    case 'I':
		if (nick && MyClient(acptr) && change == '-') {
		    remove_matching_invites(chptr, acptr, &me);
		    sendmsg--;
		}
		break;

	    default:
		sendmsg++;
		break;
	}

    if (!sendmsg)
	return 0;

    if (nick)
	sendto_serv_butone(cptr, ":%s SVSMODE %s %s %s %ld", parv[0],
			   parv[1], parv[2], nick, acptr->tsinfo);
    else
	sendto_serv_butone(cptr, ":%s SVSMODE %s %s", parv[0], parv[1],
			   parv[2]);

    return 0;
}

/*
 * m_svsmode - df function integrated
 * *  - Raistlin
 * * -- Behaviour changed - Epi (11/30/99)
 * * parv[0] - sender
 * * parv[1] - nick  
 * * parv[2] - TS (or mode, depending on svs version)
 * * parv[3] - mode (or services id if old svs version)
 * * parv[4] - optional arguement (services id)
 */
int m_svsmode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int flag, *s, what, setflags;
    char *m, *modes, *optarg;
    aClient *acptr;
    ts_val ts = 0;

    if (!IsULine(sptr) || (parc < 3))
	return 0;

    if (parv[1][0] == '#')
	return channel_svsmode(cptr, sptr, parc, parv);

    if ((parc >= 4) && ((parv[3][0] == '+') || (parv[3][0] == '-'))) {
	ts = atol(parv[2]);
	modes = parv[3];
	optarg = (parc > 4) ? parv[4] : NULL;
    } else {
	modes = parv[2];
	optarg = (parc > 3) ? parv[3] : NULL;
    }

    if (!(acptr = find_person(parv[1], NULL)))
	return 0;

    if (ts && (ts != acptr->tsinfo))
	return 0;

    what = MODE_ADD;
    setflags = 0;

    for (s = user_modes; (flag = *s); s += 2)
	if (acptr->umode & flag)
	    setflags |= flag;

    for (m = modes; *m; m++)
	switch (*m) {
	    case '+':
		what = MODE_ADD;
		break;
	    case '-':
		what = MODE_DEL;
		break;
	    case ' ':
	    case 'q':		/* Never ever SVSMODE to +q on someone!!! NEVER 
				 * -TimeMr14C */
	    case '\n':
	    case '\r':
	    case '\t':
		break;
#ifdef USE_ZOMBIE
	    case 'z':
		if (MyClient(acptr)) {
		    if (what == MODE_ADD) {
			SetZombie(acptr);
			sendto_one(acptr,
				   ":%s NOTICE %s :You are now marked as zombie",
				   me.name, acptr->name);
		    } else if (what == MODE_DEL) {
			ClearZombie(acptr);
			sendto_one(acptr,
				   ":%s NOTICE %s :You are no longer marked as zombie",
				   me.name, acptr->name);
		    }
		    break;	/* Only break if our client -TimeMr14C */
		}
#endif
	    case 'd':
		if (optarg && isdigit(*optarg))
		    acptr->user->servicestamp = strtoul(optarg, NULL, 0);
		break;
	    default:
		for (s = user_modes; (flag = *s); s += 2) {
		    if (*m == (char) (*(s + 1))) {
			if (what == MODE_ADD)
			    acptr->umode |= flag;
			else
			    acptr->umode &= ~flag;
			break;
		    }
		}
		break;
	}

    if (optarg)
	sendto_serv_butone(cptr, ":%s SVSMODE %s %ld %s %s",
			   parv[0], parv[1], acptr->tsinfo, modes, optarg);
    else
	sendto_serv_butone(cptr, ":%s SVSMODE %s %ld %s", parv[0], parv[1],
			   acptr->tsinfo, modes);

    return 0;
}

/*
 * m_svsnoop - Remove all ops from a server
 * *  Once again pretty much straight outt df
 * *    - Raistlin
 * * parv[0] = sender prefix
 * * parv[1] = server 
 * * parv[2] = +/-
 */

int m_svsnoop(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aConfItem *aconf;

    if (!(IsULine(sptr) && parc > 2))
	return 0;

    if (hunt_server(cptr, sptr, ":%s SVSNOOP %s :%s", 1, parc, parv) ==
	HUNTED_ISME) {
	if (parv[2][0] == '+')
	    for (aconf = conf; aconf; aconf = aconf->next) {
		if (aconf->status & CONF_OPERATOR || aconf->status & CONF_LOCOP)
		    aconf->status = CONF_ILLEGAL;
	} else
	    rehash(&me, &me, 2);
    }
    return 0;
}
