/************************************************************************
 *   IRC - Internet Relay Chat, src/irctr.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * -TimeMr14C
 *
 * 06.01.2001
 *
 * This file contains every newer function used in the ircd
 * of the IRCTR network
 * Any laws and rules applicable to other source files of
 * the Bahamut ircd are applicable to this file, too.
 *
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <sys/stat.h>
#include <utmp.h>
#include <fcntl.h>
#include <netinet/in.h>
#include "nameser.h"
#include "resolv.h"
#include "h.h"
#include "inet.h"
#include "md5.h"

#ifdef FLUD
#include "blalloc.h"
#endif /* 
        * FLUD 
        */

#if defined( HAVE_STRING_H)
#include <string.h>
#else
#include <strings.h>
#endif

static char nickbuf[BUFSIZE], buf[BUFSIZE];
static char modebuf[REALMODEBUFLEN], parabuf[REALMODEBUFLEN];

extern Link *find_channel_link(Link *, aChannel *);
extern char *smalldate(time_t);	/* defined in s_misc.c */

static aConfItem *temporary_excludes = (aConfItem *) NULL;
extern int rehashed;
extern int zline_in_progress;

/*
 * -TimeMr14C: This is the old calcmask function
 * * char *calcmask(char *myip, char *yourip) {
 * *
 * *  char sonucumuz[HOSTLEN], kript[HOSTLEN];
 * *  char *p = NULL;
 * *  extern char *crypt();
 * *
 * *  strcpy(sonucumuz, strtoken(&p, myip, "."));
 * *  strcat(sonucumuz, ".");   
 * *  strcat(sonucumuz, strtoken(&p, NULL, "."));
 * *  strcat(sonucumuz, ".");
 * *  strcpy(kript, strtoken(&p, NULL, "."));
 * *  strcat(kript, ".");
 * *  strcat(kript, strtoken(&p, NULL, "."));
 * *  strcat(sonucumuz, (char *)crypt(kript,"TR"));
 * *  strcat(sonucumuz, "q.");
 * *  strcat(sonucumuz, NETWORKNAME);
 * *
 * *  strcpy(yourip, sonucumuz);
 * *  return yourip;
 * * }
 */

/*
 * -Talesin: This is the new Calcmask Function 
 */

/*
 * char *calcmask(char *myip, char *yourip) {
 * *
 * *  char res[HOSTLEN], out[HOSTLEN];
 * *  char *na;
 * *  char myseed[2];
 * *  char *p = NULL;
 * *  extern char *crypt();
 * *  
 * *  na = NETWORKNAME;
 * * 
 * *  strcpy(res, strtoken(&p, myip, "."));
 * *  strcat(res, ".");
 * *  strcat(res, strtoken(&p, NULL, "."));
 * *  strcat(res, ".");
 * *  strncat(res, NETWORKNAME, strlen(NETWORKNAME)-1);
 * *
 * *  strcpy(out, strtoken(&p, NULL, "."));
 * *  strcat(out, ".");
 * *  strcat(out, strtoken(&p, NULL, "."));
 * *
 * *  myseed[0] = na[strlen(NETWORKNAME)-1];
 * *  myseed[1] = 0;
 * *  strcat(myseed,"-"); 
 * *
 * *  strcat(res, (char *)crypt(out,myseed));
 * *  strcat(res, "X.TR");
 * *     
 * *  strcpy(yourip, res);
 * *  return yourip;
 * * }
 */

/*
 * -TimeMr14C: This is a better calcmask function 
 */

char *calcmask(char *myip, char *yourip)
{

    int cnt;
    unsigned char arr[16];
    char res[HOSTLEN], out[HOSTLEN];

    md5_buffer(myip, HOSTLEN, arr);

    sprintf(res, "%s.%01X", NETWORKNAME, arr[0]);

    for (cnt = 1; cnt < 15; ++cnt) {
	sprintf(out, "%01X", arr[cnt]);
	strcat(res, out);
    }

    strcat(res, ".TR");
    strcpy(yourip, res);
    return yourip;
}

/*
 * m_operdo 
 * * -TimeMr14C
 * * operdo mode/kick/topic
 * * parv[0] = sender
 * * parv[1] = opcommand
 * * parv[2] = channel
 * * parv[3] = params
 */

static int get_opcommand(char *param)
{
    int i = 0;

    if (!strcasecmp(param, "MODE")) {
	i = 1;
    } else if (!strcasecmp(param, "TOPIC")) {
	i = 2;
    } else if (!strcasecmp(param, "KICK")) {
	i = 3;
    } else if (!strcasecmp(param, "JOIN")) {
	i = 4;
    } else if (!strcasecmp(param, "HELP")) {
	i = 5;
    }
    return i;

}

int m_operdo(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    aChannel *chptr;
    int o_param;
    int sendtsm;
    time_t ts;

    if (!IsAnOper(cptr) || parc < 4)
	return 0;

    o_param = get_opcommand(parv[1]);
    if (!check_channelname(sptr, (unsigned char *) parv[2])) {
	sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], parv[2]);
	return 0;
    }
    chptr = find_channel(parv[2], NULL);
    if ((chptr == NullChn) && (o_param < 4)) {
	sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], parv[2]);
	return 0;
    }
    if ((o_param < 4) && IsChanAnon(chptr)) {
	sendto_one(sptr,
		   ":%s NOTICE %s :OPERDO Cannot be used for anonymous channels",
		   me.name, parv[0]);
	send_globops("from %s: %s tried to use OPERDO on anonymous channel %s",
		     me.name, parv[0], chptr->chname);
	return 0;
    }

    switch (o_param) {

	case 1:{
		sendtsm =
		    set_mode(cptr, sptr, chptr, 3, parc - 3, parv + 3, modebuf,
			     parabuf);
		if (strlen(modebuf) > (size_t) 1) {

		    sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s",
					   parv[0], chptr->chname, modebuf,
					   parabuf);
		    sendto_match_servs(chptr, cptr, ":%s MODE %s %s %s",
				       parv[0], chptr->chname, modebuf,
				       parabuf);
		    if (MyClient(sptr)) {
			sendto_serv_butone(NULL,
					   ":%s GLOBOPS :%s used OPERDO MODE on %s %s%s%s",
					   me.name, sptr->name, chptr->chname,
					   modebuf, (*parabuf != 0 ? " " : ""),
					   parabuf);
			send_globops
			    ("from %s: %s used OPERDO MODE on %s %s%s%s",
			     me.name, sptr->name, chptr->chname, modebuf,
			     (*parabuf != 0 ? " " : ""), parabuf);
		    }
		}
		break;
	    }

	case 2:{
		ts = timeofday;
		strncpyzt(chptr->topic, parv[3], TOPICLEN + 1);
		strcpy(chptr->topic_nick, sptr->name);
		chptr->topic_time = ts;

		sendto_match_servs(chptr, cptr, ":%s TOPIC %s %s %lu :%s",
				   parv[0], chptr->chname,
				   chptr->topic_nick, chptr->topic_time,
				   chptr->topic);
		sendto_channel_butserv(chptr, sptr, ":%s TOPIC %s :%s", parv[0],
				       chptr->chname, chptr->topic);
		if (MyClient(sptr)) {
		    sendto_serv_butone(NULL,
				       ":%s GLOBOPS :%s used OPERDO TOPIC on %s",
				       me.name, sptr->name, chptr->chname);
		    send_globops("from %s: %s used OPERDO TOPIC on %s",
				 me.name, sptr->name, chptr->chname);
		}
		break;
	    }
	case 3:{

    aClient *who;
    aChannel *chptr;
    int chasing = 0;
    int user_count;
    char *comment, *name, *p = NULL, *user, *p2 = NULL;

		if (parc < 4 || *parv[2] == '\0') {
		    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name,
			       parv[0], "KICK");
		    return 0;
		}

		if (IsServer(sptr) && !IsULine(sptr))
		    sendto_ops("KICK from %s for %s %s", parv[0], parv[2],
			       parv[3]);

		comment = (BadPtr(parv[4])) ? parv[0] : parv[4];

		if (strlen(comment) > (size_t) TOPICLEN)
		    comment[TOPICLEN] = '\0';

		*nickbuf = *buf = '\0';
		name = strtoken(&p, parv[2], ",");
		while (name) {
		    chptr = get_channel(sptr, name, !CREATE);
		    if (!chptr) {
			sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name,
				   parv[0], name);
			name = strtoken(&p, (char *) NULL, ",");
			continue;
		    }

		    user = strtoken(&p2, parv[3], ",");
		    user_count = 4;
		    while (user && user_count) {
			user_count--;
			if (!(who = find_chasing(sptr, user, &chasing))) {
			    user = strtoken(&p2, (char *) NULL, ",");
			    continue;
			}

			if (IsMember(who, chptr)) {
			    sendto_channel_butserv(chptr, sptr,
						   ":%s KICK %s %s :%s",
						   parv[0], name, who->name,
						   comment);
			    sendto_match_servs(chptr, cptr,
					       ":%s KICK %s %s :%s", parv[0],
					       name, who->name, comment);
			    if (MyClient(sptr)) {
				sendto_serv_butone(NULL,
						   ":%s GLOBOPS :%s used OPERDO KICK on %s for %s",
						   me.name, sptr->name,
						   chptr->chname, who->name);
				send_globops
				    ("from %s: %s used OPERDO KICK on %s for %s",
				     me.name, sptr->name, chptr->chname,
				     who->name);
			    }
			    remove_user_from_channel(who, chptr);
			} else
			    sendto_one(sptr, err_str(ERR_USERNOTINCHANNEL),
				       me.name, parv[0], user, name);
			user = strtoken(&p2, (char *) NULL, ",");
		    }
		    name = strtoken(&p, (char *) NULL, ",");
		}
		break;
	    }

	case 4:{
		parv[1] = parv[2];
		(void) m_join(cptr, sptr, 2, parv);
		chptr = find_channel(parv[1], NULL);
       		if (chptr == NullChn)
        	    return 0;
		sendtsm =
		    set_mode(&me, &me, chptr, 3, parc - 3, parv + 3, modebuf,
			     parabuf);
		if (strlen(modebuf) > (size_t) 1) {

		    sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s",
					   me.name, chptr->chname, modebuf,
					   parabuf);
		    sendto_match_servs(chptr, cptr, ":%s MODE %s %s %s",
				       me.name, chptr->chname, modebuf,
				       parabuf);
		    if (MyClient(sptr)) {
			sendto_serv_butone(NULL,
					   ":%s GLOBOPS :%s used OPERDO JOIN on %s %s%s%s",
					   me.name, sptr->name, chptr->chname,
					   modebuf, (*parabuf != 0 ? " " : ""),
					   parabuf);
			send_globops
			    ("from %s: %s used OPERDO JOIN on %s %s%s%s",
			     me.name, sptr->name, chptr->chname, modebuf,
			     (*parabuf != 0 ? " " : ""), parabuf);
		    }
		}
		break;
	    }

	case 5:{
		sendto_one(sptr, ":%s NOTICE %s :Available commands are:",
			   me.name, parv[0]);
		sendto_one(sptr, ":%s NOTICE %s : ", me.name, parv[0]);
		sendto_one(sptr, ":%s NOTICE %s :\002MODE\002 #channel params",
			   me.name, parv[0]);
		sendto_one(sptr, ":%s NOTICE %s :\002TOPIC\002 #channel :text",
			   me.name, parv[0]);
		sendto_one(sptr,
			   ":%s NOTICE %s :\002KICK\002 #channel user :reason",
			   me.name, parv[0]);
		sendto_one(sptr, ":%s NOTICE %s :\002JOIN\002 #channel modes",
			   me.name, parv[0]);
		break;
	    }

	case 0:{
		sendto_one(sptr, ":%s NOTICE %s :No such operdo command %s",
			   me.name, parv[0], parv[1]);
		sendto_one(sptr, ":%s NOTICE %s :Try \002HELP\002",
			   me.name, parv[0]);
		break;
	    }
	default:
	    break;
    }
    return 0;

}

/*
 * s_svsjoin 
 * *  - TimeMr14C
 * * parv[0] = sender
 * * parv[1] = nickname
 * * parv[2] = channel
 * * parv[3] = timestamp
 */

int m_svsjoin(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    aClient *acptr;

    if (!IsULine(sptr) && !IsServer(sptr))
	return 0;

    if (parc < 3 || !(acptr = find_person(parv[1], NULL)))
	return 0;

    if (MyClient(acptr)) {

	parv[0] = parv[1];
	parv[1] = parv[2];
	(void) m_join(acptr, acptr, 2, parv);

    } else {

	sendto_serv_butone(sptr, "SVSJOIN %s %s", parv[1], parv[2]);

    }
    return 0;

}

/*
 * m_nethtm
 *
 * -TimeMr14c : Tm serverlar ayni anda htm e gecsinler diye
 */

int m_nethtm(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    aClient *acptr;

    if (!parv[1])
	return 0;
    if (!IsServer(sptr)) {
	if ((acptr = find_person(parv[0], NULL)) != NULL) {

	    if (!IsAnOper(acptr))
		return 0;

	    (void) m_htm(acptr, acptr, parc, parv);
	    if (parv[2])
		sendto_serv_butone(sptr, "NETHTM %s %s :%s", acptr->name,
				   parv[1], parv[2]);
	    else
		sendto_serv_butone(sptr, "NETHTM %s %s", acptr->name, parv[1]);
	}
    } else {
	if ((acptr = find_person(parv[1], NULL)) != NULL) {
	    parv[0] = parv[1];
	    parv[1] = parv[2];

	    if (parv[3]) {
		parv[2] = parv[3];
		(void) m_htm(acptr, acptr, parc, parv);
		sendto_serv_butone(sptr, "NETHTM %s %s :%s", acptr->name,
				   parv[1], parv[2]);
	    } else {
		(void) m_htm(acptr, acptr, parc, parv);
		sendto_serv_butone(sptr, "NETHTM %s %s", acptr->name, parv[1]);
	    }
	}
    }
    return 0;
}

/*
 * m_gline as a script to place local klines and spread to linked servers
 * -TimeMr14C
 */

int m_gline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    aClient *acptr;
    int duration;
    char argv[USERLEN + HOSTLEN + 3];
    char reason[TOPICLEN + 1];

    if ((acptr = find_person(parv[0], NULL)) != NULL) {
	if (!IsPrivileged(cptr)) {
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
    } else {
	return 0;
    }

    if (parc > 3) {
	duration = atoi(parv[1]);
	if (duration <= 0) {
	    sendto_one(acptr,
		       ":%s NOTICE %s :Please try /gline <duration> <user@host> <:reason>",
		       me.name, acptr->name);
	    return 0;
	}
	strcpy(argv, parv[2]);
	strcpy(reason, parv[3]);
	sendto_ops("Passing along GLINE by %s for %s", acptr->name, argv);
	(void) m_kline(acptr, acptr, parc, parv);
	sendto_serv_butone(sptr, ":%s GLINE %d %s :GLINED (%s)",
			   acptr->name, duration, argv, reason);
    } else {
	sendto_one(acptr,
		   ":%s NOTICE %s :Please try /gline <duration> <user@host> <:reason>",
		   me.name, acptr->name);
    }

    return 0;
}

/*
 * If there is a G:Line then also an UNGLINE ;-P
 * -TimeMr14C
 *
 */

int m_ungline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    aClient *acptr;
    char argv[USERLEN + HOSTLEN + 3];

    if ((acptr = find_person(parv[0], NULL)) != NULL) {
	if (!IsPrivileged(cptr)) {
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
    } else {
	return 0;
    }

    if (parc > 1) {
	strcpy(argv, parv[1]);
	sendto_ops("Passing along UNGLINE by %s for %s", acptr->name, argv);
	(void) m_unkline(acptr, acptr, parc, parv);
	sendto_serv_butone(sptr, ":%s UNGLINE %s", acptr->name, argv);
    } else {
	sendto_one(acptr, ":%s NOTICE %s :Please try /ungline <user@host>",
		   me.name, acptr->name);
    }

    return 0;

}

int m_jupiter(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aConfItem *aconf;

#ifdef NO_OPER_JUPITER
    if (!(IsServer(sptr) || IsUline(sptr)))
	return 0;
#else
    if (!(IsServer(sptr) || IsULine(sptr) || IsAnOper(sptr)))
	return 0;
#endif

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "JUPITER");
	return 0;
    }

    parv[1] = collapse(parv[1]);

    if (!(aconf = find_conf_name(parv[1], CONF_JUPECHAN))) {
	aconf = make_conf();

	aconf->status = (CONF_JUPECHAN | CONF_SJUPECHAN);
	DupString(aconf->name, parv[1]);
	DupString(aconf->passwd,
		  (parv[2] !=
		   NULL ? parv[2] : "Channel has been jupitered due to abuse"));
	aconf->next = conf;
	conf = aconf;
    }
    if (MyClient(sptr))
	sendto_gnotice
	    ("from %s: %s has JUPITERED following channel mask: %s",
	     me.name, sptr->name, parv[1], aconf->passwd);
    sendto_serv_butone(cptr, ":%s JUPITER %s :%s", sptr->name, parv[1],
		       aconf->passwd);
    return 0;
}

int m_unjupiter(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int matchit = 0;
    char *mask;
    aConfItem *aconf, *ac2, *ac3;

#ifdef NO_OPER_JUPITER
    if (!(IsServer(sptr) || IsUline(sptr)))
	return 0;
#else
    if (!(IsServer(sptr) || IsULine(sptr) || IsAnOper(sptr)))
	return 0;
#endif

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "UNJUPITER");
	return 0;
    }

    if (parc == 3) {
	matchit = atoi(parv[1]);
	mask = parv[2];
    } else
	mask = parv[1];

    ac2 = ac3 = aconf = conf;
    while (aconf) {
	if ((aconf->status & CONF_JUPECHAN)
	    && (aconf->status & CONF_SJUPECHAN) &&
	    aconf->name && (matchit ? !match(mask, aconf->name)
			    : !mycmp(mask, aconf->name))) {
	    if (conf == aconf)
		ac2 = ac3 = conf = aconf->next;
	    else
		ac2 = ac3->next = aconf->next;
	    MyFree(aconf->passwd);
	    MyFree(aconf->name);
	    MyFree(aconf);
	    aconf = ac2;
	} else {
	    ac3 = aconf;
	    aconf = aconf->next;
	}
    }
    if (MyClient(sptr))
	sendto_gnotice
	    ("from %s: %s has UNJUPED following channel mask: %s", me.name,
	     sptr->name, parv[1], mask);

    if (parc == 3)
	sendto_serv_butone(cptr, ":%s UNJUPITER %d :%s", sptr->name, matchit,
			   mask);
    else
	sendto_serv_butone(cptr, ":%s UNJUPITER %s", sptr->name, mask);
    return 0;
}

typedef struct {
    char *lev_chname;
    struct Channel *levptr;
    char *topic;
} LChan;

static LChan lchans[12] = {
    {N_NOTICE, NULL, T_NOTICE_LEV},
    {N_CONNECT, NULL, T_CCONN_LEV},
    {N_REJECT, NULL, T_REJ_LEV},
    {N_KILL, NULL, T_SKILL_LEV},
    {N_STAT, NULL, T_SPY_LEV},
    {N_DEBUG, NULL, T_DEBUG_LEV},
    {N_FLOOD, NULL, T_FLOOD_LEV},
    {N_SPAMBOT, NULL, T_SPAM_LEV},
    {N_DCCSEND, NULL, T_DCCSEND_LEV},
    {N_SERVICE, NULL, T_SERVICE_LEV},
    {N_ULINE, NULL, NULL},
    {NULL, NULL, NULL}
};

void init_levs(aClient *ghost)
{

    int i;
    LChan *alptr;

    SetLocOp(ghost);
    SetAnon(ghost);

    i = 0;

    alptr = lchans + i;
    alptr->levptr = get_channel(ghost, alptr->lev_chname, CREATE);
    add_user_to_channel(alptr->levptr, ghost, CHFL_CHANOP);
    alptr->levptr->mode.mode |= MODE_TOPICLIMIT;
    alptr->levptr->mode.mode |= MODE_NOPRIVMSGS;
    alptr->levptr->mode.mode |= MODE_ANONYMOUS;
    strcpy(alptr->levptr->topic, alptr->topic);
    i++;

    alptr = lchans + i;
    alptr->levptr = get_channel(ghost, alptr->lev_chname, CREATE);
    add_user_to_channel(alptr->levptr, ghost, CHFL_CHANOP);
    alptr->levptr->mode.mode |= MODE_TOPICLIMIT;
    alptr->levptr->mode.mode |= MODE_NOPRIVMSGS;
    alptr->levptr->mode.mode |= MODE_OPERONLY;
    alptr->levptr->mode.mode |= MODE_ANONYMOUS;
    strcpy(alptr->levptr->topic, alptr->topic);
    i++;

    alptr = lchans + i;
    alptr->levptr = get_channel(ghost, alptr->lev_chname, CREATE);
    add_user_to_channel(alptr->levptr, ghost, CHFL_CHANOP);
    alptr->levptr->mode.mode |= MODE_TOPICLIMIT;
    alptr->levptr->mode.mode |= MODE_NOPRIVMSGS;
    alptr->levptr->mode.mode |= MODE_OPERONLY;
    alptr->levptr->mode.mode |= MODE_ANONYMOUS;
    strcpy(alptr->levptr->topic, alptr->topic);
    i++;

    alptr = lchans + i;
    alptr->levptr = get_channel(ghost, alptr->lev_chname, CREATE);
    add_user_to_channel(alptr->levptr, ghost, CHFL_CHANOP);
    alptr->levptr->mode.mode |= MODE_TOPICLIMIT;
    alptr->levptr->mode.mode |= MODE_NOPRIVMSGS;
    alptr->levptr->mode.mode |= MODE_OPERONLY;
    alptr->levptr->mode.mode |= MODE_ANONYMOUS;
    strcpy(alptr->levptr->topic, alptr->topic);
    i++;

    alptr = lchans + i;
    alptr->levptr = get_channel(ghost, alptr->lev_chname, CREATE);
    add_user_to_channel(alptr->levptr, ghost, CHFL_CHANOP);
    alptr->levptr->mode.mode |= MODE_TOPICLIMIT;
    alptr->levptr->mode.mode |= MODE_NOPRIVMSGS;
    alptr->levptr->mode.mode |= MODE_OPERONLY;
    alptr->levptr->mode.mode |= MODE_ANONYMOUS;
    strcpy(alptr->levptr->topic, alptr->topic);
    i++;

    alptr = lchans + i;
    alptr->levptr = get_channel(ghost, alptr->lev_chname, CREATE);
    add_user_to_channel(alptr->levptr, ghost, CHFL_CHANOP);
    alptr->levptr->mode.mode |= MODE_TOPICLIMIT;
    alptr->levptr->mode.mode |= MODE_NOPRIVMSGS;
    alptr->levptr->mode.mode |= MODE_OPERONLY;
    alptr->levptr->mode.mode |= MODE_ANONYMOUS;
    strcpy(alptr->levptr->topic, alptr->topic);
    i++;

    alptr = lchans + i;
    alptr->levptr = get_channel(ghost, alptr->lev_chname, CREATE);
    add_user_to_channel(alptr->levptr, ghost, CHFL_CHANOP);
    alptr->levptr->mode.mode |= MODE_TOPICLIMIT;
    alptr->levptr->mode.mode |= MODE_NOPRIVMSGS;
    alptr->levptr->mode.mode |= MODE_OPERONLY;
    alptr->levptr->mode.mode |= MODE_ANONYMOUS;
    strcpy(alptr->levptr->topic, alptr->topic);
    i++;

    alptr = lchans + i;
    alptr->levptr = get_channel(ghost, alptr->lev_chname, CREATE);
    add_user_to_channel(alptr->levptr, ghost, CHFL_CHANOP);
    alptr->levptr->mode.mode |= MODE_TOPICLIMIT;
    alptr->levptr->mode.mode |= MODE_NOPRIVMSGS;
    alptr->levptr->mode.mode |= MODE_OPERONLY;
    alptr->levptr->mode.mode |= MODE_ANONYMOUS;
    strcpy(alptr->levptr->topic, alptr->topic);
    i++;

    alptr = lchans + i;
    alptr->levptr = get_channel(ghost, alptr->lev_chname, CREATE);
    add_user_to_channel(alptr->levptr, ghost, CHFL_CHANOP);
    alptr->levptr->mode.mode |= MODE_TOPICLIMIT;
    alptr->levptr->mode.mode |= MODE_NOPRIVMSGS;
    alptr->levptr->mode.mode |= MODE_OPERONLY;
    alptr->levptr->mode.mode |= MODE_ANONYMOUS;
    strcpy(alptr->levptr->topic, alptr->topic);
    i++;

    alptr = lchans + i;
    alptr->levptr = get_channel(ghost, alptr->lev_chname, CREATE);
    add_user_to_channel(alptr->levptr, ghost, CHFL_CHANOP);
    alptr->levptr->mode.mode |= MODE_TOPICLIMIT;
    alptr->levptr->mode.mode |= MODE_NOPRIVMSGS;
    alptr->levptr->mode.mode |= MODE_OPERONLY;
    alptr->levptr->mode.mode |= MODE_ANONYMOUS;
    strcpy(alptr->levptr->topic, alptr->topic);
    i++;

    alptr = lchans + i;
    alptr->levptr = get_channel(ghost, alptr->lev_chname, CREATE);
    add_user_to_channel(alptr->levptr, ghost, CHFL_CHANOP);
    alptr->levptr->mode.mode |= MODE_TOPICLIMIT;
    alptr->levptr->mode.mode |= MODE_NOPRIVMSGS;
    alptr->levptr->mode.mode |= MODE_OPERONLY;
    alptr->levptr->mode.mode |= MODE_ANONYMOUS;
    alptr->levptr->mode.mode |= MODE_SECRET;
    alptr->levptr->mode.mode |= MODE_PRIVATE;
}

void sendto_realops_lev(int lev, char *pattern, ...)
{

    aChannel *chptr = NullChn;
    LChan *alptr;
    char nbuf[1024];
    va_list vl;

    alptr = lchans + (lev);

    alptr->levptr = find_channel(alptr->lev_chname, NULL);

    if ((chptr = alptr->levptr)) {
	va_start(vl, pattern);
	(void) vsprintf(nbuf, pattern, vl);
	va_end(vl);
	sendto_channel_butserv(chptr, &me, ":%s NOTICE %s :%s", me.name,
			       chptr->chname, nbuf);
    }
    return;
}

void sendto_ops_lev(int lev, char *pattern, ...)
{

    aChannel *chptr = NullChn;
    LChan *alptr;
    char nbuf[1024];
    va_list vl;

    alptr = lchans + (lev);

    alptr->levptr = find_channel(alptr->lev_chname, NULL);

    if ((chptr = alptr->levptr)) {
	va_start(vl, pattern);
	(void) vsprintf(nbuf, pattern, vl);
	va_end(vl);
	sendto_channel_butserv(chptr, &me, ":%s NOTICE %s :%s", me.name,
			       chptr->chname, nbuf);
    }
    return;
}

void do_rehash_excludes()
{
aConfItem *exclude_list_ptr;
aConfItem *last_list_ptr;
aConfItem *tmp_list_ptr;

    if (temporary_excludes) {
	exclude_list_ptr = last_list_ptr = temporary_excludes;

	while (exclude_list_ptr) {
	    if (exclude_list_ptr->hold <= NOW ||
		exclude_list_ptr->hold == 0xFFFFFFFF) {
		if (temporary_excludes == exclude_list_ptr)
		    temporary_excludes = last_list_ptr = tmp_list_ptr =
			exclude_list_ptr->next;
		else
		    tmp_list_ptr = last_list_ptr->next = exclude_list_ptr->next;

		MyFree(exclude_list_ptr->host);
		MyFree(exclude_list_ptr->name);
		MyFree(exclude_list_ptr->passwd);
		MyFree(exclude_list_ptr);
		exclude_list_ptr = tmp_list_ptr;
	    } else {
		last_list_ptr = exclude_list_ptr;
		exclude_list_ptr = exclude_list_ptr->next;
	    }
	}
    }
}

aConfItem *find_exclude(aClient *cptr)
{
    char *host, *name;
    aConfItem *ret;

    if (!cptr->user)
	return 0;

    host = cptr->sockhost;
    name = cptr->user->username;

    if (strlen(host) > (size_t) HOSTLEN ||
	(name ? strlen(name) : 0) > (size_t) HOSTLEN)
	return (0);

    ret = find_is_excluded(host, name, cptr);
    if (ret != NULL)
	return ret;

    host = cptr->hostip;
    return (find_is_excluded(host, name, cptr));
}

aConfItem *find_is_excluded(char *host, char *name, aClient *cptr)
{
    aConfItem *exclude_list_ptr;	/* used for the link list only */
    aConfItem *last_list_ptr;
    aConfItem *tmp_list_ptr;

    if (temporary_excludes) {
	exclude_list_ptr = last_list_ptr = temporary_excludes;

	while (exclude_list_ptr) {
	    if (exclude_list_ptr->hold <= NOW &&
		exclude_list_ptr->hold != 0xFFFFFFFF) {
		if (temporary_excludes == exclude_list_ptr) {
		    temporary_excludes = last_list_ptr = tmp_list_ptr =
			exclude_list_ptr->next;
		} else {
		    tmp_list_ptr = last_list_ptr->next = exclude_list_ptr->next;
		}

		MyFree(exclude_list_ptr->host);
		MyFree(exclude_list_ptr->name);
		MyFree(exclude_list_ptr->passwd);
		MyFree(exclude_list_ptr);
		exclude_list_ptr = tmp_list_ptr;
	    } else {
		if ((exclude_list_ptr->name &&
		     (!name || !match(exclude_list_ptr->name, name)))
		    && (exclude_list_ptr->host &&
			(!host || !match(exclude_list_ptr->host, host)))) {
		    return (exclude_list_ptr);
		}
#ifdef USE_FAKEHOST
		if ((exclude_list_ptr->name &&
		     (!name || !match(exclude_list_ptr->name, name)))
		    && (exclude_list_ptr->host && cptr &&
			!match(exclude_list_ptr->host, cptr->user->fakehost))) {
		    return (exclude_list_ptr);
		}
#endif
		if ((exclude_list_ptr->name &&
		     (!name || !match(exclude_list_ptr->name, name)))
		    && (exclude_list_ptr->host && cptr &&
			!match_ipmask(exclude_list_ptr->host, cptr))) {
		    return (exclude_list_ptr);
		}
		last_list_ptr = exclude_list_ptr;
		exclude_list_ptr = exclude_list_ptr->next;
	    }
	}
    }

    return ((aConfItem *) NULL);
}

void add_temp_exclude(aConfItem *aconf)
{
    aconf->next = temporary_excludes;
    temporary_excludes = aconf;
}

void report_temp_excludes(aClient *sptr)
{
    aConfItem *exclude_list_ptr;
    aConfItem *last_list_ptr;
    aConfItem *tmp_list_ptr;
    char *host, *name, *reason, type;
    int len;

    if (temporary_excludes) {
	exclude_list_ptr = last_list_ptr = temporary_excludes;

	while (exclude_list_ptr) {
	    if (exclude_list_ptr->hold <= NOW &&
		exclude_list_ptr->hold != 0xFFFFFFFF) {
		if (temporary_excludes == exclude_list_ptr) {
		    temporary_excludes = last_list_ptr = tmp_list_ptr =
			exclude_list_ptr->next;
		} else {
		    tmp_list_ptr = last_list_ptr->next = exclude_list_ptr->next;
		}

		MyFree(exclude_list_ptr->host);
		MyFree(exclude_list_ptr->name);
		MyFree(exclude_list_ptr->passwd);
		MyFree(exclude_list_ptr);
		exclude_list_ptr = tmp_list_ptr;
	    } else {
		if (exclude_list_ptr->host)
		    host = exclude_list_ptr->host;
		else
		    host = "*";

		if (exclude_list_ptr->name)
		    name = exclude_list_ptr->name;
		else
		    name = "*";

		if (exclude_list_ptr->passwd)
		    reason = exclude_list_ptr->passwd;
		else
		    reason = "No Reason";

		if (exclude_list_ptr->hold == 0xFFFFFFFF) {
		    len = 0;
		} else {
		    len = (exclude_list_ptr->hold - NOW) / 60;
		}

		type = 'E';

		sendto_one(sptr, rpl_str(RPL_STATSKLINE), me.name,
			   sptr->name, type, host, name, len, reason);

		last_list_ptr = exclude_list_ptr;
		exclude_list_ptr = exclude_list_ptr->next;
	    }
	}
    }
}

int remove_temp_exclude(char *host, char *user)
{
    aConfItem *exclude_list_ptr;
    aConfItem *last_list_ptr;
    aConfItem *tmp_list_ptr;
    int num_removed = 0;

    if (temporary_excludes) {
	exclude_list_ptr = last_list_ptr = temporary_excludes;

	while (exclude_list_ptr) {
	    if (!mycmp(exclude_list_ptr->host, host)
		&& !mycmp(exclude_list_ptr->name, user)) {
		if (temporary_excludes == exclude_list_ptr)
		    temporary_excludes = last_list_ptr = tmp_list_ptr =
			exclude_list_ptr->next;
		else
		    tmp_list_ptr = last_list_ptr->next = exclude_list_ptr->next;

		MyFree(exclude_list_ptr->host);
		MyFree(exclude_list_ptr->name);
		MyFree(exclude_list_ptr->passwd);
		MyFree(exclude_list_ptr);
		exclude_list_ptr = tmp_list_ptr;
		num_removed++;
	    } else {
		last_list_ptr = exclude_list_ptr;
		exclude_list_ptr = exclude_list_ptr->next;
	    }
	}
    }
    return num_removed;
}

/*
 * m_akill -
 * * Parse AKILL command
 * * parv[1]=host, parv[2]=user, parv[3]=length, parv[4]=akiller, parv[5]=time set, parv[6]=reason
 */
/*
 * m_exclude -
 * * Parse EXCLUDE command
 * * parv[1]=host, parv[2]=user, parv[3]=length, parv[4]=excluder, parv[5]=time set, parv[6]=reason
 */
int m_exclude(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    aConfItem *aconf, *bconf;
    char *user, *host, *reason, *excluder, buffer[1024], *current_date;
    time_t length = 0, timeset = 0;

    if ((!IsServer(sptr) && !IsServiceModeA(sptr)) || (parc < 6))
	return 0;

    if (!IsULine(sptr)) {
	sendto_serv_butone(&me,
			   ":%s GLOBOPS :Non-ULined server %s trying to Exclude Autokill!",
			   me.name, sptr->name);
	send_globops
	    ("From %s: Non-ULined server %s trying to Exclude Autokill!",
	     me.name, sptr->name);
	return 0;
    }

    host = parv[1];
    user = parv[2];
    excluder = parv[4];
    length = atoi(parv[3]);
    timeset = atoi(parv[5]);
    reason = (parv[6] ? parv[6] : "<no reason>");

    current_date = smalldate((time_t) timeset);
    if (strlen(reason) > 250)
	reason[251] = 0;

    bconf = find_is_excluded(host, user, (aClient *) NULL);
    if (bconf != NULL) {
	sendto_flag_serv_butone(cptr, CAPAB_EXCLUDE, 0,
				":%s EXCLUDE %s %s %d %s %d :%s",
				sptr->name, host, user, length, excluder,
				timeset, reason);
	return 0;
    }
    aconf = make_conf();
    aconf->status = CONF_EXCLUDE;
    DupString(aconf->host, host);
    DupString(aconf->name, user);

    ircsprintf(buffer, "%s (%s)", reason, current_date);

    DupString(aconf->passwd, buffer);

    if (length > 86400 || !length)
	aconf->hold = 0xFFFFFFFF;
    else
	aconf->hold = timeset + length;

    add_temp_exclude(aconf);
    rehashed = YES;
    zline_in_progress = NO;
    sendto_flag_serv_butone(cptr, CAPAB_EXCLUDE, 0,
			    ":%s EXCLUDE %s %s %d %s %d :%s", sptr->name,
			    host, user, length, excluder, timeset, reason);
    return 0;
}

int m_rexclude(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (!IsServer(sptr) && !IsServiceModeA(sptr))
	return 0;
    if (parc < 3) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		   "REXCLUDE");
	return 0;
    }
    if (!IsULine(sptr)) {
	sendto_serv_butone(&me,
			   ":%s GLOBOPS :Non-ULined server %s trying to Remove Exclusion!",
			   me.name, sptr->name);
	send_globops
	    ("From %s: Non-ULined server %s trying to Remove Exclusion!",
	     me.name, sptr->name);
	return 0;
    }
    remove_temp_exclude(parv[1], parv[2]);
    sendto_flag_serv_butone(cptr, CAPAB_EXCLUDE, 0, ":%s REXCLUDE %s %s",
			    sptr->name, parv[1], parv[2]);
    return 0;
}

/*
 * Match address by #IP bitmask (10.11.12.128/27)
 * returns -1 on error, 0 on match, 1 when NO match.
 */

int match_ipmask(char *mask, aClient *cptr)
{
    int m;
    char *p;
    struct in_addr addr;
    char dummy[128];
    u_long lmask;

    strncpyzt(dummy, mask, sizeof(dummy));
    mask = dummy;

    if (!(p = strchr(mask, '/')))
	return -1;

    *p = '\0';

    m = atoi(p + 1);

    if (m < 0 || m > 32)
	return -1;

    lmask = htonl((u_long) 0xffffffffL << (32 - m));
    addr.s_addr = inet_addr(mask);
    return ((addr.s_addr ^ cptr->ip.s_addr) & lmask) ? 1 : 0;

}

int check_forbidden_words(char *text)
{

    if ((!match("*/server*", text)) || (!match("*irc.*", text)) ||
	(!match("*script*", text)) || (strchr(text, '#')) ||
	(!match("*www.*", text))) {

	return 1;
    }
    return 0;

}

char *find_displaypass()
{
aConfItem *aconf;

    for (aconf = conf; aconf; aconf = aconf->next)
	if (aconf->status & CONF_DRPASS)
	    return (aconf->name);

    return NULL;
}

char *stealth_server(char *name, char *out)
{
    char *tmp = NULL;
    char arr[HOSTLEN];

    tmp = strdup(name);
    tmp = strpbrk(tmp, ".");
    sprintf(arr, "*%s", tmp);
    strcpy(out, arr);
    return out;
}

/*
** m_post
**      parv[0] = sender prefix
**      parv[1] = comment
*/
int m_post(aClient *cptr, aClient *sptr, int parc, char *parv[]) {

    sendto_realops_lev(REJ_LEV, "Client rejected for POST command: [%s]",
               		cptr->sockhost);
    return exit_client(cptr, sptr, sptr, "Client Exit");
}

