/************************************************************************
 *   IRC - Internet Relay Chat, src/m_capab.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#if defined( HAVE_STRING_H )
# include <string.h>
#else
  /*
   * older unices don't have strchr/strrchr .. help them out 
   */
# include <strings.h>
# undef strchr
# define strchr index
#endif

struct capablist {
    char *name;
    unsigned int flag;
};

struct capablist my_capabs[] = {
    {"TS3", CAPAB_TS3},
    {"NOQUIT", CAPAB_NOQUIT},
    {"SSJOIN", CAPAB_NSJOIN},
    {"BURST", CAPAB_BURST},
    {"UNCONNECT", CAPAB_UNCONN},
    {"ZIP", CAPAB_ZIP},
#ifdef HAVE_ENCRYPTION_ON
    {"DKEY", CAPAB_DKEY},
#endif
#ifdef USE_FAKEHOST
    {"DT1", CAPAB_DTONE},
#endif
    {"SERVICE", CAPAB_SERVICES},
    {"EX-REX", CAPAB_EXCLUDE},
    {"NICKIP", CAPAB_NICKIP},
    /*   { "CANHIDE",               CAPAB_CANHIDE },   */
    /*   { "DT2",           CAPAB_DTONE|CAPAB_CANHIDE }, */
    /* The above two are commented out, because the feature is not
     * completed. It will be completed, after the hunt_server() case
     * is totally solved, which requires us to modify every me.name
     * cases to be stealth_server()'ed.
     *
     * The new SERVER command can still be used, but is not needed,
     * if the feature is not applicable.
     *
     * -TimeMr14C
     */
    {NULL, 0}
};

void make_capab_string(char *buf, unsigned int noflags)
{
    int i;
    int c;

    *buf = '\0';

    for (i = c = 0; my_capabs[i].name; i++) {
	if (!(my_capabs[i].flag & noflags)) {
	    if (c)
		strcat(buf, " ");
	    strcat(buf, my_capabs[i].name);
	    c++;
	}
    }
}

/*  
 * m_capab
 * Communicate what I can do to another server
 * This has to be able to be sent and understood while
 * the client is UNREGISTERED. Therefore, we
 * absolutely positively must not check to see if
 * this is a server or a client. It's probably an unknown!
 */
int m_capab(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int i, j;

    /* If it's not local, or it has already set capabilities, silently ignore it. */

    if (cptr != sptr || cptr->capabilities)
	return 0;

    for (i = 1; i < parc; i++) {
	for (j = 0; my_capabs[j].name; j++) {
	    if (strcmp(parv[i], my_capabs[j].name) == 0)
		cptr->capabilities |= my_capabs[j].flag;
	}
    }

    return 0;
}

void send_capab_to(aClient *cptr, int noflag)
{
    char capabbuf[512];
    make_capab_string(capabbuf, noflag);
    sendto_one(cptr, "CAPAB %s", capabbuf);
    return;
}
