/************************************************************************
 *   IRC - Internet Relay Chat, src/m_capab.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "dh.h"

extern int do_server_estab(aClient *);

#define DKEY_GOTIN  0x01
#define DKEY_GOTOUT 0x02

#define DKEY_DONE(x) (((x) & (DKEY_GOTIN|DKEY_GOTOUT)) == (DKEY_GOTIN|DKEY_GOTOUT))

int m_dkey(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (!(IsNegoServer(sptr) && parc > 1)) {
	if (IsPerson(sptr))
	    return 0;
	return exit_client(sptr, sptr, sptr, "Not negotiating now");
    }
#ifdef HAVE_ENCRYPTION_ON
    if (mycmp(parv[1], "START") == 0) {
    char keybuf[1024];

	if (parc != 2)
	    return exit_client(sptr, sptr, sptr, "DKEY START failure");

	if (sptr->serv->sessioninfo_in != NULL &&
	    sptr->serv->sessioninfo_out != NULL)
	    return exit_client(sptr, sptr, sptr, "DKEY START duplicate?!");

	sptr->serv->sessioninfo_in = dh_start_session();
	sptr->serv->sessioninfo_out = dh_start_session();
	sendto_realops("Initiating diffie-hellman key exchange with %s",
		       sptr->name);

	dh_get_s_public(keybuf, 1024, sptr->serv->sessioninfo_in);
	sendto_one(sptr, "DKEY PUB I %s", keybuf);

	dh_get_s_public(keybuf, 1024, sptr->serv->sessioninfo_out);
	sendto_one(sptr, "DKEY PUB O %s", keybuf);
	return 0;
    }

    if (mycmp(parv[1], "PUB") == 0) {
    char keybuf[1024];

	if (parc != 4 || !sptr->serv->sessioninfo_in ||
	    !sptr->serv->sessioninfo_out)
	    return exit_client(sptr, sptr, sptr, "DKEY PUB failure");

	if (mycmp(parv[2], "O") == 0) {	/* their out is my in! */
	    if (!dh_generate_shared(sptr->serv->sessioninfo_in, parv[3]))
		return exit_client(sptr, sptr, sptr, "DKEY PUB O invalid");
	    sptr->serv->dkey_flags |= DKEY_GOTOUT;
	} else if (mycmp(parv[2], "I") == 0) {	/* their out is my in! */
	    if (!dh_generate_shared(sptr->serv->sessioninfo_out, parv[3]))
		return exit_client(sptr, sptr, sptr, "DKEY PUB I invalid");
	    sptr->serv->dkey_flags |= DKEY_GOTIN;
	} else
	    return exit_client(sptr, sptr, sptr, "DKEY PUB bad option");

	if (DKEY_DONE(sptr->serv->dkey_flags)) {
    int hexlen;
    unsigned char hexbuf[1024];

	    sendto_one(sptr, "DKEY DONE");
	    SetRC4OUT(sptr);

	    if (dh_get_s_shared(keybuf, 1024, sptr->serv->sessioninfo_in)
		== NULL || dh_hexstr_to_raw(keybuf, hexbuf, &hexlen) == 0)
		return exit_client(sptr, sptr, sptr,
				   "Could not setup encrypted session");
	    sptr->serv->rc4_in = rc4_initstate(hexbuf, hexlen);

	    if (dh_get_s_shared(keybuf, 1024, sptr->serv->sessioninfo_out)
		== NULL || dh_hexstr_to_raw(keybuf, hexbuf, &hexlen) == 0)
		return exit_client(sptr, sptr, sptr,
				   "Could not setup encrypted session");
	    sptr->serv->rc4_out = rc4_initstate(hexbuf, hexlen);

	    dh_end_session(sptr->serv->sessioninfo_in);
	    dh_end_session(sptr->serv->sessioninfo_out);

	    sptr->serv->sessioninfo_in = sptr->serv->sessioninfo_out = NULL;
	    return 0;
	}
	return 0;
    }

    if (mycmp(parv[1], "DONE") == 0) {
	if (!
	    ((sptr->serv->sessioninfo_in == NULL &&
	      sptr->serv->sessioninfo_out == NULL)
	     && (sptr->serv->rc4_in != NULL && sptr->serv->rc4_out != NULL)))
	    return exit_client(sptr, sptr, sptr, "DKEY DONE when not done!");
	SetRC4IN(sptr);
	sendto_realops
	    ("Diffie-Hellman exchange with %s complete, connection encrypted.",
	     sptr->name);
	sendto_one(sptr, "DKEY EXIT");
	return RC4_NEXT_BUFFER;
    }

    if (mycmp(parv[1], "EXIT") == 0) {
	if (!(IsRC4IN(sptr) && IsRC4OUT(sptr)))
	    return exit_client(sptr, sptr, sptr,
			       "DKEY EXIT when not in proper stage");
	ClearNegoServer(sptr);
	return do_server_estab(sptr);
    }
#endif

    return 0;
}
