/************************************************************************
 *   IRC - Internet Relay Chat, src/m_capab.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "msg.h"
#include "numeric.h"
#include "channel.h"
#if defined( HAVE_STRING_H )
# include <string.h>
#else
# include <strings.h>
# undef strchr
# define strchr index
#endif

extern int check_for_flud(aClient *, aClient *, aChannel *, int);

/*
 * check to see if the message has any color chars in it.
 */
int msg_has_colors(char *msg)
{
    char *c = msg;

    while (*c) {
	if (*c == '\003' || *c == '\033')
	    break;
	else
	    c++;
    }

    if (*c)
	return 1;

    return 0;
}

/*                          
 * m_message (used in m_private() and m_notice()) the general
 * function to deliver MSG's between users/channels
 *
 * parv[0] = sender prefix
 * parv[1] = receiver list
 * parv[2] = message text
 *
 * massive cleanup * rev argv 6/91
 *
 */

static inline int m_message(aClient *cptr, aClient *sptr, int parc,
			    char *parv[], int notice)
{
    aClient *acptr;
    char *s;
    int i, ret, ischan;
    aChannel *chptr;
    char *nick, *server, *p, *cmd, *dccmsg;

    cmd = notice ? MSG_NOTICE : MSG_PRIVATE;

    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NORECIPIENT), me.name, parv[0], cmd);
	return -1;
    }

    if (parc < 3 || *parv[2] == '\0') {
	sendto_one(sptr, err_str(ERR_NOTEXTTOSEND), me.name, parv[0]);
	return -1;
    }

    if (MyConnect(sptr)) {
#ifdef ANTI_SPAMBOT
#ifndef ANTI_SPAMBOT_WARN_ONLY
	/*
	 * if its a spambot, just ignore it
	 */
	if (sptr->join_leave_count >= MAX_JOIN_LEAVE_COUNT)
	    return 0;
#endif
#endif
	parv[1] = canonize(parv[1]);
    }

    for (p = NULL, nick = strtoken(&p, parv[1], ","), i = 0;
	 nick && i < 20; nick = strtoken(&p, NULL, ",")) {
	/*
	 * If someone is spamming via "/msg nick1,nick2,nick3,nick4 SPAM"
	 * (or even to channels) then subject them to flood control!
	 * -Taner
	 */
	if (i++ > 10)
#ifdef NO_OPER_FLOOD
	    if (!IsAnOper(sptr) && !IsULine(sptr))
#endif
		sptr->since += 4;

	/*
	 * channel msg?
	 */
	ischan = IsChannelName(nick);
	if (ischan && (chptr = find_channel(nick, NullChn))) {
/****** NOTICE OP ******/
	    if ((!is_chan_op(sptr, chptr) && !IsULine(sptr) && !IsAnOper(sptr)
		 && !IsServer(sptr)
		) && (notice)) {
		sendto_one(sptr,
			   ":%s NOTICE %s :ONLY ChanOps can send NOTICE to channels",
			   me.name, parv[0]);
		return 0;
	    }
/************/

	    if (!notice)
		switch (check_for_ctcp(parv[2], NULL)) {
		    case CTCP_NONE:
			break;

		    case CTCP_DCCSEND:
		    case CTCP_DCC:
			sendto_one(sptr,
				   ":%s NOTICE %s :You may not send a DCC command to a channel (%s)",
				   me.name, parv[0], nick);
			continue;

		    default:
#ifdef FLUD
			if (check_for_flud(sptr, NULL, chptr, 1))
			    return 0;
#endif
			break;
		}
	    ret = IsULine(sptr) ? 0 : can_send(sptr, chptr, parv[2]);
	    if (MyClient(sptr) && (ret == ERR_NOCOLORSONCHAN)) {
		sendto_one(sptr, err_str(ERR_NOCOLORSONCHAN), me.name, parv[0],
			   nick, parv[2]);
		continue;
	    }
	    if (MyClient(sptr) && (ret == ERR_NEEDREGGEDNICK)) {
		sendto_one(sptr, err_str(ERR_NEEDREGGEDNICK), me.name, parv[0],
			   nick, "speak in");
		continue;
	    }
	    if (!ret && MyClient(sptr) && (chptr->mode.mode & MODE_NOCOLOR)
		&& msg_has_colors(parv[2])) {
		if (!notice)
		    sendto_one(sptr, err_str(ERR_NOCOLORSONCHAN), me.name,
			       parv[0], nick, parv[2]);
		continue;
	    }

	    if (ret) {
		if (!notice)
		    sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name,
			       parv[0], nick);
	    } else
		sendto_channel_butone(cptr, 0, sptr, chptr, ":%s %s %s :%s",
				      parv[0], cmd, nick, parv[2]);
	    continue;
	}
	/*
	 * nickname addressed?
	 */
	if (!ischan && (acptr = find_person(nick, NULL))) {
	    if (IsNoNonReg(acptr) && !IsRegNick(sptr) && !IsULine(sptr)
		&& !IsOper(sptr) && !IsService(sptr)) {
		sendto_one(sptr, rpl_str(ERR_NONONREG), me.name, parv[0],
			   acptr->name);
		return 0;
	    }
#ifdef FLUD
	    if (!notice && MyFludConnect(acptr))
#else
	    if (!notice && MyConnect(acptr))
#endif
	    {

		switch (check_for_ctcp(parv[2], &dccmsg)) {
		    case CTCP_NONE:
			break;

		    case CTCP_DCCSEND:
#ifdef FLUD
			if (check_for_flud(sptr, acptr, NULL, 1))
			    return 0;
#endif

			if (check_dccsend(sptr, acptr, dccmsg))
			    continue;
			break;

		    default:
#ifdef FLUD
			if (check_for_flud(sptr, acptr, NULL, 1))
			    return 0;
#endif
			break;
		}
	    }
	    if (!is_silenced(sptr, acptr)) {
		if (!notice && MyClient(acptr) && acptr->user &&
		    acptr->user->away)
		    sendto_one(sptr, rpl_str(RPL_AWAY), me.name, parv[0],
			       acptr->name, acptr->user->away);
		sendto_prefix_one(acptr, sptr, ":%s %s %s :%s", parv[0], cmd,
				  nick, parv[2]);
	    }
	    continue;
	}

	if (nick[1] == '#' && nick[0] != '#') {
	    if (nick[0] == '@') {
		if ((chptr = find_channel(nick + 1, NullChn))) {
		    if (can_send(sptr, chptr, parv[2]) == 0 || IsULine(sptr))
			sendto_channel_butone(cptr, CHFL_CHANOP, sptr, chptr,
					      ":%s %s %s :%s", parv[0], cmd,
					      nick, parv[2]);
		    else if (!notice)
			sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name,
				   parv[0], nick + 1);
		}
	    } else if (nick[0] == '+') {
		if ((chptr = find_channel(nick + 1, NullChn))) {
		    if (can_send(sptr, chptr, parv[2]) == 0 || IsULine(sptr))
			sendto_channel_butone(cptr, CHFL_VOICE, sptr, chptr,
					      ":%s %s %s :%s", parv[0], cmd,
					      nick, parv[2]);
		    else if (!notice)
			sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN),
				   me.name, parv[0], nick + 1);
		}
	    } else
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0],
			   nick + 1);

	    continue;
	}

	if (nick[0] == '@' && nick[1] == '+' && nick[2] == '#') {
	    if ((chptr = find_channel(nick + 2, NullChn))) {
		if (can_send(sptr, chptr, parv[2]) == 0 || IsULine(sptr))
		    sendto_channel_butone(cptr, CHFL_CHANOP | CHFL_VOICE,
					  sptr, chptr, ":%s %s %s :%s",
					  parv[0], cmd, nick, parv[2]);
		else if (!notice)
		    sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name,
			       parv[0], nick + 1);
	    } else
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0],
			   nick + 1);
	    continue;
	}

	if (IsAnOper(sptr)) {
	    /*
	     * the following two cases allow masks in NOTICEs
	     * (for OPERs only)
	     * 
	     * Armin, 8Jun90 (gruner@informatik.tu-muenchen.de)
	     */
	    if ((*nick == '$' || *nick == '#')) {
		if (!(s = (char *) strrchr(nick, '.'))) {
		    sendto_one(sptr, err_str(ERR_NOTOPLEVEL), me.name,
			       parv[0], nick);
		    continue;
		}
		while (*++s)
		    if (*s == '.' || *s == '*' || *s == '?')
			break;
		if (*s == '*' || *s == '?') {
		    sendto_one(sptr, err_str(ERR_WILDTOPLEVEL), me.name,
			       parv[0], nick);
		    continue;
		}
		sendto_match_butone(IsServer(cptr) ? cptr : NULL, sptr,
				    nick + 1,
				    (*nick ==
				     '#') ? MATCH_HOST : MATCH_SERVER,
				    ":%s %s %s :%s", parv[0], cmd, nick,
				    parv[2]);
		continue;
	    }
	}
	/*
	 * user@server addressed?
	 */
	if (!ischan && (server = (char *) strchr(nick, '@'))
	    && (acptr = find_server(server + 1, NULL))) {
    int count = 0;

	    /*
	     * Not destined for a user on me :-(
	     */
	    if (!IsMe(acptr)) {
		sendto_one(acptr, ":%s %s %s :%s", parv[0], cmd, nick, parv[2]);
		continue;
	    }
	    *server = '\0';

	    /*
	     * Look for users which match the destination host
	     * (no host == wildcard) and if one and one only is found
	     * connected to me, deliver message!
	     */
	    acptr = find_person(nick, NULL);
	    if (server)
		*server = '@';
	    if (acptr) {
		if (count == 1)
		    sendto_prefix_one(acptr, sptr, ":%s %s %s :%s", parv[0],
				      cmd, nick, parv[2]);
		else if (!notice)
		    sendto_one(sptr, err_str(ERR_TOOMANYTARGETS), me.name,
			       parv[0], nick);
	    }
	    if (acptr)
		continue;
	}
	sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], nick);
    }
    if ((i > 20) && sptr->user)
	sendto_realops_lev(SPY_LEV,
			   "User %s (%s@%s) tried to msg %d users",
			   sptr->name, sptr->user->username,
			   sptr->user->host, i);
    return 0;
}

    /*
     * * m_private *      parv[0] = sender prefix *       parv[1] =
     * receiver list *      parv[2] = message text
     */

int m_private(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

#ifdef USE_ZOMBIE
    if (IsZombie(sptr)) {
	sendto_one(sptr, getreply(ERR_YOUAREZOMBIE), me.name, sptr->name,
		   parv[1]);
	return 0;
    } else {
#endif
	return m_message(cptr, sptr, parc, parv, 0);
#ifdef USE_ZOMBIE
    }
#endif
}

    /*
     * * m_notice *       parv[0] = sender prefix *       parv[1] = receiver list *
     * parv[2] = notice text
     */

int m_notice(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
#ifdef USE_ZOMBIE
    if (IsZombie(sptr)) {
	sendto_one(sptr, getreply(ERR_YOUAREZOMBIE), me.name, sptr->name,
		   parv[1]);
	return 0;
    } else {
#endif
	return m_message(cptr, sptr, parc, parv, 1);
#ifdef USE_ZOMBIE
    }
#endif
}
