/*
 *   IRC - Internet Relay Chat, src/m_mode.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Co Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: m_mode.c,v 1.17 2001/12/05 18:31:50 tr-ircd Exp $ 
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "channel.h"
#include "h.h"

/*
 * some buffers for rebuilding channel/nick lists with ,'s 
 */
static char buf[BUFSIZE];
static char modebuf[REALMODEBUFLEN], parabuf[REALMODEBUFLEN];

/*
 * externally defined function 
 */
extern Link *find_channel_link(Link *, aChannel *);	/* defined in list.c */

/*
 * Fixes a string so that the first white space found becomes an end of
 * string marker (`\-`).  returns the 'fixed' string or "*" if the
 * string was NULL length or a NULL pointer.
 */
static char *check_string(char *s)
{
    static char star[2] = "*";
    char *str = s;

    if (BadPtr(s))
	return star;

    for (; *s; s++)
	if (isspace(*s)) {
	    *s = '\0';
	    break;
	}

    return (BadPtr(str)) ? star : str;
}

/*
 * create a string of form "foo!bar@fubar" given foo, bar and fubar as
 * the parameters.  If NULL, they become "*".
 */
static char *make_nick_user_host(char *nick, char *name, char *host)
{
    static char namebuf[NICKLEN + USERLEN + HOSTLEN + 6];
    int n;
    char *ptr1, *ptr2;

    ptr1 = namebuf;
    for (ptr2 = check_string(nick), n = NICKLEN; *ptr2 && n--;)
	*ptr1++ = *ptr2++;
    *ptr1++ = '!';
    for (ptr2 = check_string(name), n = USERLEN; *ptr2 && n--;)
	*ptr1++ = *ptr2++;
    *ptr1++ = '@';
    for (ptr2 = check_string(host), n = HOSTLEN; *ptr2 && n--;)
	*ptr1++ = *ptr2++;
    *ptr1 = '\0';
    return (namebuf);
}

int add_banid(aClient *cptr, aChannel *chptr, char *banid)
{
    aBan *ban;
    int cnt = 0;
    chanMember *cm;
    char *s, nickuhost[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
    char fakeuhost[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
    for (ban = chptr->banlist; ban; ban = ban->next) {
	if (MyClient(cptr) && (++cnt >= MAXBANS)) {
	    sendto_one(cptr, getreply(ERR_BANLISTFULL), me.name, cptr->name,
		       chptr->chname, banid);
	    return -1;
	}
	/*
	 * yikes, we were doing all sorts of weird crap here before, now
	 * * we ONLY want to know if current bans cover this ban, not if this
	 * * ban covers current ones, since it may cover other things too -wd 
	 */
	else if (!match(ban->banstr, banid))
	    return -1;
    }

    ban = (aBan *) MyMalloc(sizeof(aBan));
    ban->banstr = (char *) MyMalloc(strlen(banid) + 1);
    (void) strcpy(ban->banstr, banid);
    ban->next = chptr->banlist;

    if (IsPerson(cptr)) {
	ban->who =
	    (char *) MyMalloc(strlen(cptr->name) +
			      strlen(cptr->user->username) +
#ifdef USE_FAKEHOST
			      strlen(IsFake(cptr) ? cptr->user->
				     fakehost : cptr->user->host) + 3);
#else
			      strlen(cptr->user->host) + 3);
#endif

	(void) ircsprintf(ban->who, "%s!%s@%s", cptr->name,
			  cptr->user->username,
#ifdef USE_FAKEHOST
			  IsFake(cptr) ? cptr->user->fakehost : cptr->user->host
#else
			  cptr->user->host
#endif
	    );
    } else {
	ban->who = (char *) MyMalloc(strlen(cptr->name) + 1);
	(void) strcpy(ban->who, cptr->name);
    }

    /*
     * determine what 'type' of mask this is, for less matching later 
     */

    if (banid[0] == '*' && banid[1] == '!') {
	if (banid[2] == '*' && banid[3] == '@')
	    ban->type = MTYP_HOST;
	else
	    ban->type = MTYP_USERHOST;
    } else
	ban->type = MTYP_FULL;

    ban->when = timeofday;
    chptr->banlist = ban;

    for (cm = chptr->members; cm; cm = cm->next) {
	if (!MyConnect(cm->cptr))
	    continue;

	strcpy(nickuhost,
	       make_nick_user_host(cm->cptr->name, cm->cptr->user->username,
				   cm->cptr->hostip));
#ifdef USE_FAKEHOST
	strcpy(fakeuhost,
	       make_nick_user_host(cm->cptr->name,
				   cm->cptr->user->username,
				   cm->cptr->user->fakehost));
#endif
	s = make_nick_user_host(cm->cptr->name, cm->cptr->user->username,
				cm->cptr->user->host);
	if (match(banid, nickuhost) == 0 || match(banid, s) == 0
#ifdef USE_FAKEHOST
	    || match(banid, fakeuhost) == 0
#endif
	    )
	    cm->bans++;
    }
    return 0;
}

/*
 * del_banid - delete an id belonging to cptr if banid is null,
 * deleteall banids belonging to cptr.
 */
int del_banid(aChannel *chptr, char *banid)
{
    aBan **ban;
    aBan *tmp;
    chanMember *cm;
    char *s, nickuhost[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
    char fakeuhost[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
    if (!banid)
	return -1;
    for (ban = &(chptr->banlist); *ban; ban = &((*ban)->next))
	if (mycmp(banid, (*ban)->banstr) == 0) {
	    tmp = *ban;
	    *ban = tmp->next;

	    for (cm = chptr->members; cm; cm = cm->next) {
		if (!MyConnect(cm->cptr) || cm->bans == 0)
		    continue;

		strcpy(nickuhost,
		       make_nick_user_host(cm->cptr->name,
					   cm->cptr->user->username,
					   cm->cptr->hostip));
#ifdef USE_FAKEHOST
		strcpy(fakeuhost,
		       make_nick_user_host(cm->cptr->name,
					   cm->cptr->user->username,
					   cm->cptr->user->fakehost));
#endif
		s = make_nick_user_host(cm->cptr->name,
					cm->cptr->user->username,
					cm->cptr->user->host);
		if (match(banid, nickuhost) == 0 || match(banid, s) == 0
#ifdef USE_FAKEHOST
		    || match(banid, fakeuhost) == 0
#endif
		    )
		    cm->bans--;
	    }
	    MyFree(tmp->banstr);
	    MyFree(tmp->who);
	    MyFree(tmp);
	    break;
	}
    return 0;
}

/*
 * is_banned - returns a pointer to the ban structure if banned else
 * NULL
 * 
 * IP_BAN_ALL from comstud always on...
 */

aBan *is_banned(aClient *cptr, aChannel *chptr)
{
aBan *tmp;
char s[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
char fakeuhost[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
char *s2;

    if (!IsPerson(cptr))
	return NULL;

    strcpy(s,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->host));
    s2 = make_nick_user_host(cptr->name, cptr->user->username, cptr->hostip);
#ifdef USE_FAKEHOST
    strcpy(fakeuhost,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->fakehost));
#endif
    for (tmp = chptr->banlist; tmp; tmp = tmp->next)
	if ((match(tmp->banstr, s) == 0) || (match(tmp->banstr, s2) == 0)
#ifdef USE_FAKEHOST
	    || (match(tmp->banstr, fakeuhost) == 0)
#endif
	    )
	    break;
    return (tmp);
}

aBan *nick_is_banned(aChannel *chptr, char *nick, aClient *cptr)
{
aBan *tmp;
char *s, s2[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
char fakeuhost[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
    if (!IsPerson(cptr))
	return NULL;

    strcpy(s2,
	   make_nick_user_host(nick, cptr->user->username, cptr->user->host));
#ifdef USE_FAKEHOST
    strcpy(fakeuhost,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->fakehost));
#endif
    s = make_nick_user_host(nick, cptr->user->username, cptr->hostip);

    for (tmp = chptr->banlist; tmp; tmp = tmp->next)
	if (tmp->type == MTYP_FULL &&	/* only check applicable bans */
	    ((match(tmp->banstr, s2) == 0) ||	/* check host before IP */
	     (match(tmp->banstr, s) == 0)
#ifdef USE_FAKEHOST
	     || (match(tmp->banstr, fakeuhost) == 0)
#endif
	    ))
	    break;
    return (tmp);
}

void remove_matching_bans(aChannel *chptr, aClient *cptr, aClient *from)
{
    aBan *ban, *bnext;
    char targhost[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
    char fakeuhost[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
    char targip[NICKLEN + USERLEN + HOSTLEN + 6];
    char *m;
    int count = 0, send = 0;

    if (!IsPerson(cptr))
	return;

    strcpy(targhost,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->host));
#ifdef USE_FAKEHOST
    strcpy(fakeuhost,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->fakehost));
#endif
    strcpy(targip,
	   make_nick_user_host(cptr->name, cptr->user->username, cptr->hostip));

    m = modebuf;
    *m++ = '-';
    *m = '\0';

    *parabuf = '\0';

    ban = chptr->banlist;

    while (ban) {
	bnext = ban->next;
	if ((match(ban->banstr, targhost) == 0)
	    || (match(ban->banstr, targip) == 0)
#ifdef USE_FAKEHOST
	    || (match(ban->banstr, fakeuhost) == 0)
#endif
	    ) {
	    if (strlen(parabuf) + strlen(ban->banstr) + 10 <
		(size_t) MODEBUFLEN) {
		if (*parabuf)
		    strcat(parabuf, " ");
		strcat(parabuf, ban->banstr);
		count++;
		*m++ = 'b';
		*m = '\0';
	    } else if (*parabuf)
		send = 1;

	    if (count == MAXMODEPARAMS)
		send = 1;

	    if (send) {
		sendto_channel_butserv(chptr, from, ":%s MODE %s %s %s",
				       from->name, chptr->chname, modebuf,
				       parabuf);
		sendto_serv_butone(from, ":%s MODE %s %s %s", from->name,
				   chptr->chname, modebuf, parabuf);
		send = 0;
		*parabuf = '\0';
		m = modebuf;
		*m++ = '-';
		if (count != MAXMODEPARAMS) {
		    strcpy(parabuf, ban->banstr);
		    *m++ = 'b';
		    count = 1;
		} else
		    count = 0;
		*m = '\0';
	    }

	    del_banid(chptr, ban->banstr);
	}
	ban = bnext;
    }

    if (*parabuf) {
	sendto_channel_butserv(chptr, from, ":%s MODE %s %s %s",
			       from->name, chptr->chname, modebuf, parabuf);
	sendto_serv_butone(from, ":%s MODE %s %s %s", from->name, chptr->chname,
			   modebuf, parabuf);
    }

    return;
}

int add_inviteid(aClient *cptr, aChannel *chptr, char *inviteid)
{
    anInvite *invite;
    int cnt = 0;
    chanMember *cm;
    char *s, nickuhost[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
    char fakeuhost[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
    for (invite = chptr->invitelist; invite; invite = invite->next) {
	if (MyClient(cptr) && (++cnt >= MAXINVITES)) {
	    sendto_one(cptr, getreply(ERR_BANLISTFULL), me.name,
		       cptr->name, chptr->chname, inviteid);
	    return -1;
	} else if (!match(invite->invitestr, inviteid))
	    return -1;
    }

    invite = (anInvite *) MyMalloc(sizeof(anInvite));
    invite->invitestr = (char *) MyMalloc(strlen(inviteid) + 1);
    (void) strcpy(invite->invitestr, inviteid);
    invite->next = chptr->invitelist;
    if (IsPerson(cptr)) {
	invite->who =
	    (char *) MyMalloc(strlen(cptr->name) +
			      strlen(cptr->user->username) +
#ifdef USE_FAKEHOST
			      strlen(IsFake(cptr) ? cptr->user->
				     fakehost : cptr->user->host)
#else
			      strlen(cptr->user->host)
#endif
			      + 3);

	(void) ircsprintf(invite->who, "%s!%s@%s", cptr->name,
			  cptr->user->username,
#ifdef USE_FAKEHOST
			  IsFake(cptr) ? cptr->user->fakehost : cptr->user->host
#else
			  cptr->user->host
#endif
	    );
    } else {
	invite->who = (char *) MyMalloc(strlen(cptr->name) + 1);
	(void) strcpy(invite->who, cptr->name);
    }

    if (inviteid[0] == '*' && inviteid[1] == '!') {
	if (inviteid[2] == '*' && inviteid[3] == '@')
	    invite->type = MTYP_HOST;
	else
	    invite->type = MTYP_USERHOST;
    } else
	invite->type = MTYP_FULL;

    invite->when = timeofday;
    chptr->invitelist = invite;

    for (cm = chptr->members; cm; cm = cm->next) {
	if (!MyConnect(cm->cptr))
	    continue;

	strcpy(nickuhost,
	       make_nick_user_host(cm->cptr->name, cm->cptr->user->username,
				   cm->cptr->hostip));
#ifdef USE_FAKEHOST
	strcpy(fakeuhost,
	       make_nick_user_host(cm->cptr->name,
				   cm->cptr->user->username,
				   cm->cptr->user->fakehost));
#endif
	s = make_nick_user_host(cm->cptr->name, cm->cptr->user->username,
				cm->cptr->user->host);
	if (match(inviteid, nickuhost) == 0 || match(inviteid, s) == 0
#ifdef USE_FAKEHOST
	    || match(inviteid, fakeuhost) == 0
#endif
	    )
	    cm->invites++;
    }
    return 0;
}

int del_inviteid(aChannel *chptr, char *inviteid)
{
    anInvite **invite;
    anInvite *tmp;
    chanMember *cm;
    char *s, nickuhost[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
    char fakeuhost[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
    if (!inviteid)
	return -1;
    for (invite = &(chptr->invitelist); *invite; invite = &((*invite)->next))
	if (mycmp(inviteid, (*invite)->invitestr) == 0) {
	    tmp = *invite;
	    *invite = tmp->next;

	    for (cm = chptr->members; cm; cm = cm->next) {
		if (!MyConnect(cm->cptr) || cm->bans == 0)
		    continue;

		strcpy(nickuhost,
		       make_nick_user_host(cm->cptr->name,
					   cm->cptr->user->username,
					   cm->cptr->hostip));
#ifdef USE_FAKEHOST
		strcpy(fakeuhost,
		       make_nick_user_host(cm->cptr->name,
					   cm->cptr->user->username,
					   cm->cptr->user->fakehost));
#endif

		s = make_nick_user_host(cm->cptr->name,
					cm->cptr->user->username,
					cm->cptr->user->host);
		if (match(inviteid, nickuhost) == 0 || match(inviteid, s) == 0
#ifdef USE_FAKEHOST
		    || match(inviteid, fakeuhost) == 0
#endif
		    )
		    cm->invites--;
	    }
	    MyFree(tmp->invitestr);
	    MyFree(tmp->who);
	    MyFree(tmp);
	    break;
	}
    return 0;
}

anInvite *has_invite(aClient *cptr, aChannel *chptr)
{
anInvite *tmp;
char s[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
char s1[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
char *s2;

    if (!IsPerson(cptr))
	return NULL;

    strcpy(s,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->host));
#ifdef USE_FAKEHOST
    strcpy(s1,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->fakehost));
#endif
    s2 = make_nick_user_host(cptr->name, cptr->user->username, cptr->hostip);

    for (tmp = chptr->invitelist; tmp; tmp = tmp->next)
	if ((match(tmp->invitestr, s) == 0)
	    || (match(tmp->invitestr, s2) == 0)
#ifdef USE_FAKEHOST
	    || (match(tmp->invitestr, s1) == 0)
#endif
	    )
	    break;
    return (tmp);
}

void remove_matching_invites(aChannel *chptr, aClient *cptr, aClient *from)
{
    anInvite *invi, *bnext;
    char targhost[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
    char fakeuhost[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
    char targip[NICKLEN + USERLEN + HOSTLEN + 6];
    char *m;
    int count = 0, send = 0;

    if (!IsPerson(cptr))
	return;

    strcpy(targhost,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->host));
#ifdef USE_FAKEHOST
    strcpy(fakeuhost,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->fakehost));
#endif
    strcpy(targip,
	   make_nick_user_host(cptr->name, cptr->user->username, cptr->hostip));

    m = modebuf;
    *m++ = '-';
    *m = '\0';

    *parabuf = '\0';

    invi = chptr->invitelist;

    while (invi) {
	bnext = invi->next;
	if ((match(invi->invitestr, targhost) == 0)
	    || (match(invi->invitestr, targip) == 0)
#ifdef USE_FAKEHOST
	    || (match(invi->invitestr, fakeuhost) == 0)
#endif
	    ) {
	    if (strlen(parabuf) + strlen(invi->invitestr) + 10 <
		(size_t) MODEBUFLEN) {
		if (*parabuf)
		    strcat(parabuf, " ");
		strcat(parabuf, invi->invitestr);
		count++;
		*m++ = 'I';
		*m = '\0';
	    } else if (*parabuf)
		send = 1;

	    if (count == MAXMODEPARAMS)
		send = 1;

	    if (send) {
		sendto_channel_butserv(chptr, from, ":%s MODE %s %s %s",
				       from->name, chptr->chname, modebuf,
				       parabuf);
		sendto_serv_butone(from, ":%s MODE %s %s %s", from->name,
				   chptr->chname, modebuf, parabuf);
		send = 0;
		*parabuf = '\0';
		m = modebuf;
		*m++ = '-';
		if (count != MAXMODEPARAMS) {
		    strcpy(parabuf, invi->invitestr);
		    *m++ = 'I';
		    count = 1;
		} else
		    count = 0;
		*m = '\0';
	    }

	    del_inviteid(chptr, invi->invitestr);
	}
	invi = bnext;
    }

    if (*parabuf) {
	sendto_channel_butserv(chptr, from, ":%s MODE %s %s %s",
			       from->name, chptr->chname, modebuf, parabuf);
	sendto_serv_butone(from, ":%s MODE %s %s %s", from->name, chptr->chname,
			   modebuf, parabuf);
    }

    return;
}

void kill_invites_list(aClient *cptr, aChannel *chptr)
{
    chanMember *cm;
    anInvite *bp, *bpn;
    char *cp;
    int count = 0, send = 0;

    cp = modebuf;
    *cp++ = '-';
    *cp = '\0';

    *parabuf = '\0';

    for (bp = chptr->invitelist; bp; bp = bp->next) {
	if (strlen(parabuf) + strlen(bp->invitestr) + 10 < (size_t) MODEBUFLEN) {
	    if (*parabuf)
		strcat(parabuf, " ");
	    strcat(parabuf, bp->invitestr);
	    count++;
	    *cp++ = 'I';
	    *cp = '\0';
	} else if (*parabuf)
	    send = 1;

	if (count == MAXMODEPARAMS)
	    send = 1;

	if (send) {
	    sendto_channel_butserv(chptr, &me, ":%s MODE %s %s %s",
				   cptr->name, chptr->chname, modebuf, parabuf);
	    send = 0;
	    *parabuf = '\0';
	    cp = modebuf;
	    *cp++ = '-';
	    if (count != MAXMODEPARAMS) {
		strcpy(parabuf, bp->invitestr);
		*cp++ = 'I';
		count = 1;
	    } else
		count = 0;
	    *cp = '\0';
	}
    }

    if (*parabuf) {
	sendto_channel_butserv(chptr, &me, ":%s MODE %s %s %s",
			       cptr->name, chptr->chname, modebuf, parabuf);
    }

    /*
     * physically destroy channel ban list 
     */

    bp = chptr->invitelist;
    while (bp) {
	bpn = bp->next;
	MyFree(bp->invitestr);
	MyFree(bp->who);
	MyFree(bp);
	bp = bpn;
    }

    chptr->invitelist = NULL;

    /*
     * reset bquiet on all channel members 
     */
    for (cm = chptr->members; cm; cm = cm->next) {
	if (MyConnect(cm->cptr))
	    cm->invites = 0;
    }
}

int add_banexid(aClient *cptr, aChannel *chptr, char *banexid)
{
    aBanEx *banex;
    int cnt = 0;
    chanMember *cm;
    char *s, nickuhost[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
    char fakeuhost[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
    for (banex = chptr->banexlist; banex; banex = banex->next) {
	if (MyClient(cptr) && (++cnt >= MAXINVITES)) {
	    sendto_one(cptr, getreply(ERR_BANLISTFULL), me.name,
		       cptr->name, chptr->chname, banexid);
	    return -1;
	} else if (!match(banex->banexstr, banexid))
	    return -1;
    }

    banex = (aBanEx *) MyMalloc(sizeof(aBanEx));
    banex->banexstr = (char *) MyMalloc(strlen(banexid) + 1);
    (void) strcpy(banex->banexstr, banexid);
    banex->next = chptr->banexlist;

    if (IsPerson(cptr)) {
	banex->who =
	    (char *) MyMalloc(strlen(cptr->name) +
			      strlen(cptr->user->username) +
#ifdef USE_FAKEHOST
			      strlen(IsFake(cptr) ? cptr->user->
				     fakehost : cptr->user->host)
#else
			      strlen(cptr->user->host)
#endif
			      + 3);

	(void) ircsprintf(banex->who, "%s!%s@%s", cptr->name,
			  cptr->user->username,
#ifdef USE_FAKEHOST
			  IsFake(cptr) ? cptr->user->fakehost : cptr->user->host
#else
			  cptr->user->host
#endif
	    );

    } else {
	banex->who = (char *) MyMalloc(strlen(cptr->name) + 1);
	(void) strcpy(banex->who, cptr->name);
    }
    if (banexid[0] == '*' && banexid[1] == '!') {
	if (banexid[2] == '*' && banexid[3] == '@')
	    banex->type = MTYP_HOST;
	else
	    banex->type = MTYP_USERHOST;
    } else
	banex->type = MTYP_FULL;
    banex->when = timeofday;
    chptr->banexlist = banex;

    for (cm = chptr->members; cm; cm = cm->next) {
	if (!MyConnect(cm->cptr))
	    continue;

	strcpy(nickuhost,
	       make_nick_user_host(cm->cptr->name, cm->cptr->user->username,
				   cm->cptr->hostip));
#ifdef USE_FAKEHOST
	strcpy(fakeuhost,
	       make_nick_user_host(cm->cptr->name,
				   cm->cptr->user->username,
				   cm->cptr->user->fakehost));
#endif
	s = make_nick_user_host(cm->cptr->name, cm->cptr->user->username,
				cm->cptr->user->host);
	if (match(banexid, nickuhost) == 0 || match(banexid, s) == 0
#ifdef USE_FAKEHOST
	    || match(banexid, fakeuhost) == 0
#endif
	    )
	    cm->banexceptions++;
    }
    return 0;
}

int del_banexid(aChannel *chptr, char *banexid)
{
    aBanEx **banex;
    aBanEx *tmp;
    chanMember *cm;
    char *s, nickuhost[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
    char fakeuhost[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
    if (!banexid)
	return -1;
    for (banex = &(chptr->banexlist); *banex; banex = &((*banex)->next))
	if (mycmp(banexid, (*banex)->banexstr) == 0) {
	    tmp = *banex;
	    *banex = tmp->next;

	    for (cm = chptr->members; cm; cm = cm->next) {
		if (!MyConnect(cm->cptr) || cm->bans == 0)
		    continue;

		strcpy(nickuhost,
		       make_nick_user_host(cm->cptr->name,
					   cm->cptr->user->username,
					   cm->cptr->hostip));
#ifdef USE_FAKEHOST
		strcpy(fakeuhost,
		       make_nick_user_host(cm->cptr->name,
					   cm->cptr->user->username,
					   cm->cptr->user->fakehost));
#endif
		s = make_nick_user_host(cm->cptr->name,
					cm->cptr->user->username,
					cm->cptr->user->host);
		if (match(banexid, nickuhost) == 0 || match(banexid, s) == 0
#ifdef USE_FAKEHOST
		    || match(banexid, fakeuhost) == 0
#endif
		    )
		    cm->banexceptions--;
	    }
	    MyFree(tmp->banexstr);
	    MyFree(tmp->who);
	    MyFree(tmp);
	    break;
	}
    return 0;
}

aBanEx *has_banex(aClient *cptr, aChannel *chptr)
{
aBanEx *tmp;
char s[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
char s1[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
char *s2;

    if (!IsPerson(cptr))
	return NULL;

    strcpy(s,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->host));
#ifdef USE_FAKEHOST
    strcpy(s1,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->fakehost));
#endif
    s2 = make_nick_user_host(cptr->name, cptr->user->username, cptr->hostip);

    for (tmp = chptr->banexlist; tmp; tmp = tmp->next)
	if ((match(tmp->banexstr, s) == 0)
	    || (match(tmp->banexstr, s2) == 0)
#ifdef USE_FAKEHOST
	    || (match(tmp->banexstr, s1) == 0)
#endif
	    )
	    break;
    return (tmp);
}

void remove_matching_banex(aChannel *chptr, aClient *cptr, aClient *from)
{
    aBanEx *banex, *bnext;
    char targhost[NICKLEN + USERLEN + HOSTLEN + 6];

#ifdef USE_FAKEHOST
    char fakeuhost[NICKLEN + USERLEN + HOSTLEN + 6];
#endif
    char targip[NICKLEN + USERLEN + HOSTLEN + 6];
    char *m;
    int count = 0, send = 0;

    if (!IsPerson(cptr))
	return;

    strcpy(targhost,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->host));
#ifdef USE_FAKEHOST
    strcpy(fakeuhost,
	   make_nick_user_host(cptr->name, cptr->user->username,
			       cptr->user->fakehost));
#endif
    strcpy(targip,
	   make_nick_user_host(cptr->name, cptr->user->username, cptr->hostip));

    m = modebuf;
    *m++ = '-';
    *m = '\0';

    *parabuf = '\0';

    banex = chptr->banexlist;

    while (banex) {
	bnext = banex->next;
	if ((match(banex->banexstr, targhost) == 0)
	    || (match(banex->banexstr, targip) == 0)
#ifdef USE_FAKEHOST
	    || (match(banex->banexstr, fakeuhost) == 0)
#endif
	    ) {
	    if (strlen(parabuf) + strlen(banex->banexstr) + 10 <
		(size_t) MODEBUFLEN) {
		if (*parabuf)
		    strcat(parabuf, " ");
		strcat(parabuf, banex->banexstr);
		count++;
		*m++ = 'e';
		*m = '\0';
	    } else if (*parabuf)
		send = 1;

	    if (count == MAXMODEPARAMS)
		send = 1;

	    if (send) {
		sendto_channel_butserv(chptr, from, ":%s MODE %s %s %s",
				       from->name, chptr->chname, modebuf,
				       parabuf);
		sendto_serv_butone(from, ":%s MODE %s %s %s", from->name,
				   chptr->chname, modebuf, parabuf);
		send = 0;
		*parabuf = '\0';
		m = modebuf;
		*m++ = '-';
		if (count != MAXMODEPARAMS) {
		    strcpy(parabuf, banex->banexstr);
		    *m++ = 'e';
		    count = 1;
		} else
		    count = 0;
		*m = '\0';
	    }

	    del_banexid(chptr, banex->banexstr);
	}
	banex = bnext;
    }

    if (*parabuf) {
	sendto_channel_butserv(chptr, from, ":%s MODE %s %s %s",
			       from->name, chptr->chname, modebuf, parabuf);
	sendto_serv_butone(from, ":%s MODE %s %s %s", from->name, chptr->chname,
			   modebuf, parabuf);
    }

    return;
}

void kill_banex_list(aClient *cptr, aChannel *chptr)
{
    chanMember *cm;
    aBanEx *bp, *bpn;
    char *cp;
    int count = 0, send = 0;

    cp = modebuf;
    *cp++ = '-';
    *cp = '\0';

    *parabuf = '\0';

    for (bp = chptr->banexlist; bp; bp = bp->next) {
	if (strlen(parabuf) + strlen(bp->banexstr) + 10 < (size_t) MODEBUFLEN) {
	    if (*parabuf)
		strcat(parabuf, " ");
	    strcat(parabuf, bp->banexstr);
	    count++;
	    *cp++ = 'e';
	    *cp = '\0';
	} else if (*parabuf)
	    send = 1;

	if (count == MAXMODEPARAMS)
	    send = 1;
	if (send) {
	    sendto_channel_butserv(chptr, &me, ":%s MODE %s %s %s",
				   cptr->name, chptr->chname, modebuf, parabuf);
	    send = 0;
	    *parabuf = '\0';
	    cp = modebuf;
	    *cp++ = '-';
	    if (count != MAXMODEPARAMS) {
		strcpy(parabuf, bp->banexstr);
		*cp++ = 'e';
		count = 1;
	    } else
		count = 0;
	    *cp = '\0';
	}
    }

    if (*parabuf) {
	sendto_channel_butserv(chptr, &me, ":%s MODE %s %s %s",
			       cptr->name, chptr->chname, modebuf, parabuf);
    }

    /*
     * physically destroy channel ban list 
     */

    bp = chptr->banexlist;
    while (bp) {
	bpn = bp->next;
	MyFree(bp->banexstr);
	MyFree(bp->who);
	MyFree(bp);
	bp = bpn;
    }

    chptr->banexlist = NULL;

    /*
     * reset bquiet on all channel members 
     */
    for (cm = chptr->members; cm; cm = cm->next) {
	if (MyConnect(cm->cptr))
	    cm->banexceptions = 0;
    }
}

/*
 * write the "simple" list of channel modes for channel chptr onto
 * buffer mbuf with the parameters in pbuf.
 */
void channel_modes(aClient *cptr, char *mbuf, char *pbuf, aChannel *chptr)
{
    *mbuf++ = '+';
    if (chptr->mode.mode & MODE_SECRET)
	*mbuf++ = 's';
    else if (chptr->mode.mode & MODE_PRIVATE)
	*mbuf++ = 'p';
    if (chptr->mode.mode & MODE_MODERATED)
	*mbuf++ = 'm';
    if (chptr->mode.mode & MODE_TOPICLIMIT)
	*mbuf++ = 't';
    if (chptr->mode.mode & MODE_INVITEONLY)
	*mbuf++ = 'i';
    if (chptr->mode.mode & MODE_NOPRIVMSGS)
	*mbuf++ = 'n';
    if (chptr->mode.mode & MODE_REGISTERED)
	*mbuf++ = 'r';
    if (chptr->mode.mode & MODE_REGONLY)
	*mbuf++ = 'R';
    if (chptr->mode.mode & MODE_NOCOLOR)
	*mbuf++ = 'c';
    if (chptr->mode.mode & MODE_OPERONLY)
	*mbuf++ = 'O';
    if (chptr->mode.mode & MODE_ANONYMOUS)
	*mbuf++ = 'q';
    if (chptr->mode.limit) {
	*mbuf++ = 'l';
	if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr)) {
	    if (*chptr->mode.key)
		ircsprintf(pbuf, "%d ", chptr->mode.limit);
	    else
		ircsprintf(pbuf, "%d", chptr->mode.limit);
	}
    }
    if (*chptr->mode.key) {
	*mbuf++ = 'k';
	if (IsMember(cptr, chptr) || IsServer(cptr) || IsULine(cptr))
	    strcat(pbuf, chptr->mode.key);
    }
    *mbuf++ = '\0';
    return;
}

static void send_ban_list(aClient *cptr, aChannel *chptr)
{
    aBan *bp;
    char *cp;
    int count = 0, send = 0;

    cp = modebuf + strlen(modebuf);

    if (*parabuf)		/* mode +l or +k xx */
	count = 1;

    for (bp = chptr->banlist; bp; bp = bp->next) {
	if (strlen(parabuf) + strlen(bp->banstr) + 10 < (size_t) MODEBUFLEN) {
	    if (*parabuf)
		strcat(parabuf, " ");
	    strcat(parabuf, bp->banstr);
	    count++;
	    *cp++ = 'b';
	    *cp = '\0';
	} else if (*parabuf)
	    send = 1;

	if (count == MAXMODEPARAMS)
	    send = 1;

	if (send) {
	    sendto_one(cptr, ":%s MODE %s %s %s", me.name, chptr->chname,
		       modebuf, parabuf);
	    send = 0;
	    *parabuf = '\0';
	    cp = modebuf;
	    *cp++ = '+';
	    if (count != MAXMODEPARAMS) {
		strcpy(parabuf, bp->banstr);
		*cp++ = 'b';
		count = 1;
	    } else
		count = 0;
	    *cp = '\0';
	}
    }
}

void send_banex_list(aClient *cptr, aChannel *chptr)
{
    aBanEx *bp;
    char *cp;
    int count = 0, send = 0;

    cp = modebuf + strlen(modebuf);

    if (*parabuf)		/* mode +l or +k xx */
	count = 1;

    for (bp = chptr->banexlist; bp; bp = bp->next) {
	if (strlen(parabuf) + strlen(bp->banexstr) + 10 < (size_t) MODEBUFLEN) {
	    if (*parabuf)
		strcat(parabuf, " ");
	    strcat(parabuf, bp->banexstr);
	    count++;
	    *cp++ = 'e';
	    *cp = '\0';
	} else if (*parabuf)
	    send = 1;

	if (count == MAXMODEPARAMS)
	    send = 1;

	if (send) {
	    sendto_one(cptr, ":%s MODE %s %s %s", me.name, chptr->chname,
		       modebuf, parabuf);
	    send = 0;
	    *parabuf = '\0';
	    cp = modebuf;
	    *cp++ = '+';
	    if (count != MAXMODEPARAMS) {
		strcpy(parabuf, bp->banexstr);
		*cp++ = 'e';
		count = 1;
	    } else
		count = 0;
	    *cp = '\0';
	}
    }
}

void send_invites_list(aClient *cptr, aChannel *chptr)
{
    anInvite *bp;
    char *cp;
    int count = 0, send = 0;

    cp = modebuf + strlen(modebuf);

    if (*parabuf)		/* mode +l or +k xx */
	count = 1;

    for (bp = chptr->invitelist; bp; bp = bp->next) {
	if (strlen(parabuf) + strlen(bp->invitestr) + 10 < (size_t) MODEBUFLEN) {
	    if (*parabuf)
		strcat(parabuf, " ");
	    strcat(parabuf, bp->invitestr);
	    count++;
	    *cp++ = 'I';
	    *cp = '\0';
	} else if (*parabuf)
	    send = 1;

	if (count == MAXMODEPARAMS)
	    send = 1;

	if (send) {
	    sendto_one(cptr, ":%s MODE %s %s %s", me.name, chptr->chname,
		       modebuf, parabuf);
	    send = 0;
	    *parabuf = '\0';
	    cp = modebuf;
	    *cp++ = '+';
	    if (count != MAXMODEPARAMS) {
		strcpy(parabuf, bp->invitestr);
		*cp++ = 'I';
		count = 1;
	    } else
		count = 0;
	    *cp = '\0';
	}
    }
}

/*
 * send "cptr" a full list of the modes for channel chptr. 
 */
void send_channel_modes(aClient *cptr, aChannel *chptr)
{
    chanMember *l;
    /* *anop = NULL, *skip = NULL; */
    int n = 0;
    char *t;

    if (*chptr->chname != '#')
	return;

    *modebuf = *parabuf = '\0';
    channel_modes(cptr, modebuf, parabuf, chptr);

    if (IsSSJoin(cptr))
	ircsprintf(buf, ":%s SJOIN %ld %s %s %s :", me.name,
		   chptr->channelts, chptr->chname, modebuf, parabuf);
    else
	ircsprintf(buf, ":%s SJOIN %ld %ld %s %s %s :", me.name,
		   chptr->channelts, chptr->channelts, chptr->chname, modebuf,
		   parabuf);
    t = buf + strlen(buf);

    l = chptr->members;
    /*
     *   for (l = chptr->members; l; l = l->next) {
     *      if (IsAnon(l->cptr)) {
     *          skip = l;
     *          break;
     *      }
     *
     *      if (l->flags & MODE_CHANOP) {
     *          anop = l;
     *          break;
     *      }
     *
     *   }
     *
     *   l = NULL;
     *   for (;;) {
     *
     *      if (anop) {
     *          l = anop;
     *          anop = NULL;
     *      } else {
     *          if (l == NULL || l == skip)
     *              l = chptr->members;
     *          else
     *              l = l->next;
     *          if (l && l == skip)
     *              l = l->next;
     *          if (l == NULL)
     *              break;
     *      }
     */
    do {
	if (l == NULL)
	    break;
	if (IsAnon(l->cptr)) {
	    l = l->next;
	    continue;
	}
	if (l->flags & MODE_OWNER)
	    *t++ = '.';
	if (l->flags & MODE_PROTECT)
	    *t++ = '~';
	if (l->flags & MODE_CHANOP)
	    *t++ = '@';
	if (l->flags & MODE_VOICE)
	    *t++ = '+';
	strcpy(t, l->cptr->name);
	t += strlen(t);
	*t++ = ' ';
	n++;
	if (t - buf > BUFSIZE - 80) {
	    *t++ = '\0';
	    if (t[-1] == ' ')
		t[-1] = '\0';
	    sendto_one(cptr, "%s", buf);
	    if (IsSSJoin(cptr))
		sprintf(buf, ":%s SJOIN %ld %s 0 :", me.name, chptr->channelts,
			chptr->chname);
	    else
		sprintf(buf, ":%s SJOIN %ld %ld %s 0 :", me.name,
			chptr->channelts, chptr->channelts, chptr->chname);
	    t = buf + strlen(buf);
	    n = 0;
	}
	l = l->next;
    } while (l != NULL);

    if (n) {
	*t++ = '\0';
	if (t[-1] == ' ')
	    t[-1] = '\0';
	sendto_one(cptr, "%s", buf);
    }
    *parabuf = '\0';
    *modebuf = '+';
    modebuf[1] = '\0';
    send_ban_list(cptr, chptr);
    if (modebuf[1] || *parabuf)
	sendto_one(cptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf,
		   parabuf);
    *parabuf = '\0';
    *modebuf = '+';
    modebuf[1] = '\0';
    send_banex_list(cptr, chptr);
    if (modebuf[1] || *parabuf)
	sendto_one(cptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf,
		   parabuf);
    *parabuf = '\0';
    *modebuf = '+';
    modebuf[1] = '\0';
    send_invites_list(cptr, chptr);
    if (modebuf[1] || *parabuf)
	sendto_one(cptr, ":%s MODE %s %s %s", me.name, chptr->chname, modebuf,
		   parabuf);
}

/*
 * m_mode parv[0] - sender parv[1] - channel 
 */

int dont_send_ts_with_mode;

int m_mode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int mcount = 0, chanop = 0;
    aChannel *chptr;

    /*
     * Now, try to find the channel in question 
     */
    if (parc > 1) {
	chptr = find_channel(parv[1], NullChn);
	if (chptr == NullChn)
	    return m_umode(cptr, sptr, parc, parv);
    } else {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODE");
	return 0;
    }

    if (!check_channelname(sptr, (unsigned char *) parv[1]))
	return 0;

    if (is_chan_op(sptr, chptr))
	chanop = 1;
    else if ((IsULine(sptr)) ||
	     (IsAnOper(sptr) && !MyClient(sptr)) ||
	     (IsServer(sptr) && chptr->channelts != 0))
	chanop = 2;		/* extra speshul access */

    if (IsChanAnon(chptr))
	chanop = 0;

    if (parc < 3) {
	*modebuf = *parabuf = '\0';
	modebuf[1] = '\0';
	channel_modes(sptr, modebuf, parabuf, chptr);
	sendto_one(sptr, rpl_str(RPL_CHANNELMODEIS), me.name, parv[0],
		   chptr->chname, modebuf, parabuf);
	sendto_one(sptr, rpl_str(RPL_CREATIONTIME), me.name, parv[0],
		   chptr->chname, chptr->channelts);
	return 0;
    }
    mcount =
	set_mode(cptr, sptr, chptr, chanop, parc - 2, parv + 2, modebuf,
		 parabuf);
    if (strlen(modebuf) > (size_t) 1)
	switch (mcount) {
	    case 0:
		break;
	    case -1:
		if (MyClient(sptr))
		    sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name,
			       parv[0], chptr->chname);
		else
		    ircstp->is_fake++;
		break;
	    default:
		sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s",
				       parv[0], chptr->chname, modebuf,
				       parabuf);
		sendto_match_servs(chptr, cptr, ":%s MODE %s %s %s", parv[0],
				   chptr->chname, modebuf, parabuf);
		sendto_service(SERVICE_SEE_MODES, SERVICE_SEE_PREFIX,
			       ":%s MODE %s %s %s", parv[0], chptr->chname,
			       modebuf, parabuf);
		if (cptr->user)
		    sendto_service(SERVICE_SEE_MODES | SERVICE_SEE_PREFIX, 0,
				   ":%s!%s@%s MODE %s %s %s", parv[0],
				   cptr->user->username, cptr->user->host,
				   chptr->chname, modebuf, parabuf);
	}
    return 0;
}

/*
 * the old set_mode was pissing me off with it's disgusting
 * * hackery, so I rewrote it.  Hope this works. }:> --wd
 */
int
set_mode(aClient *cptr, aClient *sptr, aChannel *chptr, int level,
	 int parc, char *parv[], char *mbuf, char *pbuf)
{
#define SM_ERR_NOPRIVS 0x0001	/* is not an op */
#define SM_ERR_MOREPARMS 0x0002	/* needs more parameters */
#define SM_ERR_RESTRICTED 0x0004	/* not allowed to op others or be op'd */
#define SM_ERR_NOTOPER    0x0008	/* not an irc op */
#define SM_MAXMODES 6

    /*
     * this macro appends to pbuf 
     */
#define ADD_PARA(p) pptr = p; if(pidx) pbuf[pidx++] = ' '; while(*pptr) pbuf[pidx++] = *pptr++;

    static int flags[] = {
	MODE_PRIVATE, 'p', MODE_SECRET, 's',
	MODE_MODERATED, 'm', MODE_NOPRIVMSGS, 'n',
	MODE_TOPICLIMIT, 't', MODE_REGONLY, 'R',
	MODE_INVITEONLY, 'i', MODE_NOCOLOR, 'c',
	MODE_OPERONLY, 'O', MODE_ANONYMOUS, 'q',
	0x0, 0x0
    };

    Link *lp;			/* for walking lists */
    chanMember *cm;		/* for walking channel member lists */
    aBan *bp;			/* for walking banlists */
    anInvite *iip;		/* -TimeMr14C */
    aBanEx *ep;			/* -TimeMr14C */
    char *modes = parv[0];	/* user's idea of mode changes */
    int args;			/* counter for what argument we're on */
    int banlsent = 0;		/* Only list bans once in a command. */
    int invilsent = 0;		/* -TimeMr14C */
    int banexlsent = 0;		/* -TimeMr14C */
    char change = '+';		/* by default we + things... */
    int errors = 0;		/* errors returned, set with bitflags so we
				 * only return them once */

    /*
     * from remote servers, ungodly numbers of modes can be sent, but
     * * from local users only SM_MAXMODES are allowed 
     */
    int maxmodes = ((IsServer(sptr) || IsULine(sptr)) ? 512 : SM_MAXMODES);
    int nmodes = 0;		/* how many modes we've set so far */
    aClient *who = NULL;	/* who we're doing a mode for */
    int chasing = 0;
    int i = 0;
    char moreparmsstr[] = "MODE   ";
    char nuhbuf[NICKLEN + USERLEN + HOSTLEN + 6];	/* for bans */
    char tmp[16];		/* temporary buffer */
    int pidx = 0;		/* index into pbuf */
    char *pptr;			/* temporary paramater pointer */
    char *morig = mbuf;		/* beginning of mbuf */

    /*
     * :cptr-name MODE chptr->chname [MBUF] [PBUF] (buflen - 3 max and NULL) 
     */
    int prelen = strlen(cptr->name) + strlen(chptr->chname) + 16;

    args = 1;
    if (parc < 1)
	return 0;

    *mbuf++ = '+';		/* add the plus, even if they don't */
    /*
     * go through once to clean the user's mode string so we can
     * * have a simple parser run through it...
     */
    while (*modes && (nmodes < maxmodes)) {
	switch (*modes) {
	    case '+':
		if (*(mbuf - 1) == '-') {
		    *(mbuf - 1) = '+';	/* change it around now */
		    change = '+';
		    break;
		} else if (change == '+')	/* we're still doing a +, we *
						 * * * * * * * * * * * * * * * * don't care */
		    break;
		change = *modes;
		*mbuf++ = '+';
		break;

	    case '-':
		if (*(mbuf - 1) == '+') {
		    *(mbuf - 1) = '-';	/* change it around now */
		    change = '-';
		    break;
		} else if (change == '-')
		    break;	/* we're still doing a -, we don't care */
		change = *modes;
		*mbuf++ = '-';
		break;

	    case 'O':
		if (level < 1 || !IsOper(sptr)) {
		    errors |= SM_ERR_NOTOPER;
		    break;
		} else {
		    if (change == '+')
			chptr->mode.mode |= MODE_OPERONLY;
		    else
			chptr->mode.mode &= ~MODE_OPERONLY;
		    *mbuf++ = *modes;
		    nmodes++;
		}
		break;
	    case 'u':
	    case 'a':
	    case 'o':
	    case 'v':
		if (level < 1) {
		    errors |= SM_ERR_NOPRIVS;
		    break;
		}
		if (parv[args] == NULL) {
		    /*
		     * silently drop the spare +o/v's 
		     */
		    break;
		}
		who = find_chasing(sptr, parv[args], &chasing);
		cm = find_user_member(chptr->members, who);
		if (cm == NULL) {
		    sendto_one(sptr, err_str(ERR_USERNOTINCHANNEL),
			       me.name, cptr->name, parv[args], chptr->chname);
		    /*
		     * swallow the arg 
		     */
		    args++;
		    break;
		}
		if ((level < 2) && ((*modes == 'u') || (*modes == 'a'))) {
		    errors |= SM_ERR_NOPRIVS;
		    args++;
		    break;
		}
		if (!IsPrivileged(cptr) && is_chan_owner(who, chptr)
		    && (change == '-') && (*modes == 'o')) {
		    errors |= SM_ERR_NOPRIVS;
		    args++;
		    break;
		}
		/*
		 * if we're going to overflow our mode buffer,
		 * * drop the change instead 
		 */
		if ((prelen + (mbuf - morig) + pidx + NICKLEN + 1) >
		    REALMODEBUFLEN) {
		    args++;
		    break;
		}
		/*
		 * if we have the user, set them +/-[vo] 
		 * 
		 * if(change=='+')
		 *  cm->flags|=(*modes=='o' ? CHFL_CHANOP : CHFL_VOICE);
		 * else
		 *  cm->flags&=~((*modes=='o' ? CHFL_CHANOP : CHFL_VOICE));
		 */

		if (change == '+') {
		    if (*modes == 'u') {
			cm->flags |= CHFL_OWNER;
		    } else if (*modes == 'a') {
			cm->flags |= CHFL_PROTECT;
		    } else if (*modes == 'o') {
			cm->flags |= CHFL_CHANOP;
		    } else if (*modes == 'v') {
			cm->flags |= CHFL_VOICE;
		    }
		} else {
		    if (*modes == 'u') {
			cm->flags &= ~CHFL_OWNER;
		    } else if (*modes == 'a') {
			cm->flags &= ~CHFL_PROTECT;
		    } else if (*modes == 'o') {
			cm->flags &= ~CHFL_CHANOP;
		    } else if (*modes == 'v') {
			cm->flags &= ~CHFL_VOICE;
		    }
		}
		/*
		 * we've decided their mode was okay, cool 
		 */
		*mbuf++ = *modes;
		ADD_PARA(cm->cptr->name)
		    args++;
		nmodes++;
		if (IsServer(sptr) && *modes == 'o' && change == '+') {
		    chptr->channelts = 0;
		    sendto_ops("Server %s setting +o and blasting TS on %s",
			       sptr->name, chptr->chname);
		}
		break;

	    case 'b':
		/*
		 * if the user has no more arguments, then they just want
		 * * to see the bans, okay, cool. 
		 */
		if (level < 1 && parv[args] != NULL) {
		    errors |= SM_ERR_NOPRIVS;
		    break;
		}
		/*
		 * show them the bans, woowoo 
		 */
		if (parv[args] == NULL) {
		    if (banlsent)
			break;	/* Send only once */
		    for (bp = chptr->banlist; bp; bp = bp->next)
			sendto_one(sptr, rpl_str(RPL_BANLIST), me.name,
				   cptr->name, chptr->chname, bp->banstr,
				   bp->who, bp->when);
		    sendto_one(cptr, rpl_str(RPL_ENDOFBANLIST), me.name,
			       cptr->name, chptr->chname);
		    banlsent = 1;
		    break;	/* we don't pass this along, either.. */
		}

		/*
		 * do not allow : in bans, or a null ban 
		 */
		if (*parv[args] == ':' || *parv[args] == '\0') {
		    args++;
		    break;
		}

		/*
		 * make a 'pretty' ban mask here, then try and set it 
		 */
		/*
		 * okay kids, let's do this again.
		 * * the buffer returned by pretty_mask is from 
		 * * make_nick_user_host. This buffer is eaten by add/del banid.
		 * * Thus, some poor schmuck gets himself on the banlist. Fixed. - lucas 
		 */

		strcpy(nuhbuf, collapse(pretty_mask(parv[args])));
		parv[args] = nuhbuf;
		/*
		 * if we're going to overflow our mode buffer,
		 * * drop the change instead 
		 */
		if ((prelen + (mbuf - morig) + pidx + strlen(nuhbuf) + 1) >
		    REALMODEBUFLEN) {
		    args++;
		    break;
		}
		/*
		 * if we can't add or delete (depending) the ban, change is
		 * * worthless anyhow 
		 */

		if (!(change == '+' && !add_banid(sptr, chptr, parv[args]))
		    && !(change == '-' && !del_banid(chptr, parv[args]))) {
		    args++;
		    break;
		}

		*mbuf++ = 'b';
		ADD_PARA(parv[args])
		    args++;
		nmodes++;
		break;

	    case 'I':
		if (level < 1 && parv[args] != NULL) {
		    errors |= SM_ERR_NOPRIVS;
		    break;
		}

		if (parv[args] == NULL) {
		    if (invilsent)
			break;
		    for (iip = chptr->invitelist; iip; iip = iip->next)
			sendto_one(sptr, rpl_str(RPL_INVITELIST), me.name,
				   cptr->name, chptr->chname, iip->invitestr,
				   iip->who, iip->when);
		    sendto_one(cptr, rpl_str(RPL_ENDOFINVITELIST), me.name,
			       cptr->name, chptr->chname);
		    invilsent = 1;
		    break;
		}

		if (*parv[args] == ':' || *parv[args] == '\0') {
		    args++;
		    break;
		}

		strcpy(nuhbuf, collapse(pretty_mask(parv[args])));
		parv[args] = nuhbuf;
		if ((prelen + (mbuf - morig) + pidx + strlen(nuhbuf) + 1) >
		    REALMODEBUFLEN) {
		    args++;
		    break;
		}

		if (!(change == '+' && !add_inviteid(sptr, chptr, parv[args]))
		    && !(change == '-' && !del_inviteid(chptr, parv[args]))) {
		    args++;
		    break;
		}

		*mbuf++ = 'I';
		ADD_PARA(parv[args])
		    args++;
		nmodes++;
		break;

	    case 'e':
		if (level < 1 && parv[args] != NULL) {
		    errors |= SM_ERR_NOPRIVS;
		    break;
		}

		if (parv[args] == NULL) {
		    if (banexlsent)
			break;
		    for (ep = chptr->banexlist; ep; ep = ep->next)
			sendto_one(sptr, rpl_str(RPL_EXCEPTIONLIST), me.name,
				   cptr->name, chptr->chname, ep->banexstr,
				   ep->who, ep->when);
		    sendto_one(cptr, rpl_str(RPL_ENDOFEXCEPLIST), me.name,
			       cptr->name, chptr->chname);
		    banexlsent = 1;
		    break;
		}

		if (*parv[args] == ':' || *parv[args] == '\0') {
		    args++;
		    break;
		}

		strcpy(nuhbuf, collapse(pretty_mask(parv[args])));
		parv[args] = nuhbuf;
		if ((prelen + (mbuf - morig) + pidx + strlen(nuhbuf) + 1) >
		    REALMODEBUFLEN) {
		    args++;
		    break;
		}

		if (!(change == '+' && !add_banexid(sptr, chptr, parv[args]))
		    && !(change == '-' && !del_banexid(chptr, parv[args]))) {
		    args++;
		    break;
		}
		*mbuf++ = 'e';
		ADD_PARA(parv[args])
		    args++;
		nmodes++;
		break;

	    case 'l':
		if (level < 1) {
		    errors |= SM_ERR_NOPRIVS;
		    break;
		}

		/*
		 * if it's a -, just change the flag, we have no arguments 
		 */
		if (change == '-') {
		    if ((prelen + (mbuf - morig) + pidx + 1) > REALMODEBUFLEN)
			break;
		    *mbuf++ = 'l';
		    chptr->mode.mode &= ~MODE_LIMIT;
		    chptr->mode.limit = 0;
		    nmodes++;
		    break;
		} else {
		    if (parv[args] == NULL) {
			errors |= SM_ERR_MOREPARMS;
			break;
		    }

		    /*
		     * if we're going to overflow our mode buffer,
		     * * drop the change instead 
		     */
		    if ((prelen + (mbuf - morig) + pidx + 16) > REALMODEBUFLEN) {
			args++;
			break;
		    }

		    i = atoi(parv[args]);

		    /*
		     * toss out invalid modes 
		     */
		    if (i < 1) {
			args++;
			break;
		    }
		    ircsprintf(tmp, "%d", i);
		    chptr->mode.limit = i;
		    chptr->mode.mode |= MODE_LIMIT;
		    *mbuf++ = 'l';
		    ADD_PARA(tmp);
		    args++;
		    nmodes++;
		    break;
		}

	    case 'k':
		if (level < 1) {
		    errors |= SM_ERR_NOPRIVS;
		    break;
		}
		if (parv[args] == NULL)
		    break;

		/*
		 * do not allow keys to start with :! ack! - lucas 
		 */
		/*
		 * another ack: don't let people set null keys! 
		 */
		if (*parv[args] == ':' || *parv[args] == '\0' ||
		    strchr(parv[args], ' ')) {
		    args++;
		    break;
		}

		/*
		 * Do not let *'s in keys in preperation for key hiding
		 * * - Raist
		 */

		if (strchr(parv[args], '*') != NULL) {
		    args++;
		    break;
		}
		/*
		 * if we're going to overflow our mode buffer,
		 * * drop the change instead 
		 */
		if ((prelen + (mbuf - morig) + pidx + KEYLEN + 2) >
		    REALMODEBUFLEN) {
		    args++;
		    break;
		}

		/*
		 * if they're an op, they can futz with the key in
		 * * any manner they like, we're not picky 
		 */
		if (change == '+') {
		    parv[args][KEYLEN] = '\0';
		    strncpy(chptr->mode.key, parv[args], KEYLEN);
		    ADD_PARA(chptr->mode.key)
		} else {
		    ADD_PARA(chptr->mode.key)
			* chptr->mode.key = '\0';
		}
		*mbuf++ = 'k';
		args++;
		nmodes++;
		break;

	    case 'r':
		if (!IsServer(sptr) && !IsULine(sptr)) {
		    sendto_one(sptr, err_str(ERR_ONLYSERVERSCANCHANGE),
			       me.name, cptr->name, chptr->chname);
		    break;
		} else {
		    if ((prelen + (mbuf - morig) + pidx + 1) > REALMODEBUFLEN)
			break;

		    if (change == '+')
			chptr->mode.mode |= MODE_REGISTERED;
		    else
			chptr->mode.mode &= ~MODE_REGISTERED;
		}
		*mbuf++ = 'r';
		nmodes++;
		break;

	    case 'q':
		if (!IsMe(sptr)) {
		    sendto_one(sptr, err_str(ERR_ONLYSERVERSCANCHANGE),
			       me.name, cptr->name, chptr->chname);
		    break;
		} else {
		    if ((prelen + (mbuf - morig) + pidx + 1) > REALMODEBUFLEN)
			break;

		    if (change == '+')
			chptr->mode.mode |= MODE_ANONYMOUS;
		    else
			chptr->mode.mode &= ~MODE_ANONYMOUS;
		}
		*mbuf++ = 'q';
		nmodes++;
		break;

	    case 'i':
		if (level < 1) {
		    errors |= SM_ERR_NOPRIVS;
		    break;
		}
		if (change == '-')
		    while ((lp = chptr->invites))
			del_invite(lp->value.cptr, chptr);
		/*
		 * fall through to default case 
		 */

	    default:
		/*
		 * phew, no more tough modes. }:>, the rest are all covered in one step 
		 * * with the above array 
		 */
		if (level < 1) {
		    errors |= SM_ERR_NOPRIVS;
		    break;
		}
		for (i = 1; flags[i] != 0x0; i += 2) {
		    if ((prelen + (mbuf - morig) + pidx + 1) > REALMODEBUFLEN)
			break;

		    if (*modes == (char) flags[i]) {
			if (change == '+')
			    chptr->mode.mode |= flags[i - 1];
			else
			    chptr->mode.mode &= ~flags[i - 1];
			*mbuf++ = *modes;
			nmodes++;
			break;
		    }
		}
		/*
		 * unknown mode.. 
		 */
		if (flags[i] == 0x0) {
		    /*
		     * we still spew lots of unknown mode bits...
		     */
		    /*
		     * but only to our own clients, silently ignore bogosity
		     * * from other servers... 
		     */
		    if (MyClient(sptr))
			sendto_one(sptr, err_str(ERR_UNKNOWNMODE), me.name,
				   sptr->name, *modes);

		}
		break;
	}

	/*
	 * spit out more parameters error here 
	 */
	if (errors & SM_ERR_MOREPARMS && MyClient(sptr)) {
	    moreparmsstr[5] = change;
	    moreparmsstr[6] = *modes;
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, sptr->name,
		       moreparmsstr);
	    errors &= ~SM_ERR_MOREPARMS;	/* oops, kill it in this case */
	}
	modes++;
    }
    /*
     * clean up the end of the string... 
     */
    if (*(mbuf - 1) == '+' || *(mbuf - 1) == '-')
	*(mbuf - 1) = '\0';
    else
	*mbuf = '\0';
    pbuf[pidx] = '\0';
    if (MyClient(sptr)) {
	if (errors & SM_ERR_NOTOPER)
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, sptr->name);
	if (errors & SM_ERR_NOPRIVS)
	    sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, sptr->name,
		       chptr->chname);
	if (errors & SM_ERR_RESTRICTED)
	    sendto_one(sptr,
		       ":%s NOTICE %s :*** Notice -- You are restricted and cannot chanop others",
		       me.name, sptr->name);
    }
    /*
     * all done! 
     */
    return nmodes;
#undef ADD_PARA
}

void kill_ban_list(aClient *cptr, aChannel *chptr)
{
    chanMember *cm;
    aBan *bp, *bpn;
    char *cp;
    int count = 0, send = 0;

    cp = modebuf;
    *cp++ = '-';
    *cp = '\0';

    *parabuf = '\0';

    for (bp = chptr->banlist; bp; bp = bp->next) {
	if (strlen(parabuf) + strlen(bp->banstr) + 10 < (size_t) MODEBUFLEN) {
	    if (*parabuf)
		strcat(parabuf, " ");
	    strcat(parabuf, bp->banstr);
	    count++;
	    *cp++ = 'b';
	    *cp = '\0';
	} else if (*parabuf)
	    send = 1;

	if (count == MAXMODEPARAMS)
	    send = 1;

	if (send) {
	    sendto_channel_butserv(chptr, &me, ":%s MODE %s %s %s",
				   cptr->name, chptr->chname, modebuf, parabuf);
	    send = 0;
	    *parabuf = '\0';
	    cp = modebuf;
	    *cp++ = '-';
	    if (count != MAXMODEPARAMS) {
		strcpy(parabuf, bp->banstr);
		*cp++ = 'b';
		count = 1;
	    } else
		count = 0;
	    *cp = '\0';
	}
    }

    if (*parabuf) {
	sendto_channel_butserv(chptr, &me, ":%s MODE %s %s %s",
			       cptr->name, chptr->chname, modebuf, parabuf);
    }

    /*
     * physically destroy channel ban list 
     */

    bp = chptr->banlist;
    while (bp) {
	bpn = bp->next;
	MyFree(bp->banstr);
	MyFree(bp->who);
	MyFree(bp);
	bp = bpn;
    }

    chptr->banlist = NULL;

    /*
     * reset bquiet on all channel members 
     */
    for (cm = chptr->members; cm; cm = cm->next) {
	if (MyConnect(cm->cptr))
	    cm->bans = 0;
    }
}

/*   
 * m_umode() added 15/10/91 By Darren Reed. parv[0] - sender parv[1] -
 * username to change mode for parv[2] - modes to change
 */
int m_umode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int flag;
    int *s;
    char **p, *m;
    aClient *acptr;
    int what, setflags;
    int badflag = NO;		/*
				 * 
				 * * Only send one bad flag notice
				 * * * -Dianora
				 */
    what = MODE_ADD;

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "MODE");
	return 0;
    }

    if (!(acptr = find_person(parv[1], NULL))) {
	if (MyConnect(sptr))
	    sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0],
		       parv[1]);
	return 0;
    }

    if ((IsServer(sptr) || (sptr != acptr) || (acptr->from != sptr->from))) {
	if (!IsServer(cptr))
	    sendto_one(sptr, err_str(ERR_USERSDONTMATCH), me.name, parv[0]);
	return 0;
    }
    if (parc < 3) {
	m = buf;
	*m++ = '+';
	for (s = user_modes; (flag = *s) && (m - buf < BUFSIZE - 4); s += 2)
	    if (sptr->umode & flag)
		*m++ = (char) (*(s + 1));
	*m = '\0';
	sendto_one(sptr, rpl_str(RPL_UMODEIS), me.name, parv[0], buf);
	return 0;
    }

    /*
     * find flags already set for user
     */
    setflags = 0;
    for (s = user_modes; (flag = *s); s += 2)
	if (sptr->umode & flag)
	    setflags |= flag;
    /*
     * parse mode change string(s)
     */
    for (p = &parv[2]; p && *p; p++)
	for (m = *p; *m; m++)
	    switch (*m) {
		case '+':
		    what = MODE_ADD;
		    break;
		case '-':
		    what = MODE_DEL;
		    break;
		    /*
		     * we may not get these, but they shouldnt be in
		     * default
		     */
		case ' ':
		case '\r':
		case '\n':
		case '\t':
		    break;
		case 'r':
		    break;	/* users can't set themselves +r! */
		case 'A':
		    /*   
		     * set auto +a if user is setting +A
		     */
		    if (MyClient(sptr) && (what == MODE_ADD))
			sptr->umode |= UMODE_a;
		default:
		    for (s = user_modes; (flag = *s); s += 2)
			if (*m == (char) (*(s + 1))) {
			    if (what == MODE_ADD)
				sptr->umode |= flag;
			    else
				sptr->umode &= ~flag;
			    break;
			}
		    if (flag == 0 && MyConnect(sptr))
			badflag = YES;
		    break;
	    }

    if (badflag)
	sendto_one(sptr, err_str(ERR_UMODEUNKNOWNFLAG), me.name, parv[0]);

    if (!(setflags & UMODE_o) && IsOper(sptr) && !IsServer(cptr))
	ClearOper(sptr);

    if (!(setflags & UMODE_O) && IsLocOp(sptr) && !IsServer(cptr))
	sptr->umode &= ~UMODE_O;

    if (!(setflags & UMODE_O) && IsLocOp(sptr) && !IsServer(cptr))
	sptr->umode &= ~UMODE_O;

    if ((setflags & (UMODE_o | UMODE_O)) && !IsAnOper(sptr) && MyConnect(sptr)) {
	det_confs_butmask(sptr, CONF_CLIENT & ~CONF_OPS);
	sptr->sendqlen = get_sendq(sptr);
	sptr->oflag = 0;
    }

    if (!(setflags & (UMODE_o | UMODE_O)) && IsAnOper(sptr)) {
	sendto_service(SERVICE_SEE_OPERS, SERVICE_SEE_PREFIX, ":%s OPER :%s",
		       parv[0], parv[0]);
	sendto_service(SERVICE_SEE_OPERS | SERVICE_SEE_PREFIX, 0,
		       ":%s!%s@%s OPER :%s", parv[0], sptr->user->username,
		       sptr->user->host, parv[0]);
	Count.oper++;
    }

    if ((setflags & (UMODE_o | UMODE_O)) && !IsAnOper(sptr)) {
	Count.oper--;
	if (MyConnect(sptr))
	    delfrom_fdlist(sptr->fd, &oper_fdlist);
    }

    /*
     * if (!(setflags & UMODE_b) && (!IsOper(sptr) && !IsLocOp(sptr))
     *    && !IsServer(cptr))
     *       sptr->umode &= ~UMODE_b;
     *       
     * if (!(setflags & UMODE_R) && (!IsOper(sptr) && !IsLocOp(sptr))
     *    && !IsServer(cptr))
     *      sptr->umode &= ~UMODE_R;
     *      
     * if (!(setflags & UMODE_e) && (!IsOper(sptr) && !IsLocOp(sptr))
     *    && !IsServer(cptr))
     *      sptr->umode &= ~UMODE_e;
     *      
     * if (!(setflags & UMODE_y) && (!IsOper(sptr) && !IsLocOp(sptr))
     *    && !IsServer(cptr))
     *       sptr->umode &= ~UMODE_y;
     *       
     * if (!(setflags & UMODE_d) && (!IsOper(sptr) && !IsLocOp(sptr))
     *    && !IsServer(cptr))
     *       sptr->umode &= ~UMODE_d;
     */

    if (!(setflags & UMODE_n) && (!IsOper(sptr) && !IsLocOp(sptr))
	&& !IsServer(cptr))
	sptr->umode &= ~UMODE_n;

    if (!(setflags & UMODE_h) && (!IsOper(sptr) && !IsLocOp(sptr))
	&& !IsServer(cptr))
	sptr->umode &= ~UMODE_h;

    if (!(setflags & UMODE_i) && IsInvisible(sptr))
	Count.invisi++;
    if ((setflags & UMODE_i) && !IsInvisible(sptr))
	Count.invisi--;
    /*  
     * compare new flags with old flags and send string which will cause
     * servers to update correctly.
     */
    if (!IsAnOper(sptr) && !IsServer(sptr)) {
	if (IsAdmin(sptr))
	    ClearAdmin(sptr);
	if (IsSAdmin(sptr))
	    ClearSAdmin(sptr);
	// if (IsUmodef(sptr)) ClearUmodef(sptr);
	// if (IsUmodec(sptr)) ClearUmodec(sptr);
	// if (IsUmodey(sptr)) ClearUmodey(sptr);
	// if (IsUmoded(sptr)) ClearUmoded(sptr);
	// if (IsUmodeb(sptr)) ClearUmodeb(sptr);
	if (IsUmoden(sptr))
	    ClearUmoden(sptr);
	// if (IsUmodem(sptr)) ClearUmodem(sptr);
	// if (IsRouting(sptr)) ClearRouting(sptr);
	// if (IsUmodee(sptr)) ClearUmodee(sptr);  
    }
    if (MyClient(sptr)) {
	if (IsAdmin(sptr) && !OPIsAdmin(sptr))
	    ClearAdmin(sptr);
	if (IsSAdmin(sptr) && !OPIsSAdmin(sptr))
	    ClearSAdmin(sptr);
	// if (IsUmodef(sptr) && !OPSetUModeF(sptr)) ClearUmodef(sptr);
	// if (IsUmodec(sptr) && !OPSetUModeC(sptr)) ClearUmodec(sptr);
	// if (IsUmodey(sptr) && !OPSetUModeY(sptr)) ClearUmodey(sptr);
	// if (IsUmoded(sptr) && !OPSetUModeD(sptr)) ClearUmoded(sptr);
	// if (IsUmodeb(sptr) && !OPSetUModeB(sptr)) ClearUmodeb(sptr);
	// if (IsRouting(sptr) && !OPIsRStaff(sptr)) ClearRouting(sptr);
    }
    send_umode_out(cptr, sptr, setflags);

    return 0;
}

void send_umode(aClient *cptr, aClient *sptr, int old, int sendmask,
		char *umode_buf)
{
    int *s, flag;
    char *m;
    int what = MODE_NULL;

    /*
     * build a string in umode_buf to represent the change in the user's
     * mode between the new (sptr->flag) and 'old'.
     */
    m = umode_buf;
    *m = '\0';
    for (s = user_modes; (flag = *s); s += 2) {
	if (MyClient(sptr) && !(flag & sendmask))
	    continue;
	if ((flag & old) && !(sptr->umode & flag)) {
	    if (what == MODE_DEL)
		*m++ = *(s + 1);
	    else {
		what = MODE_DEL;
		*m++ = '-';
		*m++ = *(s + 1);
	    }
	} else if (!(flag & old) && (sptr->umode & flag)) {
	    if (what == MODE_ADD)
		*m++ = *(s + 1);
	    else {
		what = MODE_ADD;
		*m++ = '+';
		*m++ = *(s + 1);
	    }
	}
    }
    *m = '\0';
    if (*umode_buf && cptr)
	sendto_one(cptr, ":%s MODE %s :%s", sptr->name, sptr->name, umode_buf);
}

/*
 * added Sat Jul 25 07:30:42 EST 1992
 * extra argument evenTS added to send to TS servers or not -orabidoo
 * extra argument evenTS no longer needed with TS only th+hybrid server
 * -Dianora
 */
void send_umode_out(aClient *cptr, aClient *sptr, int old)
{
    int i, j;
    aClient *acptr;
    fdlist fdl = serv_fdlist;

    send_umode(NULL, sptr, old, SEND_UMODES, buf);
    /*
     * Cycling through serv_fdlist here should be MUCH faster than
     * looping through every client looking for servers. -ThemBones
     */

    for (i = fdl.entry[j = 1]; j <= fdl.last_entry; i = fdl.entry[++j])
	if ((acptr = local[i]) && (acptr != cptr) && (acptr != sptr) && (*buf))
	    sendto_one(acptr, ":%s MODE %s :%s", sptr->name, sptr->name, buf);

    sendto_service(SERVICE_SEE_UMODES, SERVICE_SEE_PREFIX,
		   ":%s MODE %s :%s", sptr->name, sptr->name, buf);
    if (sptr->user)
	sendto_service(SERVICE_SEE_UMODES | SERVICE_SEE_PREFIX, 0,
		       ":%s!%s@%s MODE %s :%s",
		       sptr->name, sptr->user->username, sptr->user->host,
		       sptr->name, buf);

    if (cptr && MyClient(cptr))
	send_umode(cptr, sptr, old, ALL_UMODES, buf);
}
