/************************************************************************
 *   IRC - Internet Relay Chat, src/m_whois.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <sys/stat.h>
#include <utmp.h>
#include <fcntl.h>
#include "h.h"

static char buf[BUFSIZE];
extern Link *find_channel_link(Link *, aChannel *);

/*
 * * m_whois *        parv[0] = sender prefix *       parv[1] = nickname
 * masklist
 */
int m_whois(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    static anUser UnknownUser = {
	NULL,			/* channel */
	NULL,			/* invited */
	NULL,			/* away */
	0,			/* last */
	0,			/* joined */
	"<Unknown>",		/* user */
	"<Unknown>",		/* host */
#ifdef USE_FAKEHOST
	"<unknown>",		/* fakehost */
#endif
	"<Unknown>",		/* server */
	0,			/* servicestamp */
	NULL			/* silenced */
    };

    Link *lp;
    anUser *user;
    aClient *acptr, *a2cptr;
    aChannel *chptr;
    char *nick, *tmp, *name;
    char *p = NULL;
    int found, len, mlen;
    char out[HOSTLEN];		/* TimeMr14C */

    if (parc < 2) {
	sendto_one(sptr, err_str(ERR_NONICKNAMEGIVEN), me.name, parv[0]);
	return 0;
    }

    if (parc > 2) {
	if (hunt_server(cptr, sptr, ":%s WHOIS %s :%s", 1, parc, parv) !=
	    HUNTED_ISME)
	    return 0;
	parv[1] = parv[2];
    }

    for (tmp = parv[1]; (nick = strtoken(&p, tmp, ",")); tmp = NULL) {
    int invis, member, showchan;

	found = 0;

	(void) collapse(nick);
	acptr = hash_find_client(nick, (aClient *) NULL);
	if (!acptr || !IsPerson(acptr)) {
	    sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], nick);
	    continue;
	}

	user = acptr->user ? acptr->user : &UnknownUser;
	name = (!*acptr->name) ? "?" : acptr->name;
	invis = IsInvisible(acptr);
	member = (user->channel) ? 1 : 0;

	a2cptr = find_server(user->server, NULL);

	if (IsAnOper(acptr) && IsUmodeh(acptr)) {
	    sendto_one(acptr,
		       ":%s NOTICE %s :%s [%s@%s] did a /whois on you ",
		       me.name, acptr->name, sptr->name, sptr->user->username,
		       sptr->user->host);
	}
	/*
	 * -TimeMr14C: 06.01.2001 
	 */

	sendto_one(sptr, rpl_str(RPL_WHOISUSER), me.name, parv[0], name,
		   user->username,
#ifdef USE_FAKEHOST
		   IsFake(acptr) ? acptr->user->fakehost : acptr->user->host,
#else
		   acptr->user->host,
#endif
		   acptr->info);

#ifdef USE_FAKEHOST
	if ((IsAnOper(sptr) || (sptr == acptr)) && IsFake(acptr))
	    sendto_one(sptr, rpl_str(RPL_REALHOST), me.name, parv[0], name,
		       user->host);
#endif

	if (user->real_oper_host && IsOper(sptr)) {
	    sendto_one(sptr, rpl_str(RPL_WHOISACTUALLY),
		       me.name, sptr->name, name,
		       user->real_oper_username, user->real_oper_host,
		       user->real_oper_ip);
	}

	mlen = strlen(me.name) + strlen(parv[0]) + 6 + strlen(name);
	for (len = 0, *buf = '\0', lp = user->channel; lp; lp = lp->next) {
	    chptr = lp->value.chptr;
	    if (IsChanAnon(chptr))
		continue;
	    showchan = ShowChannel(sptr, chptr);
	    if ((showchan && !(acptr->umode & UMODE_w)) || IsOper(sptr)
		|| (acptr == sptr)) {
		if (len + strlen(chptr->chname)
		    > (size_t) BUFSIZE - 4 - mlen) {
		    sendto_one(sptr,
			       ":%s %d %s %s :%s", me.name, RPL_WHOISCHANNELS,
			       parv[0], name, buf);
		    *buf = '\0';
		    len = 0;
		}
		if (!showchan)
		    *(buf + len++) = '%';
		if (is_chan_op(acptr, chptr))
		    *(buf + len++) = '@';
		else if (has_voice(acptr, chptr))
		    *(buf + len++) = '+';
		if (len)
		    *(buf + len) = '\0';
		(void) strcpy(buf + len, chptr->chname);
		len += strlen(chptr->chname);
		(void) strcat(buf + len, " ");
		len++;
	    }
	}
	if (buf[0] != '\0')
	    sendto_one(sptr, rpl_str(RPL_WHOISCHANNELS), me.name, parv[0], name,
		       buf);

	sendto_one(sptr, rpl_str(RPL_WHOISSERVER),
		   me.name, parv[0], name,
		   (WillHide(a2cptr) ? stealth_server(user->server, out) :
		    user->server), a2cptr ? a2cptr->info : "*Not On This Net*");
	if (IsRegNick(acptr))
	    sendto_one(sptr, rpl_str(RPL_WHOISREGNICK), me.name, parv[0], name);
	if (user->away)
	    sendto_one(sptr, rpl_str(RPL_AWAY), me.name, parv[0], name,
		       user->away);

	buf[0] = '\0';
	if (IsAnOper(acptr))
	    strcat(buf, "an IRC Operator");
	if (IsAdmin(acptr))
	    strcat(buf, " - Server Administrator");
	else if (IsSAdmin(acptr))
	    strcat(buf, " - Services Administrator");
	if (buf[0])
	    sendto_one(sptr, rpl_str(RPL_WHOISOPERATOR), me.name, parv[0], name,
		       buf);

	if (acptr->user && MyConnect(acptr))
	    sendto_one(sptr, rpl_str(RPL_WHOISIDLE),
		       me.name, parv[0], name, timeofday - user->last,
		       acptr->firsttime);

	continue;
	if (!found)
	    sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, parv[0], nick);
	if (p)
	    p[-1] = ',';
    }
    sendto_one(sptr, rpl_str(RPL_ENDOFWHOIS), me.name, parv[0], parv[1]);

    return 0;
}
