/************************************************************************
 *   IRC - Internet Relay Chat, src/s_flood.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <sys/stat.h>
#include <utmp.h>
#include <fcntl.h>
#include "throttle.h"
#include "h.h"

#ifdef FLUD
#include "blalloc.h"
int flud_num = FLUD_NUM;
int flud_time = FLUD_TIME;
int flud_block = FLUD_BLOCK;
extern BlockHeap *free_fludbots;
extern BlockHeap *free_Links;
extern void outofmemory();

void announce_fluder(aClient *, aClient *, aChannel *, int);
struct fludbot *remove_fluder_reference(struct fludbot **, aClient *);
Link *remove_fludee_reference(Link **, void *);
int check_for_fludblock(aClient *, aClient *, aChannel *, int);
int check_for_flud(aClient *, aClient *, aChannel *, int);
void free_fluders(aClient *, aChannel *);
void free_fludees(aClient *);
#endif
extern int killnoident;
int botwarn(char *, char *, char *, char *);

/**
 ** botreject(host)
 **   Reject a bot based on a fake hostname...
 **           -Taner
 **/
int botreject(char *host)
{
    /*
     * Eggdrop Bots:   "USER foo 1 1 :foo" Vlad, Com, joh Bots:
     * "USER foo null null :foo" Annoy/OJNKbots: "user foo . . :foo"
     * (disabled) Spambots that are based on OJNK: "user foo x x :foo"
     */
#undef CHECK_FOR_ANNOYOJNK
    if (!strcmp(host, "1"))
	return 1;
    if (!strcmp(host, "null"))
	return 2;
    if (!strcmp(host, "x"))
	return 3;
#ifdef CHECK_FOR_ANNOYOJNK
    if (!strcmp(host, "."))
	return 4;
#endif
    return 0;
}

/**
 ** botwarn(host, nick, user, realhost)
 **     Warn that a bot MAY be connecting (added by ThemBones)
 **/
int botwarn(char *host, char *nick, char *user, char *realhost)
{
    /*
     * Eggdrop Bots:      "USER foo 1 1 :foo" Vlad, Com, joh Bots:
     * "USER foo null null :foo" Annoy/OJNKbots:    "user foo . . :foo"
     * (disabled)
     */
#undef CHECK_FOR_ANNOYOJNK
    if (!strcmp(host, "1"))
	sendto_realops_lev(CCONN_LEV, "Possible Eggdrop: %s (%s@%s)", nick,
			   user, realhost);
    if (!strcmp(host, "null"))
	sendto_realops_lev(CCONN_LEV, "Possible ComBot: %s (%s@%s)", nick, user,
			   realhost);
    if (!strcmp(host, "x"))
	sendto_realops_lev(CCONN_LEV, "Possible SpamBot: %s (%s@%s)", nick,
			   user, realhost);
#ifdef CHECK_FOR_ANNOYOJNK
    if (!strcmp(host, "."))
	sendto_realops_lev(CCONN_LEV, "Possible AnnoyBot: %s (%s@%s)", nick,
			   user, realhost);
#endif
    return 0;
}

/*
 * Shadowfax's FLUD code
 */

#ifdef FLUD

void announce_fluder(aClient *fluder,	/* 
					 * fluder, client being fluded
					 */
		     aClient *cptr, aChannel *chptr,	/* 
							 * channel being fluded
							 */
		     int type)
{				/* 
				 * for future use
				 */
    char *fludee;

    if (cptr)
	fludee = cptr->name;
    else
	fludee = chptr->chname;

    sendto_ops_lev(FLOOD_LEV, "Flooder %s [%s@%s] on %s target: %s",
		   fluder->name, fluder->user->username,
		   fluder->user->host, fluder->user->server, fludee);
}

/*
 * This is really just a "convenience" function.  I can only keep three
 * or * four levels of pointer dereferencing straight in my head.  This
 * remove * an entry in a fluders list.  Use this when working on a
 * fludees list :)
 */
struct fludbot *remove_fluder_reference(struct fludbot **fluders,
					aClient *fluder)
{
    struct fludbot *current, *prev, *next;

    prev = NULL;
    current = *fluders;
    while (current) {
	next = current->next;
	if (current->fluder == fluder) {
	    if (prev)
		prev->next = next;
	    else
		*fluders = next;

	    BlockHeapFree(free_fludbots, current);
	} else
	    prev = current;
	current = next;
    }

    return (*fluders);
}

/*
 * Another function to unravel my mind.
 */
Link *remove_fludee_reference(Link **fludees, void *fludee)
{
    Link *current, *prev, *next;

    prev = NULL;
    current = *fludees;
    while (current) {
	next = current->next;
	if (current->value.cptr == (aClient *) fludee) {
	    if (prev)
		prev->next = next;
	    else
		*fludees = next;

	    BlockHeapFree(free_Links, current);
	} else
	    prev = current;
	current = next;
    }

    return (*fludees);
}

int check_for_fludblock(aClient *fluder,	/* 
						 * fluder being fluded
						 */
			aClient *cptr,	/* 
					 * client being fluded
					 */
			aChannel *chptr,	/* 
						 * channel being fluded
						 */
			int type)
{

    time_t now;
    int blocking;

    /*
     * If it's disabled, we don't need to process all of this
     */
    if (flud_block == 0)
	return 0;

    /*
     * It's either got to be a client or a channel being fluded
     */
    if ((cptr == NULL) && (chptr == NULL))
	return 0;

    if (cptr && !MyFludConnect(cptr)) {
	sendto_ops("check_for_fludblock() called for non-local client");
	return 0;
    }

    /*
     * Are we blocking fluds at this moment?
     */
    time(&now);
    if (cptr)
	blocking = (cptr->fludblock > (now - flud_block));
    else
	blocking = (chptr->fludblock > (now - flud_block));

    return (blocking);
}

int check_for_flud(aClient *fluder,
		   /*
		    * fluder, client being fluded
		    */
		   aClient *cptr, aChannel *chptr,	/* 
							 * channel being fluded
							 */
		   int type)
{				/* 
				 * for future use
				 */
    time_t now;
    struct fludbot *current, *prev, *next;
    int blocking, count, found;
    Link *newfludee;

    /*
     * If it's disabled, we don't need to process all of this
     */
    if (flud_block == 0)
	return 0;

    /*
     * It's either got to be a client or a channel being fluded
     */
    if ((cptr == NULL) && (chptr == NULL))
	return 0;

    if (cptr && !MyFludConnect(cptr)) {
	sendto_ops("check_for_flud() called for non-local client");
	return 0;
    }

    /*
     * Are we blocking fluds at this moment?
     */
    time(&now);
    if (cptr)
	blocking = (cptr->fludblock > (now - flud_block));
    else
	blocking = (chptr->fludblock > (now - flud_block));

    /*
     * Collect the Garbage
     */
    if (!blocking) {
	if (cptr)
	    current = cptr->fluders;
	else
	    current = chptr->fluders;
	prev = NULL;
	while (current) {
	    next = current->next;
	    if (current->last_msg < (now - flud_time)) {
		if (cptr)
		    remove_fludee_reference(&current->fluder->fludees,
					    (void *) cptr);
		else
		    remove_fludee_reference(&current->fluder->fludees,
					    (void *) chptr);

		if (prev)
		    prev->next = current->next;
		else if (cptr)
		    cptr->fluders = current->next;
		else
		    chptr->fluders = current->next;
		BlockHeapFree(free_fludbots, current);
	    } else
		prev = current;
	    current = next;
	}
    }
    /*
     * Find or create the structure for the fluder, and update the
     * counter * and last_msg members.  Also make a running total count
     */
    if (cptr)
	current = cptr->fluders;
    else
	current = chptr->fluders;
    count = found = 0;
    while (current) {
	if (current->fluder == fluder) {
	    current->last_msg = now;
	    current->count++;
	    found = 1;
	}
	if (current->first_msg < (now - flud_time))
	    count++;
	else
	    count += current->count;
	current = current->next;
    }
    if (!found) {
    if ((current = BlockHeapALLOC(free_fludbots, struct fludbot)) != NULL) {
	current->fluder = fluder;
	current->count = 1;
	current->first_msg = now;
	current->last_msg = now;
	if (cptr) {
	    current->next = cptr->fluders;
	    cptr->fluders = current;
	} else {
	    current->next = chptr->fluders;
	    chptr->fluders = current;
	}

	count++;

	if ((newfludee = BlockHeapALLOC(free_Links, Link)) != NULL) {
	    if (cptr) {
		newfludee->flags = 0;
		newfludee->value.cptr = cptr;
	    } else {
		newfludee->flags = 1;
		newfludee->value.chptr = chptr;
	    }
	    newfludee->next = fluder->fludees;
	    fluder->fludees = newfludee;
	} else
	    outofmemory();
	/*
	 * If we are already blocking now, we should go ahead * and
	 * announce the new arrival
	 */
	if (blocking)
	    announce_fluder(fluder, cptr, chptr, type);
    } else
	outofmemory();
    }
    /*
     * Okay, if we are not blocking, we need to decide if it's time to *
     * begin doing so.  We already have a count of messages received in *
     * the last flud_time seconds
     */
    if (!blocking && (count > flud_num)) {
	blocking = 1;
	ircstp->is_flud++;
	/*
	 * if we are going to say anything to the fludee, now is the *
	 * time to mention it to them.
	 */
	if (cptr)
	    sendto_one(cptr,
		       ":%s NOTICE %s :*** Notice -- Server flood protection activated for %s",
		       me.name, cptr->name, cptr->name);
	else
	    sendto_channel_butserv(chptr, &me,
				   ":%s NOTICE %s :*** Notice -- Server flood protection activated for %s",
				   me.name, chptr->chname, chptr->chname);
	/*
	 * Here we should go back through the existing list of * fluders
	 * and announce that they were part of the game as * well.
	 */
	if (cptr)
	    current = cptr->fluders;
	else
	    current = chptr->fluders;
	while (current) {
	    announce_fluder(current->fluder, cptr, chptr, type);
	    current = current->next;
	}
    }
    /*
     * update blocking timestamp, since we received a/another CTCP
     * message
     */
    if (blocking) {
	if (cptr)
	    cptr->fludblock = now;
	else
	    chptr->fludblock = now;
    }

    return (blocking);
}

void free_fluders(aClient *cptr, aChannel *chptr)
{
    struct fludbot *fluders, *next;

    if ((cptr == NULL) && (chptr == NULL)) {
	sendto_ops("free_fluders(NULL, NULL)");
	return;
    }

    if (cptr && !MyFludConnect(cptr))
	return;

    if (cptr)
	fluders = cptr->fluders;
    else
	fluders = chptr->fluders;

    while (fluders) {
	next = fluders->next;

	if (cptr)
	    remove_fludee_reference(&fluders->fluder->fludees, (void *) cptr);
	else
	    remove_fludee_reference(&fluders->fluder->fludees, (void *) chptr);

	BlockHeapFree(free_fludbots, fluders);
	fluders = next;
    }
}

void free_fludees(aClient *badguy)
{
    Link *fludees, *next;

    if (badguy == NULL) {
	sendto_ops("free_fludees(NULL)");
	return;
    }
    fludees = badguy->fludees;
    while (fludees) {
	next = fludees->next;

	if (fludees->flags)
	    remove_fluder_reference(&fludees->value.chptr->fluders, badguy);
	else {
	    if (!MyFludConnect(fludees->value.cptr))
		sendto_ops("free_fludees() encountered non-local client");
	    else
		remove_fluder_reference(&fludees->value.cptr->fluders, badguy);
	}

	BlockHeapFree(free_Links, fludees);
	fludees = next;
    }
}
#endif

/*
 * Shadowfax's server side, anti flood code 
 */
#ifdef FLUD
extern int flud_num;
extern int flud_time;
extern int flud_block;
#endif

#ifdef ANTI_SPAMBOT
extern int spam_num;
extern int spam_time;
#endif

#ifdef NO_CHANOPS_WHEN_SPLIT
extern int server_split_recovery_time;
#endif

/*
 * m_set - set options while running
 */
int m_set(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *command;

    if (!MyClient(sptr) || !IsOper(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    if (parc > 1) {
	command = parv[1];
	if (!strncasecmp(command, "MAX", 3)) {
	    if (parc > 2) {
    int new_value = atoi(parv[2]);

		if (new_value > MASTER_MAX) {
		    sendto_one(sptr,
			       ":%s NOTICE %s :You cannot set MAXCLIENTS to > MASTER_MAX (%d)",
			       me.name, parv[0], MASTER_MAX);
		    return 0;
		}
		if (new_value < 32) {
		    sendto_one(sptr,
			       ":%s NOTICE %s :You cannot set MAXCLIENTS to < 32 (%d:%d)",
			       me.name, parv[0], MAXCLIENTS, highest_fd);
		    return 0;
		}
		MAXCLIENTS = new_value;
		sendto_one(sptr,
			   ":%s NOTICE %s :NEW MAXCLIENTS = %d (Current = %d)",
			   me.name, parv[0], MAXCLIENTS, Count.local);
		sendto_realops
		    ("%s!%s@%s set new MAXCLIENTS to %d (%d current)",
		     parv[0], sptr->user->username, sptr->sockhost, MAXCLIENTS,
		     Count.local);
		return 0;
	    }
	    sendto_one(sptr, ":%s NOTICE %s :Current Maxclients = %d (%d)",
		       me.name, parv[0], MAXCLIENTS, Count.local);
	    return 0;
	}
#ifdef FLUD
	else if (!strncasecmp(command, "FLUDNUM", 7)) {
	    if (parc > 2) {
    int newval = atoi(parv[2]);

		if (newval <= 0) {
		    sendto_one(sptr, ":%s NOTICE %s :flud NUM must be > 0",
			       me.name, parv[0]);
		    return 0;
		}
		flud_num = newval;
		sendto_ops("%s has changed flud NUM to %i", parv[0], flud_num);
		sendto_one(sptr, ":%s NOTICE %s :flud NUM is now set to %i",
			   me.name, parv[0], flud_num);
		return 0;
	    } else {
		sendto_one(sptr, ":%s NOTICE %s :flud NUM is currently %i",
			   me.name, parv[0], flud_num);
		return 0;
	    }
	} else if (!strncasecmp(command, "FLUDTIME", 8)) {
	    if (parc > 2) {
    int newval = atoi(parv[2]);

		if (newval <= 0) {
		    sendto_one(sptr, ":%s NOTICE %s :flud TIME must be > 0",
			       me.name, parv[0]);
		    return 0;
		}
		flud_time = newval;
		sendto_ops("%s has changed flud TIME to %i", parv[0],
			   flud_time);
		sendto_one(sptr, ":%s NOTICE %s :flud TIME is now set to %i",
			   me.name, parv[0], flud_time);
		return 0;
	    } else {
		sendto_one(sptr, ":%s NOTICE %s :flud TIME is currently %i",
			   me.name, parv[0], flud_time);
		return 0;
	    }
	} else if (!strncasecmp(command, "FLUDBLOCK", 9)) {
	    if (parc > 2) {
    int newval = atoi(parv[2]);

		if (newval < 0) {
		    sendto_one(sptr, ":%s NOTICE %s :flud BLOCK must be >= 0",
			       me.name, parv[0]);
		    return 0;
		}
		flud_block = newval;
		if (flud_block == 0) {
		    sendto_ops("%s has disabled flud detection/protection",
			       parv[0]);
		    sendto_one(sptr, ":%s NOTICE %s :flud detection disabled",
			       me.name, parv[0]);
		} else {
		    sendto_ops("%s has changed flud BLOCK to %i", parv[0],
			       flud_block);
		    sendto_one(sptr,
			       ":%s NOTICE %s :flud BLOCK is now set to %i",
			       me.name, parv[0], flud_block);
		}
		return 0;
	    } else {
		sendto_one(sptr,
			   ":%s NOTICE %s :flud BLOCK is currently %i",
			   me.name, parv[0], flud_block);
		return 0;
	    }
	}
#endif
#ifdef NO_CHANOPS_WHEN_SPLIT
	else if (!strncasecmp(command, "SPLITDELAY", 10)) {
	    if (parc > 2) {
    int newval = atoi(parv[2]);

		if (newval < 0) {
		    sendto_one(sptr, ":%s NOTICE %s :split delay must be >= 0",
			       me.name, parv[0]);
		    return 0;
		}
		if (newval > MAX_SERVER_SPLIT_RECOVERY_TIME)
		    newval = MAX_SERVER_SPLIT_RECOVERY_TIME;
		sendto_ops
		    ("%s has changed spam SERVER SPLIT RECOVERY TIME  to %i",
		     parv[0], newval);
		sendto_one(sptr,
			   ":%s NOTICE %s :SERVER SPLIT RECOVERY TIME is now set to %i",
			   me.name, parv[0], newval);
		server_split_recovery_time = (newval * 60);
		return 0;

	    } else {
		sendto_one(sptr,
			   ":%s NOTICE %s :SERVER SPLIT RECOVERY TIME is currenly %i",
			   me.name, parv[0], server_split_recovery_time / 60);
	    }
	}
#endif
#ifdef ANTI_SPAMBOT
	/*
	 * int spam_time = MIN_JOIN_LEAVE_TIME;
	 * * int spam_num = MAX_JOIN_LEAVE_COUNT;
	 */
	else if (!strncasecmp(command, "SPAMNUM", 7)) {
	    if (parc > 2) {
    int newval = atoi(parv[2]);

		if (newval <= 0) {
		    sendto_one(sptr, ":%s NOTICE %s :spam NUM must be > 0",
			       me.name, parv[0]);
		    return 0;
		}
		if (newval < MIN_SPAM_NUM)
		    spam_num = MIN_SPAM_NUM;
		else
		    spam_num = newval;
		sendto_ops("%s has changed spam NUM to %i", parv[0], spam_num);
		sendto_one(sptr, ":%s NOTICE %s :spam NUM is now set to %i",
			   me.name, parv[0], spam_num);
		return 0;
	    } else {
		sendto_one(sptr, ":%s NOTICE %s :spam NUM is currently %i",
			   me.name, parv[0], spam_num);
		return 0;
	    }
	} else if (!strncasecmp(command, "SPAMTIME", 8)) {
	    if (parc > 2) {
    int newval = atoi(parv[2]);

		if (newval <= 0) {
		    sendto_one(sptr, ":%s NOTICE %s :spam TIME must be > 0",
			       me.name, parv[0]);
		    return 0;
		}
		if (newval < MIN_SPAM_TIME)
		    spam_time = MIN_SPAM_TIME;
		else
		    spam_time = newval;
		sendto_ops("%s has changed spam TIME to %i", parv[0],
			   spam_time);
		sendto_one(sptr, ":%s NOTICE %s :SPAM TIME is now set to %i",
			   me.name, parv[0], spam_time);
		return 0;
	    } else {
		sendto_one(sptr, ":%s NOTICE %s :spam TIME is currently %i",
			   me.name, parv[0], spam_time);
		return 0;
	    }
	}
#endif
#ifdef IDENTD_COMPLAIN
	else if (!strncasecmp(command, "KILLNOIDENT", 3)) {
	    if (parc > 2) {

		if (!strncasecmp(parv[2], "ON", 2)) {
		    sendto_one(sptr,
			       ":%s NOTICE %s :The ircd will not allow users"
			       "without a correct ident reply to connect",
			       me.name, parv[0]);
		    sendto_realops("%s!%s@%s activated KILLNOIDENT", parv[0],
				   sptr->user->username, sptr->sockhost);
		    killnoident = 1;
		} else if (!strncasecmp(parv[2], "OFF", 3)) {
		    sendto_one(sptr,
			       ":%s NOTICE %s :The ircd will now allow users"
			       "without a correct ident reply to connect",
			       me.name, parv[0]);
		    sendto_realops("%s!%s@%s deactivated KILLNOIDENT", parv[0],
				   sptr->user->username, sptr->sockhost);
		    killnoident = 0;
		} else {
		    sendto_one(sptr,
			       ":%s NOTICE %s :The correct setting is either"
			       "\002ON\002 or \002OFF\002",
			       me.name, sptr->name);
		}

		return 0;
	    } else {

		sendto_one(sptr, ":%s NOTICE %s :KILLNOIDENT is now %s",
			   me.name, parv[0],
			   (killnoident ? "\002ACTIVE\002" :
			    "\002INACTIVE\002"));
		return 0;
	    }
	}
#endif
	else if (!strncasecmp(command, "THROTTLE", 8)) {
    char *changed = NULL;
    char *to = NULL;
	    /* several values available:
	     * ENABLE [on|off] to enable the code
	     * COUNT [n] to set a max count, must be > 1
	     * TIME [n] to set a max time before expiry, must be > 5
	     * BANTIME [n] to set a time for the throttle bans to expire
	     * HASH [n] to set the size of the hash table, must be bigger than
	     *          the default */

	    /* only handle individual settings if parc > 3 (they're actually
	     * changing stuff) */
	    if (parc > 3) {
		if (!strcasecmp(parv[2], "ENABLE")) {
		    changed = "ENABLE";
		    if (tolower(*parv[3]) == 'y' || !strcasecmp(parv[3], "on")) {
			throttle_enable = 1;
			to = "ON";
		    } else if (tolower(*parv[3]) == 'n' ||
			       !strcasecmp(parv[3], "off")) {
			throttle_enable = 0;
			to = "OFF";
		    }
		} else if (!strcasecmp(parv[2], "COUNT")) {
    int cnt;
		    changed = "COUNT";
		    cnt = atoi(parv[3]);
		    if (cnt > 1) {
			throttle_tcount = cnt;
			to = parv[3];
		    }
		} else if (!strcasecmp(parv[2], "TIME")) {
    int cnt;
		    changed = "TIME";
		    cnt = atoi(parv[3]);
		    if (cnt >= 5) {
			throttle_ttime = cnt;
			to = parv[3];
		    }
		} else if (!strcasecmp(parv[2], "BANTIME")) {
    int cnt;
		    changed = "BANTIME";
		    cnt = atoi(parv[3]);
		    if (cnt >= 30) {
			throttle_ztime = cnt;
			to = parv[3];
		    }
		} else if (!strcasecmp(parv[2], "HASH")) {
    int cnt;
		    changed = "HASH";
		    cnt = atoi(parv[3]);
		    if (cnt >= THROTTLE_HASHSIZE) {
			throttle_resize(cnt);
			to = parv[3];
		    }
		}

		if (to != NULL) {
		    sendto_ops("%s has changed throttle %s to %s", parv[0],
			       changed, to);
		    sendto_one(sptr, ":%s NOTICE %s :set throttle %s to %s",
			       me.name, parv[0], changed, to);
		}
	    } else {
		/* report various things, we cannot easily get the hash size, so
		 * leave that alone. */
		sendto_one(sptr, ":%s NOTICE %s :THROTTLE %s", me.name, parv[0],
			   throttle_enable ? "enabled" : "disabled");
		sendto_one(sptr, ":%s NOTICE %s :THROTTLE COUNT=%d", me.name,
			   parv[0], throttle_tcount);
		sendto_one(sptr, ":%s NOTICE %s :THROTTLE TIME=%d", me.name,
			   parv[0], throttle_ttime);
		sendto_one(sptr, ":%s NOTICE %s :THROTTLE BANTIME=%d", me.name,
			   parv[0], throttle_ztime);
	    }
	}
#ifdef DEBUGMODE
	else if (!strncasecmp(command, "DEBUG", 3)) {
	    if (parc > 2) {
		if (atoi(parv[2])) {
		    DupString(debugmode, parv[2]);
		    debuglevel = atoi(parv[2]);
		    sendto_one(sptr,
			       ":%s NOTICE %s :DEBUGMODE set to level \002%s\002",
			       me.name, parv[0], debugmode);
		    Debug((DEBUG_FATAL, "Debug: Changed to Level: %d at %s",
			   debuglevel, myctime(time(NULL))));

		} else if (!strncasecmp(parv[2], "OFF", 3)) {
		    Debug((DEBUG_FATAL, "Debug: Stopped at %s",
			   myctime(time(NULL))));
		    debugmode = "0";
		    debuglevel = -1;
		    sendto_one(sptr,
			       ":%s NOTICE %s :DEBUGMODE is now \002OFF\002",
			       me.name, parv[0]);
		} else {
		    sendto_one(sptr,
			       ":%s NOTICE %s :Possible settings are "
			       "\0021\002 to \0029\002 or \002OFF\002",
			       me.name, parv[0]);
		}
		return 0;
	    } else {
		sendto_one(sptr,
			   ":%s NOTICE %s :DEBUGMODE is set to \002%s\002",
			   me.name, parv[0],
			   atoi(debugmode) ? debugmode : "OFF");
		return 0;
	    }
	}
#endif
    } else {
	sendto_one(sptr, ":%s NOTICE %s :Options: MAX", me.name, parv[0]);
	sendto_one(sptr, ":%s NOTICE %s :Options: THROTTLE "
		   "<ENABLE|COUNT|TIME|BANTIME|HASH> [setting]", me.name,
		   parv[0]);
#ifdef IDENTD_COMPLAIN
	sendto_one(sptr, ":%s NOTICE %s :Options: KILLNOIDENT", me.name,
		   parv[0]);
#endif

#ifdef FLUD
	sendto_one(sptr, ":%s NOTICE %s :Options: FLUDNUM, FLUDTIME, FLUDBLOCK",
		   me.name, parv[0]);
#endif

#ifdef ANTI_SPAMBOT
	sendto_one(sptr, ":%s NOTICE %s :Options: SPAMNUM, SPAMTIME", me.name,
		   parv[0]);
#endif

#ifdef DEBUGMODE
	sendto_one(sptr, ":%s NOTICE %s :Options: DEBUG", me.name, parv[0]);
#endif

#ifdef NO_CHANOPS_WHEN_SPLIT
	sendto_one(sptr, ":%s NOTICE %s :Options: SPLITDELAY", me.name,
		   parv[0]);
#endif
    }
    return 0;
}
