/************************************************************************
 *   IRC - Internet Relay Chat, src/s_user.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: s_user.c,v 1.31 2002/02/08 19:04:53 tr-ircd Exp $ */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <sys/stat.h>
#include <utmp.h>
#include <fcntl.h>
#include "h.h"
#include "throttle.h"
#ifdef FLUD
#include "blalloc.h"
#endif
#if defined( HAVE_STRING_H)
#include <string.h>
#else
#include <strings.h>
#endif

#ifdef USE_FAKEHOST
int do_user(char *, aClient *, aClient *, char *, char *,
	    char *, char *, unsigned long, unsigned int, char *);
#else
int do_user(char *, aClient *, aClient *, char *, char *,
	    char *, unsigned long, unsigned int, char *);
#endif

extern char motd_last_changed_date[];
extern int send_motd(aClient *, aClient *, int, char **);
extern void send_topic_burst(aClient *);
int check_oper_can_mask(aClient *, char *, char *, char **);
extern void outofmemory(void);
extern int killnoident;
#ifdef MAXBUFFERS
extern void reset_sock_opts();
extern int send_lusers(aClient *, aClient *, int, char **);

#endif
extern int lifesux;

#ifdef WINGATE_NOTICE
extern char ProxyMonURL[TOPICLEN + 1];
extern char ProxyMonHost[HOSTLEN + 1];
#endif

int user_modes[] = { UMODE_o, 'o',	/* ops, realops */
    UMODE_O, 'O',		/* ops */
    UMODE_i, 'i',		/* invisible on who/names */
    UMODE_w, 'w',		/* wallops */
    /*
     * UMODE_s, 's', 
     * UMODE_c, 'c', 
     */
    UMODE_r, 'r',		/* registered and identified nickname */
    UMODE_R, 'R',		/* listens to registered and identified
				 * nicknames */
    /*
     * UMODE_k, 'k', 
     * UMODE_y, 'y', 
     * UMODE_d, 'd', 
     * UMODE_e, 'e', 
     */
    UMODE_g, 'g',		/* globops */
    /*
     * UMODE_b, 'b', 
     */
    UMODE_a, 'a',		/* nick is services admin */
    UMODE_A, 'A',		/* nick is server admin */
    /*
     * UMODE_f, 'f', 
     */
    UMODE_n, 'n',		/* gnotice */
    /*
     * UMODE_m, 'm', 
     */
    UMODE_h, 'h',		/* nick is ready to help : stats P */
#ifdef USE_FAKEHOST
    UMODE_x, 'x',		/* nick has hidden hostname */
#endif
    0, 0
};

/*
 * internally defined functions 
 */
int botreject(char *);
extern int find_fline(aClient *);
extern Link *find_channel_link(Link *, aChannel *);
#ifdef ANTI_SPAMBOT
int spam_time = MIN_JOIN_LEAVE_TIME;
int spam_num = MAX_JOIN_LEAVE_COUNT;
#endif

/*
 * * m_functions execute protocol messages on this server: *
 * 
 *      cptr    is always NON-NULL, pointing to a *LOCAL* client *
 * tructure (with an open socket connected!). This *
 * es the physical socket where the message *           originated (or
 * which caused the m_function to be *          executed--some
 * m_functions may call others...). *
 * 
 *      sptr    is the source of the message, defined by the *
 * refix part of the message if present. If not *               or
 * prefix not found, then sptr==cptr. *
 * 
 *              (!IsServer(cptr)) => (cptr == sptr), because *
 * refixes are taken *only* from servers... *
 * 
 *              (IsServer(cptr)) *                      (sptr == cptr)
 * => the message didn't *                      have the prefix. *
 * 
 *                      (sptr != cptr && IsServer(sptr) means *
 * he prefix specified servername. (?) *
 * 
 *                      (sptr != cptr && !IsServer(sptr) means *
 * hat message originated from a remote *                       user
 * (not local). *
 * 
 *              combining *
 * 
 *              (!IsServer(sptr)) means that, sptr can safely *
 * aken as defining the target structure of the *               message
 * in this server. *
 * 
 *      *Always* true (if 'parse' and others are working correct): *
 * 
 *      1)      sptr->from == cptr  (note: cptr->from == cptr) *
 * 
 *      2)      MyConnect(sptr) <=> sptr == cptr (e.g. sptr *
 * annot* be a local connection, unless it's *          actually
 * cptr!). [MyConnect(x) should probably *              be defined as
 * (x == x->from) --msa ] *
 * 
 *      parc    number of variable parameter strings (if zero, *
 * arv is allowed to be NULL) *
 * 
 *      parv    a NULL terminated list of parameter pointers, *
 * 
 *                      parv[0], sender (prefix string), if not present *
 * his points to an empty string. *
 * arc-1] *                             pointers to additional
 * parameters *                 parv[parc] == NULL, *always* *
 * 
 *              note:   it is guaranteed that parv[0]..parv[parc-1] are
 * all *                        non-NULL pointers.
 */

/*
 * * next_client *    Local function to find the next matching
 * client. The search * can be continued from the specified client
 * entry. Normal *      usage loop is: *
 * 
 *      for (x = client; x = next_client(x,mask); x = x->next) *
 * andleMatchingClient; *
 * 
 */
aClient *next_client(aClient *next,	/* 
					 * First client to check 
					 */
		     char *ch)
{				/* 
				 * search string (may include wilds) 
				 */
    aClient *tmp = next;

    next = find_client(ch, tmp);
    if (tmp && tmp->prev == next)
	return ((aClient *) NULL);

    if (next != tmp)
	return next;
    for (; next; next = next->next) {
	if (!match(ch, next->name))
	    break;
    }
    return next;
}

/*
 * this slow version needs to be used for hostmasks *sigh * 
 */

aClient *next_client_double(aClient *next, char *ch)
{
    aClient *tmp = next;

    next = find_client(ch, tmp);
    if (tmp && tmp->prev == next)
	return NULL;
    if (next != tmp)
	return next;
    for (; next; next = next->next) {
	if (!match(ch, next->name) || !match(next->name, ch))
	    break;
    }
    return next;
}

/*
 * * register_user *  This function is called when both NICK and USER
 * messages *   have been accepted for the client, in whatever order.
 * Only *       after this, is the USER message propagated. *
 * 
 *      NICK's must be propagated at once when received, although *
 * it would be better to delay them too until full info is *
 * available. Doing it is not so simple though, would have *    to
 * implement the following: *
 * 
 *      (actually it has been implemented already for a while)
 * -orabidoo *
 * 
 *      1) user telnets in and gives only "NICK foobar" and waits *
 * 2) another user far away logs in normally with the nick *
 * "foobar" (quite legal, as this server didn't propagate *        it). *
 * 3) now this server gets nick "foobar" from outside, but *       has
 * already the same defined locally. Current server *      would just
 * issue "KILL foobar" to clean out dups. But, *           this is not
 * fair. It should actually request another *      nick from local user
 * or kill him/her...
 */

int register_user(aClient *cptr, aClient *sptr, char *nick, char *username)
{
    aClient *nsptr;
    aConfItem *aconf = NULL, *pwaconf = NULL, *bconf = NULL;
    char *parv[3];
    static char ubuf[12];

#ifdef USE_FAKEHOST
    char newhostarr[HOSTLEN];
    char *q;
#endif
    char *p;
    short oldstatus = sptr->status;
    anUser *user = sptr->user;

#ifdef SHORT_MOTD
    aMotd *smotd;
#endif
    int i, dots;
    int bad_dns;

#ifdef ANTI_SPAMBOT
    char spamchar = '\0';
#endif
    char tmpstr2[512];

    user->last = timeofday;
    parv[0] = sptr->name;
    parv[1] = parv[2] = NULL;

    if (MyConnect(sptr)) {
	p = inetntoa((char *) &sptr->ip);
#ifdef USE_FAKEHOST
	q = inetntoa((char *) &sptr->ip);
#endif
	strncpyzt(sptr->hostip, p, HOSTIPLEN + 1);
	if ((i = check_client(sptr))) {
	    /*
	     * -2 is a socket error, already reported.
	     */
	    if (i != -2) {
		if (i == -4) {
		    ircstp->is_ref++;
		    return exit_client(cptr, sptr, &me,
				       "Too many connections from your hostname");
		} else if (i == -3)
		    sendto_realops_lev(SPY_LEV, "%s for %s [%s] ",
				       "I-line is full", get_client_host(sptr),
				       p);
		else
		    sendto_realops_lev(CCONN_LEV, "%s from %s [%s]",
				       "Unauthorized client connection",
				       get_client_host(sptr), p);
		ircstp->is_ref++;
		return exit_client(cptr, sptr, &me, i == -3 ?
				   "No more connections allowed in your connection class"
				   :
				   "You are not authorized to use this server");
	    } else
		return exit_client(cptr, sptr, &me, "Socket Error");
	}
#ifdef ANTI_SPAMBOT
	/*
	 * This appears to be broken 
	 * Check for single char in user->host -ThemBones 
	 */
	if (*(user->host + 1) == '\0')
	    spamchar = *user->host;
#endif

	strncpyzt(user->host, sptr->sockhost, HOSTLEN);

	dots = 0;
	p = user->host;
	bad_dns = NO;
	while (*p) {
	    if (!isalnum(*p)) {
#ifdef RFC1035_ANAL
		if ((*p != '-') && (*p != '.'))
#else
		if ((*p != '-') && (*p != '.') && (*p != '_')
		    && (*p != '/'))
#endif
		    bad_dns = YES;
	    }
	    if (*p == '.')
		dots++;
	    p++;
	}
	/*
	 * Check that the hostname has AT LEAST ONE dot (.) in it. If
	 * not, drop the client (spoofed host) -ThemBones
	 */
	if (!dots) {
	    sendto_realops("Invalid hostname for %s, dumping user %s",
			   sptr->hostip, sptr->name);
	    return exit_client(cptr, sptr, &me, "Invalid hostname");
	}
#ifdef USE_FAKEHOST
	q = user->host;
	strncpyzt(user->fakehost, calcmask(q, newhostarr), HOSTLEN);
#endif
	if (bad_dns) {
	    sendto_one(sptr,
		       ":%s NOTICE %s :*** Notice -- You have a bad character in your hostname",
		       me.name, cptr->name);
	    strcpy(user->host, sptr->hostip);
	    strcpy(sptr->sockhost, sptr->hostip);
	}

	pwaconf = sptr->confs->value.aconf;

	if (sptr->flags & FLAGS_DOID && !(sptr->flags & FLAGS_GOTID)) {
	    /*
	     * because username may point to user->username 
	     */
    char temp[USERLEN + 1];

	    strncpyzt(temp, username, USERLEN + 1);
	    *user->username = '~';
	    (void) strncpy(&user->username[1], temp, USERLEN);
	    user->username[USERLEN] = '\0';
#ifdef IDENTD_COMPLAIN
	    /*
	     * tell them to install identd -Taner 
	     */
	    sendto_one(sptr,
		       ":%s NOTICE %s :*** Notice -- It seems that you don't have identd installed on your host.",
		       me.name, cptr->name);
	    if (killnoident) {
		ircstp->is_ref++;
		return exit_client(cptr, sptr, &me, "Ident Required");
	    } else {
		sendto_one(sptr,
			   ":%s NOTICE %s :*** Notice -- If you wish to have your username show up without the ~ (tilde),",
			   me.name, cptr->name);
		sendto_one(sptr,
			   ":%s NOTICE %s :*** Notice -- then install identd.",
			   me.name, cptr->name);
	    }
#endif
	}
#ifndef FOLLOW_IDENT_RFC
	else if (sptr->flags & FLAGS_GOTID && *sptr->username != '-')
	    strncpyzt(user->username, sptr->username, USERLEN + 1);
#endif
	else
	    strncpyzt(user->username, username, USERLEN + 1);

	if (!BadPtr(pwaconf->passwd)) {
    char *tmpptr = strchr(sptr->passwd, ':');
    char tmppwd[PASSWDLEN + 1];

	    /*
	     * If there's a : in the password, fix it so after this function,
	     * sptr->passwd changes from:
	     * moo:cow:test:asdf
	     * to
	     * cow:test:asdf
	     */

	    if (tmpptr) {
		*tmpptr++ = '\0';
		strcpy(tmppwd, tmpptr);
	    }

	    if (!StrEq(sptr->passwd, pwaconf->passwd)) {
		ircstp->is_ref++;
		sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
		return exit_client(cptr, sptr, &me, "Bad Password");
	    }
	    if (tmpptr)
		strcpy(sptr->passwd, tmppwd);
	    else
		sptr->passwd[0] = '\0';
	}

	if ((aconf = find_kill(sptr)) && !(bconf = find_exclude(sptr))) {
    char *reason;
    char *ktype;
    int kline;

	    kline = (aconf->status == CONF_KILL) ? 1 : 0;
	    ktype = kline ? "K-lined" : "Autokilled";
	    reason = aconf->passwd ? aconf->passwd : ktype;

	    sendto_one(sptr, err_str(ERR_YOUREBANNEDCREEP), me.name, sptr->name,
		       ktype);
	    sendto_one(sptr,
		       ":%s NOTICE %s :*** You are not welcome on this %s.",
		       me.name, sptr->name, kline ? "server" : "network");
	    sendto_one(sptr, ":%s NOTICE %s :*** %s for %s", me.name,
		       sptr->name, ktype, reason);
	    sendto_one(sptr, ":%s NOTICE %s :*** Your hostmask is %s!%s@%s",
		       me.name, sptr->name, sptr->name, sptr->user->username,
		       sptr->sockhost);
	    sendto_one(sptr,
		       ":%s NOTICE %s :*** For more information, please mail %s and include everything shown here.",
		       me.name, sptr->name,
		       kline ? SERVER_KLINE_ADDRESS : NETWORK_KLINE_ADDRESS);

#ifdef USE_REJECT_HOLD
	    cptr->flags |= FLAGS_REJECT_HOLD;
#endif
	    ircstp->is_ref++;

#ifndef USE_REJECT_HOLD
	    return exit_client(cptr, sptr, &me, reason);
#endif
	}

	/*
	 * Limit clients 
	 */
	/*
	 * We want to be able to have servers and F-line clients connect,
	 * so save room for "buffer" connections. Smaller servers may
	 * want to decrease this, and it should probably be just a
	 * percentage of the MAXCLIENTS... -Taner
	 */
	/*
	 * Except "F:" clients 
	 */
	if ((Count.local >= (MAXCLIENTS - 10)) && !(find_fline(sptr))) {
	    sendto_realops_lev(SPY_LEV,
			       "Too many clients, rejecting %s[%s].", nick,
			       sptr->sockhost);
	    ircstp->is_ref++;
	    return exit_client(cptr, sptr, &me,
			       "Sorry, server is full - try later");
	}
#ifdef ANTI_SPAMBOT
	/*
	 * It appears, this is catching normal clients 
	 * Reject single char user-given user->host's 
	 */
	if (spamchar == 'x') {
	    sendto_realops_lev(REJ_LEV,
			       "Rejecting possible Spambot: %s (Single char user-given userhost: %c)",
			       get_client_name(sptr, FALSE), spamchar);
	    ircstp->is_ref++;
	    return exit_client(cptr, sptr, sptr, "Spambot detected, rejected.");
	}
#endif

	if (oldstatus == STAT_MASTER && MyConnect(sptr))
	    (void) m_oper(&me, sptr, 1, parv);

	/*
	 * hostile username checks begin here 
	 */

	{
    char *tmpstr;
    u_char c, cc;
    int lower, upper, special;

	    lower = upper = special = cc = 0;

	    /*
	     * check for "@" in identd reply -Taner 
	     */
	    if ((strchr(user->username, '@') != NULL)
		|| (strchr(username, '@') != NULL)) {
		sendto_realops_lev(REJ_LEV,
				   "Illegal \"@\" in username: %s (%s)",
				   get_client_name(sptr, FALSE), username);
		ircstp->is_ref++;
		(void) ircsprintf(tmpstr2,
				  "Invalid username [%s] - '@' is not allowed!",
				  username);
		return exit_client(cptr, sptr, sptr, tmpstr2);
	    }
	    /*
	     * First check user->username... 
	     */
#ifdef IGNORE_FIRST_CHAR
	    tmpstr =
		(user->username[0] ==
		 '~' ? &user->username[2] : &user->username[1]);
	    /*
	     * Ok, we don't want to TOTALLY ignore the first character. We
	     * should at least check it for control characters, etc -
	     * ThemBones
	     */
	    cc = (user->username[0] ==
		  '~' ? user->username[1] : user->username[0]);
	    if ((!isalnum(cc) && !strchr(" -_.", cc)) || (cc > 127))
		special++;
#else
	    tmpstr =
		(user->username[0] ==
		 '~' ? &user->username[1] : user->username);
#endif /* IGNORE_FIRST_CHAR */

	    while (*tmpstr) {
		c = *(tmpstr++);
		if (islower(c)) {
		    lower++;
		    continue;
		}
		if (isupper(c)) {
		    upper++;
		    continue;
		}
		if ((!isalnum(c) && !strchr(" -_.", c)) || (c > 127)
		    || (c < 32))
		    special++;
	    }
#ifdef NO_MIXED_CASE
	    if (lower && upper) {
		sendto_realops_lev(REJ_LEV, "Invalid username: %s (%s@%s)",
				   nick, user->username, user->host);
		ircstp->is_ref++;
		(void) ircsprintf(tmpstr2, "Invalid username [%s]",
				  user->username);
		return exit_client(cptr, sptr, &me, tmpstr2);
	    }
#endif /* NO_MIXED_CASE */
	    if (special) {
		sendto_realops_lev(REJ_LEV, "Invalid username: %s (%s@%s)",
				   nick, user->username, user->host);
		ircstp->is_ref++;
		(void) ircsprintf(tmpstr2, "Invalid username [%s]",
				  user->username);
		return exit_client(cptr, sptr, &me, tmpstr2);
	    }
	    /*
	     * Ok, now check the username they provided, if different 
	     */
	    lower = upper = special = cc = 0;
	    if (strcmp(user->username, username)) {

#ifdef IGNORE_FIRST_CHAR
		tmpstr = (username[0] == '~' ? &username[2] : &username[1]);
		/*
		 * Ok, we don't want to TOTALLY ignore the first character.
		 * We should at least check it for control charcters, etc
		 * -ThemBones
		 */
		cc = (username[0] == '~' ? username[1] : username[0]);
		if ((!isalnum(cc) && !strchr(" -_.", cc)) || (cc > 127))
		    special++;
#else
		tmpstr = (username[0] == '~' ? &username[1] : username);
#endif /* IGNORE_FIRST_CHAR */
		while (*tmpstr) {
		    c = *(tmpstr++);
		    if (islower(c)) {
			lower++;
			continue;
		    }
		    if (isupper(c)) {
			upper++;
			continue;
		    }
		    if ((!isalnum(c) && !strchr(" -_.", c)) || (c > 127))
			special++;
		}
#ifdef NO_MIXED_CASE
		if (lower && upper) {
		    sendto_realops_lev(REJ_LEV,
				       "Invalid username: %s (%s@%s)", nick,
				       username, user->host);
		    ircstp->is_ref++;
		    (void) ircsprintf(tmpstr2, "Invalid username [%s]",
				      username);
		    return exit_client(cptr, sptr, &me, tmpstr2);
		}
#endif
		if (special) {
		    sendto_realops_lev(REJ_LEV,
				       "Invalid username: %s (%s@%s)", nick,
				       username, user->host);
		    ircstp->is_ref++;
		    (void) ircsprintf(tmpstr2, "Invalid username [%s]",
				      username);
		    return exit_client(cptr, sptr, &me, tmpstr2);
		}
	    }			/* usernames different */
	}

	/*
	 * reject single character usernames which aren't alphabetic i.e.
	 * reject jokers who have '?@somehost' or '.@somehost'
	 * 
	 * -Dianora
	 */

	if ((user->username[1] == '\0') && !isalpha(user->username[0])) {
	    sendto_realops_lev(REJ_LEV, "Invalid username: %s (%s@%s)",
			       nick, user->username, user->host);
	    ircstp->is_ref++;
	    (void) ircsprintf(tmpstr2, "Invalid username [%s]", user->username);
	    return exit_client(cptr, sptr, &me, tmpstr2);
	}

	sendto_realops_lev(CCONN_LEV,
			   "Client connecting: %s (%s@%s) [%s] {%d}",
			   nick, user->username, user->host, sptr->hostip,
			   get_client_class(sptr));
	if ((++Count.local) > Count.max_loc) {
	    Count.max_loc = Count.local;
	    if (!(Count.max_loc % 10))
		sendto_ops("New Max Local Clients: %d", Count.max_loc);
	}
	if ((NOW - Count.day) > 86400) {
	    Count.today = 0;
	    Count.day = NOW;
	}
	if ((NOW - Count.week) > 604800) {
	    Count.weekly = 0;
	    Count.week = NOW;
	}
	if ((NOW - Count.month) > 2592000) {
	    Count.monthly = 0;
	    Count.month = NOW;
	}
	if ((NOW - Count.year) > 31536000) {
	    Count.yearly = 0;
	    Count.year = NOW;
	}
	Count.today++;
	Count.weekly++;
	Count.monthly++;
	Count.yearly++;
	if (sptr->flags & FLAGS_BAD_DNS)
	    sendto_realops_lev(SPY_LEV,
			       "DNS lookup: %s (%s@%s) is a possible cache polluter",
			       sptr->name, sptr->user->username,
			       sptr->user->host);
    } else
	strncpyzt(user->username, username, USERLEN + 1);

    SetClient(sptr);
    /*
     * Increment our total user count here 
     */
    if (++Count.total > Count.max_tot)
	Count.max_tot = Count.total;

    if (IsInvisible(sptr))
	Count.invisi++;

    if (MyConnect(sptr)) {
	sptr->pingval = get_client_ping(sptr);
	sptr->sendqlen = get_sendq(sptr);
#ifdef MAXBUFFERS
	/*
	 * Let's try changing the socket options for the client here...
	 * -Taner
	 */
	reset_sock_opts(sptr->fd, 0);
	/*
	 * End sock_opt hack 
	 */
#endif
	sendto_one(sptr, rpl_str(RPL_WELCOME), me.name, nick, NETWORKNAME,
		   nick, sptr->user->username, sptr->user->host);
	/*
	 * This is a duplicate of the NOTICE but see below...
	 * um, why were we hiding it? they did make it on to the
	 * server and all..;) -wd
	 */
	sendto_one(sptr, rpl_str(RPL_YOURHOST), me.name, nick,
		   get_client_name(&me, TRUE), version);
	sendto_one(sptr, rpl_str(RPL_CREATED), me.name, nick, creation);
	sendto_one(sptr, rpl_str(RPL_MYINFO), me.name, parv[0], me.name,
		   version);
	sendto_one(sptr, rpl_str(RPL_PROTOCTL), me.name, parv[0], MAXWATCH,
		   MAXMODEPARAMS, MAXCHANNELSPERUSER, MAXBANS, NICKLEN,
		   TOPICLEN, TOPICLEN, NETWORKNAME, MAXSILES);

	if (!BadPtr(sptr->passwd)
	    && (pwaconf->flags & CONF_FLAGS_I_OPERPORT))
	    do {
    char *onptr = sptr->passwd;
    char *opptr;
    char *onick;
#ifdef OPERS_CAN_CHOOSE_FAKEHOST
    char *hostptr;
#endif
    char *tmpptr = NULL;
    char tmppwd[PASSWDLEN + 1];

		if (!(opptr = strchr(onptr, ':')))
		    break;

		*opptr++ = '\0';
#ifdef OPERS_CAN_CHOOSE_FAKEHOST
		if ((hostptr = strchr(opptr, ':'))) {
		    *hostptr++ = '\0';
		    if ((tmpptr = strchr(hostptr, ':')))
			*tmpptr++ = '\0';
		}
#else
		if ((tmpptr = strchr(opptr, ':')))
		    *tmpptr++ = '\0';
#endif
		if (check_oper_can_mask(sptr, onptr, opptr, &onick) != 0) {
		    sendto_one(sptr,
			       ":%s NOTICE %s :*** Your hostname has been masked.",
			       me.name, sptr->name);

		    sptr->user->real_oper_host =
			MyMalloc(strlen(sptr->user->host) + 1);
		    sptr->user->real_oper_username =
			MyMalloc(strlen(sptr->username) + 1);
		    sptr->user->real_oper_ip =
			MyMalloc(strlen(sptr->hostip) + 1);
		    strcpy(sptr->user->real_oper_host, sptr->user->host);
		    strcpy(sptr->user->real_oper_username, sptr->username);
		    strcpy(sptr->user->real_oper_ip, sptr->hostip);

#ifdef OPERS_CAN_CHOOSE_FAKEHOST
		    if (hostptr)
			strncpyzt(sptr->user->host, hostptr, HOSTLEN + 1);
		    else if (pwaconf->localhost)
			strncpyzt(sptr->user->host, pwaconf->localhost,
				  HOSTLEN + 1);
		    else
			strncpyzt(sptr->user->host, STAFF_ADDRESS, HOSTLEN + 1);
#else
		    if (pwaconf->localhost)
			strncpyzt(sptr->user->host, pwaconf->localhost,
				  HOSTLEN + 1);
		    else
			strncpyzt(sptr->user->host, STAFF_ADDRESS, HOSTLEN + 1);
#endif
		    strncpyzt(sptr->user->username, onick, USERLEN + 1);
		    strncpyzt(sptr->username, onick, USERLEN + 1);
		    sptr->flags |= FLAGS_GOTID;	/* fake ident */
		    if (sptr->flags & FLAGS_IPHASH) {
			remove_one_ip(sptr->ip.s_addr);
			sptr->flags &= ~FLAGS_IPHASH;
		    }

		    sptr->ip.s_addr = 0;
		    strcpy(sptr->hostip, "0.0.0.0");
#ifdef OPERS_CAN_CHOOSE_FAKEHOST
		    if (hostptr)
			strncpy(sptr->sockhost, hostptr, HOSTLEN + 1);
		    else if (pwaconf->localhost)
			strncpy(sptr->sockhost, pwaconf->localhost,
				HOSTLEN + 1);
		    else
			strncpy(sptr->sockhost, STAFF_ADDRESS, HOSTLEN + 1);
#else
		    if (pwaconf->localhost)
			strncpy(sptr->sockhost, pwaconf->localhost,
				HOSTLEN + 1);
		    else
			strncpy(sptr->sockhost, STAFF_ADDRESS, HOSTLEN + 1);
#endif
		}

		if (tmpptr) {
		    strcpy(tmppwd, tmpptr);
		    strcpy(sptr->passwd, tmppwd);
		} else
		    sptr->passwd[0] = '\0';
	    } while (0);

	(void) send_lusers(sptr, sptr, 1, parv);

	sendto_one(sptr,
		   ":%s NOTICE %s :*** Notice -- motd was last changed at %s",
		   me.name, nick, motd_last_changed_date);
#ifdef SHORT_MOTD
	sendto_one(sptr,
		   ":%s NOTICE %s :*** Notice -- Please read the motd if you haven't read it",
		   me.name, nick);

	sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, parv[0], me.name);
	if ((smotd = shortmotd) == NULL) {
	    sendto_one(sptr, rpl_str(RPL_MOTD), me.name, parv[0],
		       "*** This is the short motd ***");
	} else {
	    while (smotd) {
		sendto_one(sptr, rpl_str(RPL_MOTD), me.name, parv[0],
			   smotd->line);
		smotd = smotd->next;
	    }
	}

	sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, parv[0]);
#else
	(void) send_motd(sptr, sptr, 1, parv);
#endif
#ifdef WINGATE_NOTICE
	sendto_one(sptr,
		   ":%s NOTICE %s :*** Notice -- This server runs an open proxy/wingate detection monitor.",
		   me.name, nick);
	sendto_one(sptr,
		   ":%s NOTICE %s :*** Notice -- If you see a port 1080 or port 23 connection from %s",
		   me.name, nick, ProxyMonHost);
	sendto_one(sptr,
		   ":%s NOTICE %s :*** Notice -- please disregard it, as it is the detector in action.",
		   me.name, nick);
	sendto_one(sptr,
		   ":%s NOTICE %s :*** Notice -- For more information please see %s",
		   me.name, nick, ProxyMonURL);
#endif

    } else if (IsServer(cptr)) {
    aClient *acptr;

	if ((acptr = find_server(user->server, NULL))
	    && acptr->from != sptr->from) {
	    sendto_realops_lev(DEBUG_LEV,
			       "Bad User [%s] :%s USER %s@%s %s, != %s[%s]",
			       cptr->name, nick, user->username,
			       user->host, user->server, acptr->name,
			       acptr->from->name);
	    sendto_one(cptr,
		       ":%s KILL %s :%s (%s != %s[%s] USER from wrong direction)",
		       me.name, sptr->name, me.name, user->server,
		       acptr->from->name, acptr->from->sockhost);
	    sptr->flags |= FLAGS_KILLED;
	    return exit_client(sptr, sptr, &me, "USER server wrong direction");

	}
	/*
	 * Super GhostDetect: If we can't find the server the user is
	 * supposed to be on, then simply blow the user away.     -Taner
	 */
	if (!acptr) {
	    sendto_one(cptr,
		       ":%s KILL %s :%s GHOST (no server %s on the net)",
		       me.name, sptr->name, me.name, user->server);
	    sendto_realops("No server %s for user %s[%s@%s] from %s",
			   user->server, sptr->name, user->username, user->host,
			   sptr->from->name);
	    sptr->flags |= FLAGS_KILLED;
	    return exit_client(sptr, sptr, &me, "Ghosted Client");
	}
    }
    send_umode(NULL, sptr, 0, SEND_UMODES, ubuf);
    if (!*ubuf) {
	ubuf[0] = '+';
	ubuf[1] = '\0';
    }
    hash_check_watch(sptr, RPL_LOGON);
#ifdef USE_FAKEHOST
    sendto_flag_serv_butone(cptr, CAPAB_NICKIP, 0,
			    "NICK %s %d %ld %s %s %s %s %s %lu %lu :%s",
			    nick, sptr->hopcount + 1, sptr->tsinfo, ubuf,
			    user->username, user->host, user->fakehost,
			    user->server, sptr->user->servicestamp,
			    htonl(sptr->ip.s_addr), sptr->info);
    sendto_flag_serv_butone(cptr, 0, CAPAB_NICKIP,
			    "NICK %s %d %ld %s %s %s %s %s %lu :%s",
			    nick, sptr->hopcount + 1, sptr->tsinfo, ubuf,
			    user->username, user->host, user->fakehost,
			    user->server, sptr->user->servicestamp, sptr->info);
    sendto_service(SERVICE_SEE_RNICKS, 0,
		   "NICK %s %d %ld %s %s %s %s %s %lu :%s", nick,
		   sptr->hopcount + 1, sptr->tsinfo, ubuf,
		   user->username, user->host, user->fakehost,
		   user->server, sptr->user->servicestamp, sptr->info);
#else
    sendto_flag_serv_butone(cptr, CAPAB_NICKIP, 0,
			    "NICK %s %d %ld %s %s %s %s %lu %lu :%s",
			    nick, sptr->hopcount + 1, sptr->tsinfo, ubuf,
			    user->username, user->host, user->server,
			    sptr->user->servicestamp, htonl(sptr->ip.s_addr),
			    sptr->info);
    sendto_flag_serv_butone(cptr, 0, CAPAB_NICKIP,
			    "NICK %s %d %ld %s %s %s %s %lu :%s", nick,
			    sptr->hopcount + 1, sptr->tsinfo, ubuf,
			    user->username, user->host, user->server,
			    sptr->user->servicestamp, sptr->info);
    sendto_service(SERVICE_SEE_RNICKS, 0, "NICK %s %d %ld %s %s %s %s %lu :%s",
		   nick, sptr->hopcount + 1, sptr->tsinfo, ubuf, user->username,
		   user->host, user->server, sptr->user->servicestamp,
		   sptr->info);
#endif
#ifdef USE_REMOTECONNECT
    sendto_realops_lev(CCONN_LEV, "Client connecting from %s: %s (%s@%s) [%s]",
		       user->server, nick, user->username, user->host,
		       inet_ntoa(sptr->ip));
#endif
    if (MyClient(sptr)) {
	/*
	 * if the I:line doesn't have a password and the user does, send it over to NickServ 
	 */
	if (sptr->passwd[0]
	    && (nsptr = find_person(NickServ, NULL)) != NULL) {
	    sendto_one(nsptr, ":%s PRIVMSG %s@%s :SIDENTIFY %s",
		       sptr->name, NickServ, SERVICES_NAME, sptr->passwd);
	}

	memset(sptr->passwd, '\0', PASSWDLEN);

	if (ubuf[1])
	    send_umode(cptr, sptr, 0, ALL_UMODES, ubuf);
    }

    return 0;
}

int do_user(char *nick, aClient *cptr, aClient *sptr, char *username,
	    char *host,
#ifdef USE_FAKEHOST
	    char *mask,
#endif
	    char *server, unsigned long serviceid, unsigned int ip,
	    char *realname)
{
    anUser *user;

    long oflags;

    user = make_user(sptr);
    oflags = sptr->umode;

    if (!MyConnect(sptr)) {
	user->server = find_or_add(server);
	strncpyzt(user->host, host, sizeof(user->host));
    } else {
	if (!IsUnknown(sptr)) {
	    sendto_one(sptr, err_str(ERR_ALREADYREGISTRED), me.name, nick);
	    return 0;
	}
	sptr->umode |= UMODE_i;	/* invisible IS default. always */
#ifdef USE_FAKEHOST
	sptr->umode |= UMODE_x;
#endif
#ifdef USE_REGISTERFILTER
	sptr->umode |= UMODE_R;
#endif

	sptr->umode |= (UFLAGS & atoi(host));
	strncpyzt(user->host, host, sizeof(user->host));
	user->server = me.name;
    }
#ifdef USE_FAKEHOST
    if (mask)
	strncpyzt(user->fakehost, mask, sizeof(user->fakehost));
#endif
    strncpyzt(sptr->info, realname, sizeof(sptr->info));

    sptr->user->servicestamp = serviceid;
    if (!MyConnect(sptr)) {
	sptr->ip.s_addr = ntohl(ip);

	/* add non-local clients to the throttle checker.  obviously, we only
	 * do this for REMOTE clients!@$$@!  throttle_check() is called
	 * elsewhere for the locals! -wd */
	if (ip != 0)
	    throttle_check(inetntoa((char *) &sptr->ip), 0);
    }
    if (MyConnect(sptr))
	sptr->oflag = 0;
    if (sptr->name[0])
	return register_user(cptr, sptr, sptr->name, username);
    else
	strncpyzt(sptr->user->username, username, USERLEN + 1);
    return 0;
}

/*
 * * m_ping * parv[0] = sender prefix *       parv[1] = origin *
 * parv[2] = destination
 */
int m_ping(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    char *origin, *destination;

    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NOORIGIN), me.name, parv[0]);
	return 0;
    }
    origin = parv[1];
    destination = parv[2];	/* 
				 * Will get NULL or pointer (parc >=
				 * * 2!!) 
				 */

    acptr = find_client(origin, NULL);
    if (!acptr)
	acptr = find_server(origin, NULL);
    if (acptr && acptr != sptr)
	origin = cptr->name;
    if (!BadPtr(destination) && mycmp(destination, me.name) != 0) {
	if ((acptr = find_server(destination, NULL)))
	    sendto_one(acptr, ":%s PING %s :%s", parv[0], origin, destination);
	else {
	    sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0],
		       destination);
	    return 0;
	}
    } else
	sendto_one(sptr, ":%s PONG %s :%s", me.name,
		   (destination) ? destination : me.name, origin);
    return 0;
}

/*
 * * m_pong * parv[0] = sender prefix *       parv[1] = origin *
 * parv[2] = destination
 */
int m_pong(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    char *origin, *destination;

    if (parc < 2 || *parv[1] == '\0') {
	sendto_one(sptr, err_str(ERR_NOORIGIN), me.name, parv[0]);
	return 0;
    }

    origin = parv[1];
    destination = parv[2];
    cptr->flags &= ~FLAGS_PINGSENT;
    sptr->flags &= ~FLAGS_PINGSENT;

    /*
     * if it's my client and it's a server.. 
     */
    if (sptr == cptr && IsServer(cptr)) {
	if (sptr->flags & FLAGS_USERBURST) {
	    sptr->flags &= ~FLAGS_USERBURST;
	    sendto_gnotice
		("from %s: %s has processed user/channel burst, sending topic burst.",
		 me.name, sptr->name);
	    send_topic_burst(sptr);
	    sptr->flags |= FLAGS_PINGSENT | FLAGS_SOBSENT;
	    sendto_one(sptr, "PING :%s", me.name);
	} else if (sptr->flags & FLAGS_TOPICBURST) {
	    sptr->flags &= ~FLAGS_TOPICBURST;
	    sendto_gnotice
		("from %s: %s has processed topic burst (synched to network data).",
		 me.name, sptr->name);
#ifdef HUB
	    sendto_serv_butone(sptr,
			       ":%s GNOTICE :%s has synched to network data.",
			       me.name, sptr->name);
#endif
	}
    }

    /*
     * Now attempt to route the PONG, comstud pointed out routable PING
     * is used for SPING.  routable PING should also probably be left in
     * -Dianora That being the case, we will route, but only for
     * registered clients (a case can be made to allow them only from
     * servers). -Shadowfax
     */
    if (!BadPtr(destination) && mycmp(destination, me.name) != 0 &&
	IsRegistered(sptr)) {
	if ((acptr = find_client(destination, NULL)) ||
	    (acptr = find_server(destination, NULL)))
	    sendto_one(acptr, ":%s PONG %s %s", parv[0], origin, destination);
	else {
	    sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0],
		       destination);
	    return 0;
	}
    }
#ifdef	DEBUGMODE
    else
	Debug((DEBUG_NOTICE, "PONG: %s %s", origin,
	       destination ? destination : "*"));
#endif
    return 0;
}

int check_oper_can_mask(aClient *sptr, char *name, char *password, char **onick)
{

#ifdef STAFFHIDE

    aConfItem *aconf;
    char *encr;

# ifdef CRYPT_OPER_PASSWORD
    extern char *crypt();
# endif

    if (!
	(aconf =
	 find_conf_exact(name, sptr->username, sptr->sockhost, CONF_OPS))
&& !(aconf = find_conf_exact(name, sptr->username, sptr->hostip, CONF_OPS))) {
	sendto_realops
	    ("Failed OPERMASK attempt by %s (%s@%s) [No Entry for %s]",
	     sptr->name, sptr->user->username, sptr->user->host, name);
	return 0;
    }
# ifdef CRYPT_OPER_PASSWORD
    /*
     * use first two chars of the password they send in as salt 
     * passwd may be NULL pointer. Head it off at the pass... 
     */
    if (password && *aconf->passwd)
	encr = crypt(password, aconf->passwd);
    else
	encr = "";
# else
    encr = password;
# endif	/* CRYPT_OPER_PASSWORD */

    if (StrEq(encr, aconf->passwd)) {
# ifdef USE_SYSLOG
	syslog(LOG_INFO, "OPERMASK: %s (%s!%s@%s)", aconf->name,
	       sptr->name, sptr->user->username, sptr->user->host);
# endif
	*onick = aconf->name;
	sendto_realops("%s [%s] (%s@<hidden>) has masked their hostname.",
		       sptr->name, aconf->name, sptr->user->username);

	return 1;
    }

    sendto_realops
	("Failed OPERMASK attempt by %s (%s@%s) [Bad Password]",
	 sptr->name, sptr->user->username, sptr->user->host);

#endif

    return 0;
}

/***************************************************************************
 * m_pass() - Added Sat, 4 March 1989
 ***************************************************************************/
/*
 * * m_pass * parv[0] = sender prefix *       parv[1] = password *
 * parv[2] = optional extra version information
 */
int m_pass(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *password = parc > 1 ? parv[1] : NULL;

    if (BadPtr(password)) {
	sendto_one(cptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "PASS");
	return 0;
    }
    if (!MyConnect(sptr) || (!IsUnknown(cptr) && !IsHandshake(cptr))) {
	sendto_one(cptr, err_str(ERR_ALREADYREGISTRED), me.name, parv[0]);
	return 0;
    }
    strncpyzt(cptr->passwd, password, sizeof(cptr->passwd));
    if (parc > 2) {
    int l = strlen(parv[2]);

	if (l < 2)
	    return 0;
	/*
	 * if (strcmp(parv[2]+l-2, "TS") == 0) 
	 */
	if (parv[2][0] == 'T' && parv[2][1] == 'S')
	    cptr->tsinfo = (ts_val) TS_DOESTS;
    }
    return 0;
}
