/***********************************************************************
 *   IRC - Internet Relay Chat, src/send_oper.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *		      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: send_oper.c,v 1.6 2001/07/22 23:39:25 tr-ircd Exp $ 
 */

#include "struct.h"
#include "channel.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include <stdio.h>
#include "numeric.h"
#include "dh.h"
#include "zlink.h"

#ifdef ALWAYS_SEND_DURING_SPLIT
extern int currently_processing_netsplit;
#endif

/*
 * This routine increments our serial number so it will
 * * be unique from anything in sentalong, no need for a memset
 * * except for every MAXINT calls - lucas
 */

/*
 * This should work on any OS where an int is 32 bit, I hope.. 
 */

#define HIGHEST_SERIAL INT_MAX

#define INC_SERIAL if(sent_serial == HIGHEST_SERIAL) \
   { memset(sentalong, 0, sizeof(sentalong)); sent_serial = 0; } \
   sent_serial++;

/*
 * sendto_ops
 * 
 * Send to *local* ops only.
 */
void sendto_ops(char *pattern, ...)
{
    aClient *cptr;
    int i;
    char nbuf[1024];
    va_list vl;

    va_start(vl, pattern);
    for (i = 0; i <= highest_fd; i++)
	if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr)
	    && IsAnOper(cptr)) {
	    (void) ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- ", me.name,
			      cptr->name);
	    (void) strncat(nbuf, pattern, sizeof(nbuf) - strlen(nbuf));
	    vsendto_one(cptr, nbuf, vl);
	}
    va_end(vl);
    return;
}

/*
 * * sendto_wallops_butone *      Send message to all operators. * one
 * - client not to send message to * from- client which message is from
 * *NEVER* NULL!!
 */
void sendto_wallops_butone(aClient *one, aClient *from, char *pattern, ...)
{
    int i;
    aClient *cptr;
    va_list vl;

    va_start(vl, pattern);
    for (i = 0; i <= highest_fd; i++) {
	if ((cptr = local[i]) != NULL) {
	    if (!(IsRegistered(cptr)
		  && (SendWallops(cptr) || IsServer(cptr))) || cptr == one)
		continue;
	    vsendto_prefix_one(cptr, from, pattern, vl);
	}
    }
    va_end(vl);
    return;
}

void send_globops(char *pattern, ...)
{
    aClient *cptr;
    int i;
    char nbuf[1024];
    va_list vl;

    va_start(vl, pattern);
    for (i = 0; i <= highest_fd; i++)
	if ((cptr = local[i]) && !IsServer(cptr) && IsAnOper(cptr) &&
	    !IsMe(cptr) && SendGlobops(cptr)) {
	    (void) ircsprintf(nbuf, ":%s NOTICE %s :*** Global -- ", me.name,
			      cptr->name);
	    (void) strncat(nbuf, pattern, sizeof(nbuf) - strlen(nbuf));
	    vsendto_one(cptr, nbuf, vl);
	}
    va_end(vl);
    return;
}

/*
 * sendto_realops
 * 
 * Send to *local* ops only but NOT +s nonopers.
 * If it's to local ops only and not +s nonopers, then SendServNotice is
 * wrong. Changed to IsAnOper. -mjs
 */
void sendto_realops(char *pattern, ...)
{
    aClient *cptr;
    int i;
    char nbuf[1024];
    fdlist *l;
    int fd;
    va_list vl;

    va_start(vl, pattern);
    l = &oper_fdlist;
    for (fd = l->entry[i = 1]; i <= l->last_entry; fd = l->entry[++i]) {
	if (!(cptr = local[fd]))
	    continue;
	if (IsAnOper(cptr)) {
	    ircsprintf(nbuf, ":%s NOTICE %s :*** Notice -- %s", me.name,
		       cptr->name, pattern);
	    vsendto_one(cptr, nbuf, vl);
	}
    }
    va_end(vl);
    return;
}

/*
 * sendto_locops
 */
void sendto_locops(char *pattern, ...)
{
    aClient *cptr;
    int i;
    char nbuf[1024];
    fdlist *l;
    int fd;
    va_list vl;

    va_start(vl, pattern);
    l = &oper_fdlist;
    for (fd = l->entry[i = 1]; i <= l->last_entry; fd = l->entry[++i]) {
	if (!(cptr = local[fd]))
	    continue;
	if (SendGlobops(cptr)) {
	    ircsprintf(nbuf, ":%s NOTICE %s :*** LocOps -- ", me.name,
		       cptr->name);
	    strncat(nbuf, pattern, sizeof(nbuf) - strlen(nbuf));
	    vsendto_one(cptr, nbuf, vl);
	}
    }
    va_end(vl);
    return;
}

/*
 * sendto_gnotice - send a routing notice to all local +n users.
 */
void sendto_gnotice(char *pattern, ...)
{
    aClient *cptr;
    int i;
    char nbuf[1024];
    va_list vl;

    va_start(vl, pattern);

    for (i = 0; i <= highest_fd; i++) {
	if ((cptr = local[i]) && !IsServer(cptr) && !IsMe(cptr)
	    && SendRnotice(cptr)) {
	    ircsprintf(nbuf, ":%s NOTICE %s :*** Routing -- ", me.name,
		       cptr->name);
	    strncat(nbuf, pattern, sizeof(nbuf) - strlen(nbuf));
	    vsendto_one(cptr, nbuf, vl);
	}
    }
    va_end(vl);
    return;
}
