/***********************************************************************
 *   IRC - Internet Relay Chat, src/send_serv.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *		      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: send_serv.c,v 1.7 2001/07/22 23:39:25 tr-ircd Exp $ 
 */

#include "struct.h"
#include "channel.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include <stdio.h>
#include "numeric.h"
#include "dh.h"
#include "zlink.h"

#ifdef ALWAYS_SEND_DURING_SPLIT
extern int currently_processing_netsplit;
#endif

/*
 * This routine increments our serial number so it will
 * * be unique from anything in sentalong, no need for a memset
 * * except for every MAXINT calls - lucas
 */

/*
 * This should work on any OS where an int is 32 bit, I hope.. 
 */

#define HIGHEST_SERIAL INT_MAX

#define INC_SERIAL if(sent_serial == HIGHEST_SERIAL) \
   { memset(sentalong, 0, sizeof(sentalong)); sent_serial = 0; } \
   sent_serial++;

/*
 * sendto_server_butone
 * 
 * Send a message to all connected servers except the client 'one'.
 */
void sendto_serv_butone(aClient *one, char *pattern, ...)
{
    int i;
    aClient *cptr;
    int j, k = 0;
    fdlist send_fdlist;
    va_list vl;

    va_start(vl, pattern);
    for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry;
	 i = serv_fdlist.entry[++j]) {
	if (!(cptr = local[i]) || (one && cptr == one->from))
	    continue;
	send_fdlist.entry[++k] = i;
    }
    send_fdlist.last_entry = k;
    if (k)
	vsendto_fdlist(&send_fdlist, pattern, vl);
    va_end(vl);
    return;
}

/*
 * * send a msg to all ppl on servers/hosts that match a specified mask *
 * (used for enhanced PRIVMSGs) *
 * 
 * addition -- Armin, 8jun90 (gruner@informatik.tu-muenchen.de)
 */

static int match_it(aClient *one, char *mask, int what)
{
    if (what == MATCH_HOST)
	return (match(mask, one->user->host) == 0);
    else
	return (match(mask, one->user->server) == 0);
}

/*
 * sendto_match_servs
 * 
 * send to all servers which match the mask at the end of a channel name
 * (if there is a mask present) or to all if no mask.
 */
void sendto_match_servs(aChannel *chptr, aClient *from, char *pattern, ...)
{
    int j, k = 0;
    fdlist send_fdlist;
    int i;
    aClient *cptr;
    va_list vl;

    if (chptr) {
	if (*chptr->chname == '&')
	    return;
    }
    va_start(vl, pattern);
    for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry;
	 i = serv_fdlist.entry[++j]) {
	if (!(cptr = local[i]))
	    continue;
	if (cptr == from)
	    continue;
	send_fdlist.entry[++k] = i;
    }
    send_fdlist.last_entry = k;
    if (k)
	vsendto_fdlist(&send_fdlist, pattern, vl);
    va_end(vl);
    return;
}

void sendto_match_butone(aClient *one, aClient *from, char *mask, int what,
			 char *pattern, ...)
{
    int i;
    aClient *cptr, *acptr;
    char cansendlocal, cansendglobal;
    va_list vl;

    va_start(vl, pattern);
    if (MyConnect(from)) {
	cansendlocal = (OPCanLNotice(from)) ? 1 : 0;
	cansendglobal = (OPCanGNotice(from)) ? 1 : 0;
    } else
	cansendlocal = cansendglobal = 1;
    for (i = 0; i <= highest_fd; i++) {
	if (!(cptr = local[i]))
	    continue;		/* that clients are not mine */
	if (cptr == one)	/* must skip the origin !! */
	    continue;
	if (IsServer(cptr)) {
	    if (!cansendglobal)
		continue;
	    for (acptr = client; acptr; acptr = acptr->next)
		if (IsRegisteredUser(acptr)
		    && match_it(acptr, mask, what)
		    && acptr->from == cptr)
		    break;
	    /*
	     * a person on that server matches the mask, so we * send *one*
	     * msg to that server ...
	     */
	    if (acptr == NULL)
		continue;
	    /*
	     * ... but only if there *IS* a matching person 
	     */
	}
	/*
	 * my client, does he match ? 
	 */
	else if (!cansendlocal ||
		 !(IsRegisteredUser(cptr) && match_it(cptr, mask, what)))
	    continue;
	vsendto_prefix_one(cptr, from, pattern, vl);
    }
    va_end(vl);
    return;
}

/*
 * sendto_flag_serv_butone
 * 
 * Send a message to all connected servers except the client 'one'.
 * Only if they include INCLUDE and exclude EXCLUDE
 */
void sendto_flag_serv_butone(aClient *one, int include, int exclude,
			     char *pattern, ...)
{
    int i;
    aClient *cptr;
    int j, k = 0;
    fdlist send_fdlist;
    va_list vl;

    va_start(vl, pattern);
    for (i = serv_fdlist.entry[j = 1]; j <= serv_fdlist.last_entry;
	 i = serv_fdlist.entry[++j]) {
	if (!(cptr = local[i]) ||
	    (include && !((cptr->capabilities & include) == include)) ||
	    (exclude && (cptr->capabilities & exclude) == exclude) ||
	    one == cptr)
	    continue;
	send_fdlist.entry[++k] = i;
    }
    send_fdlist.last_entry = k;
    if (k)
	vsendto_fdlist(&send_fdlist, pattern, vl);
    va_end(vl);
    return;
}
