/***********************************************************************
 *   IRC - Internet Relay Chat, src/send_user.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *		      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: send_user.c,v 1.5 2001/07/22 23:39:25 tr-ircd Exp $ 
 */

#include "struct.h"
#include "channel.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include <stdio.h>
#include "numeric.h"
#include "dh.h"
#include "zlink.h"

#ifdef ALWAYS_SEND_DURING_SPLIT
extern int currently_processing_netsplit;
#endif

static char sendbuf[2048];

/*
 * This routine increments our serial number so it will
 * * be unique from anything in sentalong, no need for a memset
 * * except for every MAXINT calls - lucas
 */

/*
 * This should work on any OS where an int is 32 bit, I hope.. 
 */

#define HIGHEST_SERIAL INT_MAX

#define INC_SERIAL if(sent_serial == HIGHEST_SERIAL) \
   { memset(sentalong, 0, sizeof(sentalong)); sent_serial = 0; } \
   sent_serial++;

/*
 * * send message to single client
 */
void sendto_one(aClient *to, char *pattern, ...)
{
    va_list vl;
    int len;			/* used for the length of the current message */

    va_start(vl, pattern);
    len = ircvsprintf(sendbuf, pattern, vl);

    if (to->from)
	to = to->from;
    if (IsMe(to)) {
	sendto_ops("Trying to send [%s] to myself!", sendbuf);
	return;
    }
    send_message(to, sendbuf, len);
    va_end(vl);
}

void vsendto_one(aClient *to, char *pattern, va_list vl)
{
    int len;			/* used for the length of the current message */

    len = ircvsprintf(sendbuf, pattern, vl);

    if (to->from)
	to = to->from;
    if (IsMe(to) && to->fd >= 0) {
	sendto_ops("Trying to send [%s] to myself!", sendbuf);
	return;
    }
    send_message(to, sendbuf, len);
}

/*
 * to - destination client from - client which message is from
 * 
 * NOTE: NEITHER OF THESE SHOULD *EVER* BE NULL!! -avalon
 */
void sendto_prefix_one(aClient *to, aClient *from, char *pattern, ...)
{
    static char sender[HOSTLEN + NICKLEN + USERLEN + 5];
    static char temp[1024];
    anUser *user;
    char *idx;
    char *par;
    int flag = 0, sidx = 0;
    va_list vl, vl2;

    va_start(vl, pattern);
    vl2 = vl;

    par = va_arg(vl, char *);

    /*
     * Optimize by checking if (from && to) before everything 
     * uhh, there's _always_ going to be a to!
     */
    if (from) {
	if (!MyClient(from) && IsPerson(to) && (to->from == from->from)) {
	    if (IsServer(from)) {
		ircvsprintf(temp, pattern, vl2);
		sendto_ops
		    ("Send message (%s) to %s[%s] dropped from %s(Fake Dir)",
		     temp, to->name, to->from->name, from->name);
		va_end(vl);
		return;
	    }

	    sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name,
		       to->user->username,
#ifdef USE_FAKEHOST
		       IsFake(to) ? to->user->fakehost : to->user->host,
#else
		       to->user->host,
#endif
		       from->name, from->user->username,
#ifdef USE_FAKEHOST
		       IsFake(from) ? from->user->fakehost : from->user->host,
#else
		       from->user->host,
#endif
		       to->from->name);
	    sendto_serv_butone(NULL,
			       ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)",
			       me.name, to->name, me.name, to->name,
			       to->user->username,
#ifdef USE_FAKEHOST
			       IsFake(to) ? to->user->fakehost : to->user->host,
#else
			       to->user->host,
#endif
			       to->from->name);

	    to->flags |= FLAGS_KILLED;
	    exit_client(NULL, to, &me, "Ghosted client");
	    if (IsPerson(from))
		sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name,
			   from->name, to->name, to->user->username,
#ifdef USE_FAKEHOST
			   IsFake(to) ? to->user->fakehost : to->user->host,
#else
			   to->user->host,
#endif
			   to->from);
	    va_end(vl);
	    return;
	}

	if (MyClient(to) && IsPerson(from) && !mycmp(par, from->name)) {
	    user = from->user;

	    for (idx = from->name; *idx; idx++)
		sender[sidx++] = *idx;

	    if (user) {
		if (*user->username) {
		    sender[sidx++] = '!';
		    for (idx = user->username; *idx; idx++)
			sender[sidx++] = *idx;
		}
		if (*user->host && !MyConnect(from)) {
		    sender[sidx++] = '@';
		    for (
#ifdef USE_FAKEHOST
			    idx =
			    IsFake(from) ? from->user->fakehost : from->user->
			    host;
#else
			    idx = from->user->host;
#endif
			    *idx; idx++)

			sender[sidx++] = *idx;
		    flag = 1;
		}
	    }

	    /*
	     * flag is used instead of index(sender, '@') for speed and
	     * also since username/nick may have had a '@' in them.
	     * -avalon
	     */

	    if (!flag && MyConnect(from) && *user->host) {
		sender[sidx++] = '@';
		for (
#ifdef USE_FAKEHOST
			idx =
			IsFake(from) ? from->user->fakehost : from->user->host;
#else
			idx = from->user->host;
#endif
			*idx; idx++)
		    sender[sidx++] = *idx;
	    }

	    sender[sidx] = '\0';
	    par = sender;

	}
    }

    temp[0] = ':';
    sidx = 1;

    /*
     * okay, we more or less know that our sendto_prefix crap is going to be :%s <blah>,
     * * so it's easy to fix these lame problems...joy 
     */

    for (idx = par; *idx; idx++)
	temp[sidx++] = *idx;
    for (idx = (pattern + 3); *idx; idx++)
	temp[sidx++] = *idx;

    temp[sidx] = '\0';

    vsendto_one(to, temp, vl);
    va_end(vl);
}

/*
 * this is an incredibly expensive function. 
 * * removed all strcat() calls. - lucas 
 */

void vsendto_prefix_one(aClient *to, aClient *from, char *pattern, va_list vl)
{
    static char sender[HOSTLEN + NICKLEN + USERLEN + 5];
    static char temp[1024];
    anUser *user;
    char *idx;
    char *par;
    int flag = 0, sidx = 0;
    va_list vl2 = vl;

    par = va_arg(vl2, char *);

    /*
     * Optimize by checking if (from && to) before everything 
     * uhh, there's _always_ going to be a to!
     */
    if (from) {
	if (!MyClient(from) && IsPerson(to) && (to->from == from->from)) {
	    if (IsServer(from)) {
		ircvsprintf(temp, pattern, vl);
		sendto_ops
		    ("Send message (%s) to %s[%s] dropped from %s(Fake Dir)",
		     temp, to->name, to->from->name, from->name);
		return;
	    }

	    sendto_ops("Ghosted: %s[%s@%s] from %s[%s@%s] (%s)", to->name,
		       to->user->username, to->user->host, from->name,
		       from->user->username,
#ifdef USE_FAKEHOST
		       IsFake(from) ? from->user->fakehost : from->user->host,
#else
		       from->user->host,
#endif
		       to->from->name);
	    sendto_serv_butone(NULL,
			       ":%s KILL %s :%s (%s[%s@%s] Ghosted %s)",
			       me.name, to->name, me.name, to->name,
			       to->user->username,
#ifdef USE_FAKEHOST
			       IsFake(to) ? to->user->fakehost : to->user->host
#else
			       to->user->host
#endif
		);

	    to->flags |= FLAGS_KILLED;
	    exit_client(NULL, to, &me, "Ghosted client");
	    if (IsPerson(from))
		sendto_one(from, err_str(ERR_GHOSTEDCLIENT), me.name,
			   from->name, to->name, to->user->username,
#ifdef USE_FAKEHOST
			   IsFake(to) ? to->user->fakehost : to->user->host,
#else
			   to->user->host,
#endif
			   to->from);
	    return;
	}

	if (MyClient(to) && IsPerson(from) && !mycmp(par, from->name)) {
	    user = from->user;

	    for (idx = from->name; *idx; idx++)
		sender[sidx++] = *idx;

	    if (user) {
		if (*user->username) {
		    sender[sidx++] = '!';
		    for (idx = user->username; *idx; idx++)
			sender[sidx++] = *idx;
		}
		if (*user->host && !MyConnect(from)) {
		    sender[sidx++] = '@';
		    for (
#ifdef USE_FAKEHOST
			    idx =
			    IsFake(from) ? from->user->fakehost : from->user->
			    host;
#else
			    idx = from->user->host;
#endif
			    *idx; idx++)
			sender[sidx++] = *idx;
		    flag = 1;
		}
	    }

	    /*
	     * flag is used instead of index(sender, '@') for speed and
	     * also since username/nick may have had a '@' in them.
	     * -avalon
	     */

	    if (!flag && MyConnect(from) && *user->host) {
		sender[sidx++] = '@';
		for (
#ifdef USE_FAKEHOST
			idx =
			IsFake(from) ? from->user->fakehost : from->user->host;
#else
			idx = from->user->host;
#endif
			*idx; idx++)

		    sender[sidx++] = *idx;
	    }

	    sender[sidx] = '\0';
	    par = sender;

	}
    }

    temp[0] = ':';
    sidx = 1;

    /*
     * okay, we more or less know that our sendto_prefix crap is going to be :%s <blah>,
     * * so it's easy to fix these lame problems...joy 
     */

    for (idx = par; *idx; idx++)
	temp[sidx++] = *idx;
    for (idx = (pattern + 3); *idx; idx++)
	temp[sidx++] = *idx;

    temp[sidx] = '\0';

    vsendto_one(to, temp, vl2);
}
