/************************************************************************
 * IRC - Internet Relay Chat, src/bsd.c
 * Copyright (C) 1990 Jarkko Oikarinen and
 * University of Oulu, Computing Center
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * -Talesin
 *
 * 02.06.2001
 *
 * This file contains the TTNet Extensions used in the ircd
 * 
 * Any laws and rules applicable to other source files of
 * the Bahamut ircd are applicable to this file, too.
 *
 */

#include "struct.h"
#include "numeric.h"
#include "sys.h"
#include "msg.h"
#include "h.h"
#include <fcntl.h>

extern aMotd *conffile;		/* conffile is a link list of aMotd */

#include "dich_conf.h"

int m_ircops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    char buf[BUFSIZE];
    char infobuf[REALLEN];
    int opers = 0, admins = 0, globs = 0, aways = 0;

    for (acptr = client; acptr; acptr = acptr->next) {
	if (!IsULine(acptr) && IsOper(acptr)) {
	    if (!acptr->user)
		continue;
	    strncpyzt(infobuf, acptr->info, REALLEN - 1);
	    sendto_one(sptr,
		       ":%s %d %s :\2%s\2 is %s %s (%s) :%s",
		       me.name, RPL_IRCOPS, sptr->name,
		       acptr->name ? acptr->name : "<unknown>",
		       NETWORKNAME,
		       IsAdmin(acptr) ? "Server Administrator" :
		       "Operator",
		       acptr->user->away ? "- (Away) -" : "", infobuf);

	    if (!acptr->user->away) {
		if (IsAdmin(acptr))
		    admins++;
		else
		    opers++;
	    } else {
		aways++;
	    }
	}
    }
    globs = opers + admins + aways;
    sprintf(buf,
	    "Total: \2%d\2 IRCOP%s online - \2%d\2 Admin%s, \2%d\2 Oper%s and \2%d\2 Away",
	    globs, (globs) > 1 ? "s" : "", admins, admins > 1 ? "s" : "",
	    opers, opers > 1 ? "s" : "", aways);
    sendto_one(sptr, rpl_str(RPL_IRCOPS), me.name, buf);
    sendto_one(sptr, rpl_str(RPL_ENDOFIRCOPS), me.name);
    return 0;
}

int m_display(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aMotd *conffile_ptr;
    char *pass = NULL;

    if ((pass = (char *) find_displaypass())) {
	if (parc < 2) {
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
		       "DISPLAY");
	    return 0;
	}
    } else {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    if (!IsAdmin(sptr) || (conffile == (aMotd *) NULL)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	sendto_realops("\2%s\2 failed to display server config file", parv[0]);
	return 0;
    }

    if (strcmp(pass, parv[1])) {
	sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
	return 0;
    }

    conffile_ptr = conffile;
    sendto_realops("%s is displaying server config file", parv[0]);

    while (conffile_ptr) {
	sendto_one(sptr, ":%s NOTICE %s :%s", me.name, parv[0],
		   conffile_ptr->line);
	conffile_ptr = conffile_ptr->next;
    }
    return 0;
}

void read_conf(char *filename)
{
    aMotd *temp, *last;
    char buffer[MOTDLINELEN], *tmp;
    int fd;

    while (conffile) {
	temp = conffile->next;
	MyFree(conffile);
	conffile = temp;
    }
    fd = open(filename, O_RDONLY);
    if (fd == -1)
	return;
    last = (aMotd *) NULL;
    while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) {
	if ((tmp = (char *) strchr(buffer, '\n')))
	    *tmp = '\0';
	if ((tmp = (char *) strchr(buffer, '\r')))
	    *tmp = '\0';
	temp = (aMotd *) MyMalloc(sizeof(aMotd));

	strncpyzt(temp->line, buffer, MOTDLINELEN);
	temp->next = (aMotd *) NULL;
	if (!conffile)
	    conffile = temp;
	else
	    last->next = temp;
	last = temp;
    }
    close(fd);
}

int m_mkpasswd(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
#ifdef CRYPT_OPER_PASSWORD
    static char saltChars[] =
	"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
    char salt[3];
    extern char *crypt();
    int useable = 0;

    if (parc > 1) {
	if (strlen(parv[1]) >= 1)
	    useable = 1;
    }

    if (useable == 0) {
	sendto_one(sptr, ":%s NOTICE %s :*** Nothing to encrypt", me.name,
		   parv[0]);
	return 0;
    }
    srandom(time(0));
    salt[0] = saltChars[random() % 64];
    salt[1] = saltChars[random() % 64];
    salt[2] = 0;

    if ((strchr(saltChars, salt[0]) == NULL)
	|| (strchr(saltChars, salt[1]) == NULL)) {
	sendto_one(sptr, ":%s NOTICE %s :*** Illegal salt %s", me.name, parv[0],
		   salt);
	return 0;
    }

    sendto_one(sptr, ":%s NOTICE %s :*** Encryption for [%s] is %s",
	       me.name, parv[0], parv[1], crypt(parv[1], salt));
#endif
    return 0;
}

/*
 * dump_map (used by m_map)
 */
void dump_map(aClient *cptr, aClient *server, char *mask, int prompt_length,
	      int length)
{
    static char prompt[64];
    char *p = &prompt[prompt_length];
    int cnt = 0, local = 0;
    aClient *acptr;
    char *tmp = NULL;

    *p = '\0';

    for (acptr = client; acptr; acptr = acptr->next) {
	if (IsPerson(acptr)) {
	    ++cnt;
	    if (!strcmp(acptr->user->server, server->name))
		++local;
	}
    }
    if (tmp)
	*tmp = '\0';
    sendto_one(cptr, rpl_str(RPL_MAP), me.name, cptr->name, prompt,
	       length, server->name, local, (local * 100) / cnt);
    cnt = 0;
    if (tmp)
	*tmp = '_';

    if (prompt_length > 0) {
	p[-1] = ' ';
	if (p[-2] == '`')
	    p[-2] = ' ';
    }
    if (prompt_length > 60)
	return;

    strcpy(p, "|-");

    for (acptr = client; acptr; acptr = acptr->next) {
	if (!IsServer(acptr) || strcmp(acptr->serv->up, server->name))
	    continue;

#ifdef HIDEULINEDSERVS
	if (!IsOper(cptr) && IsULine(acptr))
	    continue;
	else
#else
	if (!match(mask, acptr->name))
#endif
	    cnt++;

	if (--cnt == 0)
	    *p = '`';
	dump_map(cptr, acptr, mask, prompt_length + 2, length - 2);
    }
    if (prompt_length > 0)
	p[-1] = '-';
}

/*
 * ** m_map
 * **
 * ** parv[0] = sender prefix
 * ** parv[1] = server mask
 * * */
int m_map(aClient *cptr, aClient *sptr, int parc, char *parv[])
{

    aClient *acptr;
    int longest = strlen(me.name);

    if ((bootopt & BOOT_HIDELINKS) && !IsOper(sptr)) {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    if (parc < 2)
	parv[1] = "*";

    for (acptr = client; acptr; acptr = acptr->next)
	if (IsServer(acptr)
	    && (strlen(acptr->name) + acptr->hopcount * 2) > longest)
	    longest = strlen(acptr->name) + acptr->hopcount * 2;

    if (longest > 60)
	longest = 60;
    longest += 2;
    dump_map(sptr, &me, parv[1], 0, longest);
    sendto_one(sptr, rpl_str(RPL_ENDMAP), me.name, parv[0]);

    return 0;
}
