/************************************************************************
 *   IRC - Internet Relay Chat, src/m_dccallow.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <sys/stat.h>
#include <utmp.h>
#include <fcntl.h>
#include "h.h"

int check_for_ctcp(char *, char **);
int allow_dcc(aClient *, aClient *);

char *exploits_2char[] = { "js", "pl", NULL };
char *exploits_3char[] =
    { "exe", "com", "bat", "dll", "ini", "vbs", "pif", "mrc", "scr",
    "doc", "xls", "lnk", "shs",
    "htm", NULL
};
char *exploits_4char[] = { "html", NULL };

int check_dccsend(aClient *from, aClient *to, char *msg)
{
    /*
     * we already know that msg will consist of "DCC SEND" so we can skip to the end 
     */
    char *filename = msg + 8;
    char *ext;
    char **farray;
    int arraysz;
    int len = 0, extlen = 0, i;

    /*
     * people can send themselves stuff all the like..
     * * opers need to be able to send cleaner files
     * * sanity checks..
     */

    if (from == to || !IsPerson(from) || IsAnOper(from) || !MyClient(to))
	return 0;

    while (*filename == ' ')
	filename++;

    if (!(*filename))
	return 0;

    while (*(filename + len) != ' ') {
	if (!(*(filename + len)))
	    break;
	len++;
    }

    for (ext = filename + len;; ext--) {
	if (ext == filename)
	    return 0;

	if (*ext == '.') {
	    ext++;
	    extlen--;
	    break;
	}
	extlen++;
    }

    switch (extlen) {
	case 2:
	    farray = exploits_2char;
	    arraysz = 2;
	    break;

	case 3:
	    farray = exploits_3char;
	    arraysz = 3;
	    break;

	case 4:
	    farray = exploits_4char;
	    arraysz = 4;
	    break;

	    /*
	     * no executable file here.. 
	     */
	default:
	    return 0;
    }

    for (i = 0; farray[i]; i++) {
	if (myncmp(farray[i], ext, arraysz) == 0)
	    break;
    }

    if (farray[i] == NULL)
	return 0;

    if (!allow_dcc(to, from)) {
    char tmpext[8];
    char tmpfn[128];
    Link *tlp, *flp;
    aChannel *chptr = NULL;

	strncpy(tmpext, ext, extlen);
	tmpext[extlen] = '\0';

	if (len > 127)
	    len = 127;
	strncpy(tmpfn, filename, len);
	tmpfn[len] = '\0';

	/*
	 * use notices!
	 * *   server notices are hard to script around.
	 * *   server notices are not ignored by clients.
	 */

	sendto_one(from,
		   ":%s NOTICE %s :The user %s is not accepting DCC sends of filetype *.%s from you."
		   " Your file %s was not sent.", me.name, from->name, to->name,
		   tmpext, tmpfn);

	sendto_one(to,
		   ":%s NOTICE %s :%s (%s@%s) has attempted to send you a file named %s, which was blocked.",
		   me.name, to->name, from->name, from->user->username,
		   from->user->host, tmpfn);

	if (!SeenDCCNotice(to)) {
	    SetDCCNotice(to);

	    sendto_one(to,
		       ":%s NOTICE %s :The majority of files sent of this type are malicious virii and trojan horses."
		       " In order to prevent the spread of this problem, we are blocking DCC sends of these types of"
		       " files by default.", me.name, to->name);
	    sendto_one(to,
		       ":%s NOTICE %s :If you trust %s, and want him/her to send you this file, you may obtain"
		       " more information on using the dccallow system by typing /dccallow help",
		       me.name, to->name, from->name, to->name);
	}

	for (tlp = to->user->channel; tlp && !chptr; tlp = tlp->next) {
	    for (flp = from->user->channel; flp && !chptr; flp = flp->next) {
		if (tlp->value.chptr == flp->value.chptr)
		    chptr = tlp->value.chptr;
	    }
	}

	if (chptr)
	    sendto_realops_lev(DCCSEND_LEV,
			       "%s (%s@%s) sending forbidden filetyped file %s to %s (channel %s)",
			       from->name, from->user->username,
			       from->user->host, tmpfn, to->name,
			       chptr->chname);
	else
	    sendto_realops_lev(DCCSEND_LEV,
			       "%s (%s@%s) sending forbidden filetyped file %s to %s",
			       from->name, from->user->username,
			       from->user->host, tmpfn, to->name);

	return 1;
    }

    return 0;
}

/*
 * This function checks to see if a CTCP message (other than ACTION) is
 * contained in the passed string.  This might seem easier than I am
 * doing it, but a CTCP message can be changed together, even after a
 * normal message.  
 *     
 * If the message is found, and it's a DCC message, pass it back in
 * *dccptr.
 *
 * Unfortunately, this makes for a bit of extra processing in the
 * server.
 */

int check_for_ctcp(char *str, char **dccptr)
{
    char *p = str;

    while ((p = strchr(p, 1)) != NULL) {
	if (myncmp(++p, "DCC", 3) == 0) {
	    if (dccptr)
		*dccptr = p;
	    if (myncmp(p + 3, " SEND", 5) == 0)
		return CTCP_DCCSEND;
	    else
		return CTCP_DCC;
	}
	if (myncmp(++p, "ACTION", 6) != 0)
	    return CTCP_YES;
	if ((p = strchr(p, 1)) == NULL)
	    return CTCP_NONE;
	if (!(*(++p)))
	    break;;
    }
    return CTCP_NONE;
}

int add_dccallow(aClient *sptr, aClient *optr)
{
    Link *lp;
    int cnt = 0;

    for (lp = sptr->user->dccallow; lp; lp = lp->next) {
	if (lp->flags != DCC_LINK_ME)
	    continue;
	if (++cnt >= MAXDCCALLOW) {
	    sendto_one(sptr, err_str(ERR_TOOMANYDCC), me.name, sptr->name,
		       optr->name, MAXDCCALLOW);
	    return 0;
	} else if (lp->value.cptr == optr) {
	    /*
	     * silently return 
	     */
	    return 0;
	}
    }

    lp = make_link();
    lp->value.cptr = optr;
    lp->flags = DCC_LINK_ME;
    lp->next = sptr->user->dccallow;
    sptr->user->dccallow = lp;

    lp = make_link();
    lp->value.cptr = sptr;
    lp->flags = DCC_LINK_REMOTE;
    lp->next = optr->user->dccallow;
    optr->user->dccallow = lp;

    sendto_one(sptr, rpl_str(RPL_DCCSTATUS), me.name, sptr->name, optr->name,
	       "added to");
    return 0;
}

int del_dccallow(aClient *sptr, aClient *optr)
{
    Link **lpp, *lp;
    int found = 0;

    for (lpp = &(sptr->user->dccallow); *lpp; lpp = &((*lpp)->next)) {
	if ((*lpp)->flags != DCC_LINK_ME)
	    continue;

	if ((*lpp)->value.cptr == optr) {
	    lp = *lpp;
	    *lpp = lp->next;
	    free_link(lp);
	    found++;
	    break;
	}
    }

    if (!found) {
	sendto_one(sptr, ":%s %d %s :%s is not in your DCC allow list",
		   me.name, RPL_DCCINFO, sptr->name, optr->name);
	return 0;
    }

    for (found = 0, lpp = &(optr->user->dccallow); *lpp; lpp = &((*lpp)->next)) {
	if ((*lpp)->flags != DCC_LINK_REMOTE)
	    continue;

	if ((*lpp)->value.cptr == sptr) {
	    lp = *lpp;
	    *lpp = lp->next;
	    free_link(lp);
	    found++;
	    break;
	}
    }
    if (!found)
	sendto_realops_lev(DEBUG_LEV,
			   "%s was in dccallowme list of %s but not in dccallowrem list!",
			   optr->name, sptr->name);

    sendto_one(sptr, rpl_str(RPL_DCCSTATUS), me.name, sptr->name, optr->name,
	       "removed from");

    return 0;
}

int allow_dcc(aClient *to, aClient *from)
{
    Link *lp;

    for (lp = to->user->dccallow; lp; lp = lp->next) {
	if (lp->flags == DCC_LINK_ME && lp->value.cptr == from)
	    return 1;
    }
    return 0;
}

int m_dccallow(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    Link *lp;
    char *p, *s;
    char *cn;
    aClient *acptr, *lastcptr = NULL;
    int didlist = 0, didhelp = 0, didanything = 0;
    char **ptr;
    static char *dcc_help[] = {
	"/DCCALLOW [<+|->nick[,<+|->nick, ...]] [list] [help]",
	"You may allow DCCs of filetypes which are otherwise blocked by the IRC server",
	"by specifying a DCC allow for the user you want to recieve files from.",
	"For instance, to allow the user bob to send you file.exe, you would type:",
	"/dccallow +bob",
	"and bob would then be able to send you files. bob will have to resend the file",
	"if the server gave him an error message before you added him to your allow list.",
	"/dccallow -bob",
	"Will do the exact opposite, removing him from your dcc allow list.",
	"/dccallow list",
	"Will list the users currently on your dcc allow list.",
	NULL
    };

    if (!MyClient(sptr))
	return 0;		/* don't accept dccallows from servers or
				 * clients that aren't mine.. */

    if (parc < 2) {
	sendto_one(sptr,
		   ":%s NOTICE %s :No command specified for DCCALLOW. Type /dccallow help for more information.",
		   me.name, sptr->name);
	return 0;
    }

    for (p = NULL, s = strtoken(&p, parv[1], ", "); s;
	 s = strtoken(&p, NULL, ", ")) {
	if (*s == '+') {
	    didanything++;
	    cn = s + 1;
	    if (*cn == '\0')
		continue;

	    acptr = find_person(cn, NULL);

	    if (acptr == sptr)
		continue;

	    if (!acptr) {
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, sptr->name,
			   cn);
		continue;
	    }

	    if (lastcptr == acptr)
		sendto_realops_lev(SPY_LEV,
				   "User %s (%s@%s) may be flooding dccallow: add %s",
				   sptr->name, sptr->user->username,
				   sptr->user->host, acptr->name);

	    lastcptr = acptr;
	    add_dccallow(sptr, acptr);
	} else if (*s == '-') {
	    didanything++;
	    cn = s + 1;
	    if (*cn == '\0')
		continue;

	    acptr = find_person(cn, NULL);

	    if (acptr == sptr)
		continue;

	    if (!acptr) {
		sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, sptr->name,
			   cn);
		continue;
	    }

	    if (lastcptr == acptr)
		sendto_realops_lev(SPY_LEV,
				   "User %s (%s@%s) may be flooding dccallow: del %s",
				   sptr->name, sptr->user->username,
				   sptr->user->host, acptr->name);

	    lastcptr = acptr;
	    del_dccallow(sptr, acptr);
	} else {
	    if (!didlist && myncmp(s, "list", 4) == 0) {
		didanything++;
		didlist++;
		sendto_one(sptr,
			   ":%s %d %s :The following users are on your dcc allow list:",
			   me.name, RPL_DCCINFO, sptr->name);
		for (lp = sptr->user->dccallow; lp; lp = lp->next) {
		    if (lp->flags == DCC_LINK_REMOTE)
			continue;
		    sendto_one(sptr, ":%s %d %s :%s (%s@%s)", me.name,
			       RPL_DCCLIST, sptr->name, lp->value.cptr->name,
			       lp->value.cptr->user->username,
#ifdef USE_FAKEHOST
			       IsFake(lp->value.cptr) ? lp->value.cptr->user->
			       fakehost : lp->value.cptr->user->host);
#else
			       lp->value.cptr->user->host);
#endif
		}
		sendto_one(sptr, rpl_str(RPL_ENDOFDCCLIST), me.name, sptr->name,
			   s);
	    } else if (!didhelp && myncmp(s, "help", 4) == 0) {
		didanything++;
		didhelp++;
		for (ptr = dcc_help; *ptr; ptr++)
		    sendto_one(sptr, ":%s %d %s :%s", me.name, RPL_DCCINFO,
			       sptr->name, *ptr);
		sendto_one(sptr, rpl_str(RPL_ENDOFDCCLIST), me.name, sptr->name,
			   s);
	    }
	}
    }

    if (!didanything) {
	sendto_one(sptr,
		   ":%s NOTICE %s :Invalid syntax for DCCALLOW. Type /dccallow help for more information.",
		   me.name, sptr->name);
	return 0;
    }

    return 0;
}
